/** The module name. */
#define CRPT_DEV_NAME         "itcscrpt"

#ifdef ITCS_CA_USE
/*CA support*/
struct ca_hw_info;
GCC_NOREGPARM extern void CaClientGlobalNotifier( const struct ca_hw_info* info );
EXPORT_SYMBOL(CaClientGlobalNotifier);
#endif

extern int GCC_NOREGPARM crptdevice_open(VN_STINODE inode, VN_STFILE file);
extern int GCC_NOREGPARM crptdevice_release(VN_STINODE inode, VN_STFILE file);
extern int GCC_NOREGPARM crptdevice_ioctl( int cmd, unsigned long arg );
/*
** stubs to fill file_operations struct while charaster device register
** (see comments about CONFIG_REGPARM)
*/
int my_crptdevice_open(struct inode *inode, struct file *file)
{
    return crptdevice_open( (VN_STINODE)inode, (VN_STFILE)file);
}

int my_crptdevice_release(struct inode *inode, struct file *file)
{
    return crptdevice_release( (VN_STINODE)inode, (VN_STFILE)file);
}

#ifdef HAVE_UNLOCKED_IOCTL
long my_crptdevice_ioctl( struct file *file, unsigned int cmd, unsigned long arg)
{
    return crptdevice_ioctl(cmd, arg);
}
#else
int my_crptdevice_ioctl( struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
    return crptdevice_ioctl(cmd, arg);
}
#endif

 /** The file_operations structure. */
static struct file_operations gFileOpsCrpt =
{
owner:      THIS_MODULE,
open:       my_crptdevice_open,
release:    my_crptdevice_release,
#ifdef HAVE_UNLOCKED_IOCTL
unlocked_ioctl: my_crptdevice_ioctl,
#else
ioctl:      my_crptdevice_ioctl,
#endif
#ifdef HAVE_COMPAT_IOCTL
compat_ioctl: my_crptdevice_ioctl,
#endif
};

/** The miscdevice structure. */
static struct miscdevice g_CrptDev =
{
minor:      MISC_DYNAMIC_MINOR,
name:       CRPT_DEV_NAME,
fops:       &gFileOpsCrpt,
# if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 17)
devfs_name: CRPT_DEV_NAME
# endif
};

static int itcscrpt_major = -1;
module_param(itcscrpt_major, int, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP);
MODULE_PARM_DESC(itcscrpt_major, "Major number for cryptodriver");

int use_gost_ctr = -1;
module_param(use_gost_ctr, int, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP);
MODULE_PARM_DESC(use_gost_ctr, "Use Gost_CTR mode (0|1)");

char* get_crypto_handle(void );
EXPORT_SYMBOL(get_crypto_handle);

extern int real_crptdrv_init(void);
extern void real_crptdrv_exit(void);

int __init crptdrv_init(void)
{
    int rc=0;

    if (use_gost_ctr != -1 )
    {
        if ((use_gost_ctr != 0) && (use_gost_ctr != 1))
        {
            myprintk("CryptoDrv: Option use GOST CTR is incorrectly specified (%d)\n",
                     use_gost_ctr);
            return myENODEV();
        }
    }

#ifdef CONFIG_ITCSDRV_AS_MISC
    rc = misc_register(&g_CrptDev);
    if (rc)
    {
        printk(KERN_ERR " Can't register " CRPT_DEV_NAME " device! rc=%d\n", rc);
        return rc;
    }
#else  /* !CONFIG_ITCSDRV_AS_MISC */
    /*
     * Register character device.
     */

    if (itcscrpt_major != -1)
    {
        if ((itcscrpt_major <= 0) || (itcscrpt_major > 254))
        {
            myprintk("CryptoDrv: device major number is incorrectly specified (%d)\n",
                     itcscrpt_major);
            return myENODEV();
        }
        dev_major_number = itcscrpt_major;
    }

    rc = register_chrdev((dev_t)dev_major_number, CRPT_DEV_NAME, &gFileOpsCrpt);
    if (rc < 0)
    {
        myprintk("CryptoDrv: cannot register device (code %d)\n", rc);
        return myENODEV();
    }

    /*
     * Save returned module major number
     */
    if ( dev_major_number == 0 )
        dev_major_number = rc;
    rc = 0;

# ifdef CONFIG_DEVFS_FS
    /*
     * Register a device entry
     */
    if (devfs_mk_cdev(MKDEV(dev_major_number, 0), S_IFCHR | ITCS_DEV_FMASK, CRPT_DEV_NAME) != 0)
    {
        myprintk("devfs_register failed!\n");
        return myENODEV();
    }
# endif
    myprintk("CryptoDrv: device number = %d, device node = /dev/"
             CRPT_DEV_NAME "\n", dev_major_number);

#endif /* !CONFIG_ITCSDRV_AS_MISC */

    rc = real_crptdrv_init();
    if  ( rc)
    {
        /** Failed, cleanup and return the error code.
         */
        myprintk("real_crpt_init returning %#x)\n", rc);
#if defined(CONFIG_DEVFS_FS) && !defined(CONFIG_ITCSDRV_AS_MISC)
        devfs_remove(CRPT_DEV_NAME);
#endif

#ifdef CONFIG_ITCSDRV_AS_MISC
        misc_deregister(&g_CrptDev);
#else
        unregister_chrdev(dev_major_number, CRPT_DEV_NAME);
#endif
        return rc;
    }
    myprintk("ViPNet Crypto Driver successfully loaded\n");

    return rc;
}

static void  __exit crptdrv_exit(void)
{
    int                 rc;
    real_crptdrv_exit();

    myprintk("Unload\n");
    NOREF(rc);

/** I Don't think it's possible to unload a driver which processes have
 * opened, at least we'll blindly assume that here.
 */
#ifdef CONFIG_ITCSDRV_AS_MISC
    rc = misc_deregister(&g_CrptDev);
    if (rc < 0)
    {
        myprintk("misc_deregister failed with rc=%#x\n", rc);
    }
#else  /* !CONFIG_ITCSDRV_AS_MISC */
# ifdef CONFIG_DEVFS_FS
    /*
     * Unregister a device entry
     */
    devfs_remove(CRPT_DEV_NAME);
# endif /* devfs */
    unregister_chrdev(dev_major_number, CRPT_DEV_NAME);
#endif /* !CONFIG_ITCSDRV_AS_MISC */
}

module_init(crptdrv_init);
module_exit(crptdrv_exit);
