/* Turn on debug info */
#if !defined(__arm__)
#define DEBUG_SHOW_STACK
#endif

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
#include <linux/config.h>
#else
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,33)
#include  <linux/autoconf.h>
#else
#include  <generated/autoconf.h>
#endif
#endif

// For correct VMALLOC_START(END) macro interpretation on old kernels
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,24)
#ifdef CONFIG_HIGHMEM
#include <asm/highmem.h>
#endif
#endif

#include <linux/netdevice.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/interrupt.h>

/* check kernel smp */
#ifdef CONFIG_SMP
#include <asm/smp.h>
#define USE_RWLOCK
#endif /* #ifdef CONFIG_SMP */

/* check atomic routines */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,31)
#    include <asm-generic/atomic-long.h>
#    define HAVE_KERNEL_ATOMIC_LONG
#    define HAVE_KERNEL_ATOMIC_LONG_RETURN
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,22)
#    include <asm-generic/atomic.h>
#    define HAVE_KERNEL_ATOMIC_LONG
#    define HAVE_KERNEL_ATOMIC_LONG_RETURN
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,17)
#    include <asm-generic/atomic.h>
#    define HAVE_KERNEL_ATOMIC_LONG
#endif

#ifdef CONFIG_IPLIR
#  ifndef IPLIR_PATCH_VERSION
#    define IPLIR_PATCH_VERSION 1
#  endif
#endif

#include <asm/unistd.h>
#include <asm/bitops.h>
#include <asm/div64.h>
#include <linux/etherdevice.h>
#include <linux/inetdevice.h>
#include <linux/vmalloc.h>
#include <linux/pci.h>                 // PCI #defs
#include "asm/io.h"
#include "asm/uaccess.h"
#include <linux/string.h>
#include <asm/unistd.h>
#include <net/neighbour.h>
#include <net/arp.h>
#include <net/route.h>
#include <linux/kthread.h>
#include <net/flow.h>
#include <net/xfrm.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,25)
#include <net/net_namespace.h>
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,39)
#include <linux/smp_lock.h>
#endif

#include <linux/rwsem.h>
#include <net/dst.h>
#include <linux/rtnetlink.h>
#include <linux/types.h>
#include <linux/proc_fs.h>
#include <linux/sysrq.h>
#include <linux/udp.h>
#include <linux/netdevice.h>
#include <linux/ip.h>
#include <linux/in.h>
#include <net/udp.h>
#include <net/tcp.h>
#include <linux/jhash.h>
#include <net/netns/hash.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,9,0)
#include <linux/sched/rt.h>
#else
#include <linux/sched.h>
#endif

#ifdef	CONFIG_LOCKMETER
#include <linux/lockmeter.h>
#endif	//CONFIG_LOCKMETER

#define IS_KERNEL_SUPPORTED_BY_IPCLS(kver)\
	(kver >= KERNEL_VERSION(3,10,0))

/* Since 2.6.25 netfilter was rewritten, structure names changed. */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,25)
#  define NF_USES_QUEUE_ENTRY
/* We start filtering non-IP protocols from 2.6.25 as well. */
#  define NF_FILTER_NONIP
#  if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,28)
#    define NF_FILTER_ARP
#  endif
#endif

#if IS_KERNEL_SUPPORTED_BY_IPCLS(LINUX_VERSION_CODE)
#	define KERN_IPCLS // ipcls(ipv6) friendly kernel
#endif

#ifdef NF_USES_QUEUE_ENTRY
/*
** In new kernels nf_reinject only takes 2 parameters, skb is not used
** (it is actually inside nfq). This macro provides protable way of
** reinjecting. --vel
*/
#  define NF_REINJECT_COMPAT(skb, nfq, verdict) nf_reinject((struct nf_queue_entry *)nfq, verdict)
#else
/* For old kernels, link new constant names to old ones. --vel */
#  define NF_INET_PRE_ROUTING NF_IP_PRE_ROUTING
#  define NF_INET_POST_ROUTING NF_IP_POST_ROUTING
#  define NF_REINJECT_COMPAT(skb, nfq, verdict) nf_reinject((struct sk_buff *)skb, (struct nf_info *)nfq, verdict)
#endif

#include <linux/cpu.h>
#include <linux/cpumask.h>

#include "drvkrniface.h"
#include "itcskrniface.h"

#include <linux/wait.h>
typedef struct {
  VN_WAITQUEUE wq;
  int condition;
} WQHANDLE, * WQHANDLEPTR;

struct mytimer_info
{
     void (*fn)(void *);
     void *arg;
     struct work_struct     ws;
};

#  include <linux/spinlock.h>
#  include <linux/slab.h>
#  define LINUX_DEVICE struct net_device
#  define LINUX_WAIT_QUEUE wait_queue_head_t
#  ifdef CONFIG_NETFILTER
#    include <linux/netfilter.h>
#    include <linux/netfilter_ipv4.h>
#    ifdef NF_USES_QUEUE_ENTRY
#      include <net/netfilter/nf_queue.h>
#      include <linux/netfilter_arp.h>
#      include <linux/netfilter_ipv6.h>
#    endif
#  endif

#  include <net/ip.h>

#  include <linux/moduleparam.h>
#  include <linux/kallsyms.h>
#  define MOD_INC_USE_COUNT_LALA
#  define MOD_DEC_USE_COUNT_LALA

#ifdef DEBUG_F1024
#include <asm/bug.h>
#endif

#ifndef LOCK_PREFIX
#ifdef CONFIG_SMP
#define LOCK_PREFIX "lock ; "
#else
#define LOCK_PREFIX ""
#endif
#endif

// SSE support

#ifdef CONFIG_X86
#    include <asm/i387.h>
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#if defined(__NR_query_module)
#    define HAVE_MYMODLOOKUP
#endif /* #if defined(__NR_query_module) */
#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0) */

/*****************************************************************************/
/* kernel_skb_ip_hdr(skb) */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
#    define kernel_skb_ip_hdr(s) (s)->nh.iph
#else
#    define kernel_skb_ip_hdr(s) ip_hdr(s)
#endif

/* kernel_skb_dst(skb) */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,31)
#    define kernel_skb_dst(s) (s->dst)
#else
#    define kernel_skb_dst(s) (skb_dst(s))
#endif

/* kernel_skb_dst_metrics(skb) */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,38)
#    define kernel_skb_dst_metrics(s) (kernel_skb_dst(s)->metrics)
#else
#    define kernel_skb_dst_metrics(s) (kernel_skb_dst(s)->_metrics)
#endif

/* kernel_skb_dst_is_empty(skb) */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,31)
#    define kernel_skb_dst_is_empty(s) ((s)->dst == 0)
#else
#    define kernel_skb_dst_is_empty(s) (skb_dst(s) == NULL)
#endif

/* kernel_skb_dst_set(skb, dst) */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,31)
#    define kernel_skb_dst_set(s, d) (s)->dst = (d)
#else
#    define kernel_skb_dst_set(s, d) skb_dst_set((s), (d))
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,31)
#define kernel_skb_dst_drop(skb) dst_release((skb)->dst)
#else
#define kernel_skb_dst_drop(skb) skb_dst_drop(skb);
#endif

/* kernel_rtable_dst(rt) */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,36)
#    define kernel_rtable_dst(rt) &((rt)->u.dst)
#else
#    define kernel_rtable_dst(rt) &((rt)->dst)
#endif

/* kernel_rtable_dev(rt) */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#    define kernel_rtable_dev(rt) ((rt)->u.dst.dev)
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2,6,38)
#    define kernel_rtable_dev(rt) ((rt)->idev->dev)
#else
#    define kernel_rtable_dev(rt) ((rt)->dst.dev)
#endif

/* kernel_skb_dst(skb) */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,31)
#    define kernel_skb_dst(s) (s->dst)
#else
#    define kernel_skb_dst(s) (skb_dst(s))
#endif

/* kernel_skb_dst_metrics(skb) */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,38)
#    define kernel_skb_dst_metrics(s) (kernel_skb_dst(s)->metrics)
#else
#    define kernel_skb_dst_metrics(s) (kernel_skb_dst(s)->_metrics)
#endif

/* kernel_skb_dst_is_empty(skb) */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,31)
#    define kernel_skb_dst_is_empty(s) ((s)->dst == 0)
#else
#    define kernel_skb_dst_is_empty(s) (skb_dst(s) == NULL)
#endif

/* kernel_skb_dst_set(skb, dst) */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,31)
#    define kernel_skb_dst_set(s, d) (s)->dst = (d)
#else
#    define kernel_skb_dst_set(s, d) skb_dst_set((s), (d))
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,31)
#define kernel_skb_dst_drop(skb) dst_release((skb)->dst)
#else
#define kernel_skb_dst_drop(skb) skb_dst_drop(skb);
#endif

/* kernel_rtable_dst(rt) */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,36)
#    define kernel_rtable_dst(rt) &((rt)->u.dst)
#else
#    define kernel_rtable_dst(rt) &((rt)->dst)
#endif

/* kernel_rtable_dev(rt) */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,38)
#    define kernel_rtable_dev(rt) ((rt)->idev->dev)
#else
#    define kernel_rtable_dev(rt) ((rt)->dst.dev)
#endif

#if LINUX_VERSION_CODE <  KERNEL_VERSION(3,0,21)
#define DST_NEIGHBOUR(dst)    dst->neighbour
#elif  LINUX_VERSION_CODE < KERNEL_VERSION(3,3,0)
#define DST_NEIGHBOUR(dst)    dst_get_neighbour(dst)
#elif LINUX_VERSION_CODE < KERNEL_VERSION(3,6,0)
#define DST_NEIGHBOUR(dst)    dst_get_neighbour_noref(dst)
#else // Modern kernel
#define DST_NEIGHBOUR(dst)    dst_neigh_lookup_skb(dst, skb)
#endif

#if LINUX_VERSION_CODE <  KERNEL_VERSION(3,1,0)
#define DST_NEIGHBOUR_OUTPUT(nghb)    nghb->output(skb);
#elif LINUX_VERSION_CODE <  KERNEL_VERSION(3,6,0)
#define DST_NEIGHBOUR_OUTPUT(nghb)    neigh_output(nghb, skb);
#else
#define DST_NEIGHBOUR_OUTPUT(nghb)    nghb->output(nghb, skb);
#endif

typedef struct
{
#ifdef USE_RWLOCK
     rwlock_t lock;             // Read-Write lock itself
#else
     struct rw_semaphore lock;  // Read-Write lock itself
#ifdef DEBUG_LOCK               // Define is used to debug mutex and rwlock. It adds additional print and breaks deadlocks.
     int count;                 // Amount of nesting locks. Protected by "struct_lock"
     int pid;                   // PID of the process that initiated recursive lock. Protected by "struct_lock"
     spinlock_t struct_lock;    // Spinlock for this struct
     int owner_pid;             // PID of process that performed actual lock. Can be overwritten after successfull recursive lock. Protected by "lock"
     unsigned int owner_ip;     // Address where lock was acquired. Protected by "lock"
     int read;                  // -1 - actual lock type is undefined. 0 - actual lock is write. 1 - actual lock is read
#endif
#endif
} MYRWLOCK, * VN_MYRWLOCK;

#ifdef IFACE_ITCSIPCLS
#define KERNEL_MEMORY_LEAK_TEST 1
#endif

#ifdef KERNEL_MEMORY_LEAK_TEST
unsigned long mem_allocated=0;
unsigned long mem_alloc_count=0;
unsigned long mem_free_count=0;
unsigned long common_alloc_count=0;
unsigned long common_free_count=0;
struct list_head mem_head;
struct list_head common_head;
spinlock_t mem_head_lock;
spinlock_t common_head_lock;

struct mem_list
{    // NOTE: list_head must be first field of struct mem_list,
     // becouse we will use direct conversion from (mem_list*) to (list_head*)
     // and backwards.
     struct list_head list;
     unsigned int size;
     void *ptr;
     void *func;
};

static inline struct mem_list* create_mem_list( void *ptr, unsigned int size )
{
     struct mem_list *q = kmalloc( sizeof( struct mem_list ), GFP_ATOMIC);
     if ( q )
     {
          INIT_LIST_HEAD( &q->list );
          q->ptr = ptr;
          q->size = size;
     }
     return q;
}
#endif

/*****************************************************************************/
/* ok functions for netfilter hooks */
int (*myip_rcv_finish)(struct sk_buff *skb) = 0;
int (*myip_finish_output)(struct sk_buff *skb) = 0;
/* define port where we will get ip_rcv_finish & ip_finish_output */
#define IPLIRCFG_PORT        2046
/*****************************************************************************/


#ifdef IFACE_DRVIPLIR
#include "iplir.c"

#if defined ( CONFIG_IPLIR_XFRM_INPUT ) || defined ( IPLIR_XFRM_INPUT )
#include "xfrmiface.c"
#include "xfrmiface_udp.c"
#endif

#endif /* IFACE_DRVIPLIR */

#ifdef IFACE_ITCSARP
char * itcsarp_major = NULL;

module_param(itcsarp_major, charp, 0);
#endif //IFACE_ITCSARP

#ifdef IFACE_ITCSCRPT
#include "crpt.c"
#endif //IFACE_ITCSCRPT

#ifdef IFACE_ITCSWD
#include "wd.c"
#endif /* IFACE_ITCSWD */

#ifdef IFACE_ITCSIPCLS
#include "ipcls.c"
#endif /* IFACE_ITCSCLS */

MODULE_AUTHOR("JSC Infotecs, 1992-2013. All rights reserved.");
#if defined(DEBUG_F1024) || defined(DEBUG_LOCK) || defined(CONFIG_DEBUG_KERNEL)
MODULE_LICENSE("GPL");
#else
MODULE_LICENSE("Infotecs");
#endif
/*****************************************************************************/

typedef void (*FILTFUNC)(void *, void *);
typedef void * GCC_NOREGPARM (*NF_DRV_FUNC)(VN_SKB, VN_NFINFO);

#if defined(KERNEL_USE_REGPARM)
/* universal handle implementation (see comments about CONFIG_REGPARM) */
static int univ_handle(void * data)
{
  struct _univ_handle * handle;
  int ret;

  handle = (struct _univ_handle *)data;
  if (!handle)
  {
    udebug(( "itcs_drv: bad univ handle data!" ));
    return 0;
  }
  if (handle->fn)
  {
//    myprintk("%s: data %08X, fn %08X, arg %08X\n", __FUNCTION__, (int)data, (int)handle->fn, (int)handle->arg);
    ret = handle->fn(handle->arg);
    if (handle->flclean)
    {
      udebug3(( "itcs_drv: free univ handle data" ));
      mykfree(handle);
    }
    return ret;
  }
  else
  {
    udebug(( "itcs_drv: bad univ handle!" ));
    return 0;
  }
}
#endif

void* GCC_NOREGPARM mysymbol_get( const char *symbol)
{
     return itcs_symbol_get( symbol );
}

void GCC_NOREGPARM mysymbol_put( const char *symbol)
{
     itcs_symbol_put( symbol );
}

/*****************************************************************************/
void mycli(void)
{
}

/*****************************************************************************/
void mysti(void)
{
}

/*****************************************************************************/

void mymemleaktest_init(void)
{
#ifdef KERNEL_MEMORY_LEAK_TEST
    INIT_LIST_HEAD( &mem_head );
    INIT_LIST_HEAD( &common_head );
    spin_lock_init( &mem_head_lock );
    spin_lock_init( &common_head_lock );
#endif
  return;
}

void mymemleaktest_clean(void)
{
#ifdef KERNEL_MEMORY_LEAK_TEST
  struct list_head *temp;
  spin_lock_bh( &mem_head_lock );
  while( mem_head.next != &mem_head )
  {
    temp = mem_head.next;
    list_del( mem_head.next );
    kfree( temp );
  }
  INIT_LIST_HEAD( &mem_head );
  spin_unlock_bh( &mem_head_lock );
  spin_lock_bh( &common_head_lock );
  while( common_head.next != &common_head )
  {
    temp = common_head.next;
    list_del( common_head.next );
    kfree( temp );
  }
  INIT_LIST_HEAD( &common_head );
  spin_unlock_bh( &common_head_lock );
#endif
  return;
}

void mymemleaktest(void)
{
#ifdef KERNEL_MEMORY_LEAK_TEST
  unsigned long size = 0;
  unsigned long common_size = 0;
  struct list_head *pos;

  spin_lock_bh( &mem_head_lock );
  list_for_each_prev( pos, &mem_head )
  {
    size += ( (struct mem_list *) pos )->size;
    print_symbol("The memory was allocated in %s and then not released\n", (unsigned long)((struct mem_list *)pos)->func);
  }
  spin_unlock_bh( &mem_head_lock );
  spin_lock_bh( &common_head_lock );
  list_for_each_prev( pos, &common_head )
  {
    common_size += ( (struct mem_list *) pos )->size;
    print_symbol("The common memory was allocated in %s and then not released\n", (unsigned long)((struct mem_list *)pos)->func);
  }
  spin_unlock_bh( &common_head_lock );

  myprintk("COMMON KERNEL MEMORY ALLOCATED:\n  %u bytes in %d block (Allocated:%u Free:%u)\n",
          common_size,  common_alloc_count - common_free_count, common_alloc_count, common_free_count);
  myprintk("LINUX  KERNEL MEMORY ALLOCATED:\n  %u bytes in %d block (Allocated:%u Free:%u)\n",
          size, mem_alloc_count - mem_free_count, mem_alloc_count, mem_free_count );
#endif
#ifdef SLAB_LEAK_TEST
  mykmem_print_stats();
#endif
  return;
}

#if defined(USE_RWLOCK) || !defined(USE_PACKETQUEUE)
#define	__MY_GFP_FLAGS	GFP_ATOMIC
#else
#define	__MY_GFP_FLAGS	GFP_KERNEL | __GFP_NOWARN
#endif

void * GCC_NOREGPARM mymalloc(unsigned int size)	
{
	/*	mymalloc() is used by drvlib for STL and other C++ crap in common code,
		and I hope not for any skb or system's data (kernel API) allocation	*/

	void* p;

	if (!size)
		return NULL;
	do
	{
		/*first try statisfy request by kmalloc()*/
		p = kmalloc( size, __MY_GFP_FLAGS );
		if( NULL != p )
			break;

#ifndef CONFIG_IPLIR_XFRM_INPUT
          /*in other cases we use read_lock_bh(), but vmalloc() can't be used in_interrrupt() context*/

		/* vmalloc() allocate memory on page granularity*/
		/*kmalloc relay on slab/slub allocator. So even request for small memory area can involve allocation of large contigous memory,
		  so we try use vamlloc() for any size of mymalloc() request, but it is waste of VA space and physical memory too. */
		p = __vmalloc( size, __MY_GFP_FLAGS, PAGE_KERNEL );
		if( NULL != p )
			break;
#endif

#ifndef USE_PACKETQUEUE
		myprintk("can't get %d bytes of free kmemory (atomic)!\n", size);
#else
		myprintk("can't get %d bytes of free kmemory (kernel)!\n", size);
#endif
#ifdef KERNEL_MEMORY_LEAK_TEST
		myprintk("total allocated %u bytes.\n", mem_allocated);
#endif
		return NULL;
	}
	while( 0 );

#ifdef KERNEL_MEMORY_LEAK_TEST	/*it's totaly broken by design*/
     {
          struct mem_list* const temp = create_mem_list( p, size );
          spin_lock_bh( &common_head_lock );
          ++common_alloc_count;
          if ( temp )
          {
               mem_allocated += size;
               temp->func = __builtin_return_address(0);
               list_add_tail( (struct list_head *) temp, &common_head );
          }
          spin_unlock_bh( &common_head_lock );
	}
#endif

	return p;
}
/*****************************************************************************/
void GCC_NOREGPARM myfree(void * obj)
{
	if (obj)
	{
#ifdef KERNEL_MEMORY_LEAK_TEST
          struct list_head *pos;
          struct list_head *pos_del = NULL;
          spin_lock_bh( &common_head_lock );
          ++common_free_count;
          list_for_each_prev( pos, &common_head )
          {
               if( ((struct mem_list*) pos)->ptr == obj )
               {
                    mem_allocated -= ( (struct mem_list *) pos )->size;
                    list_del( pos );
                    pos_del = pos;
                    break;
               }
          }
          spin_unlock_bh( &common_head_lock );
          if ( pos_del )
               kfree( pos_del );
#endif
#ifndef CONFIG_IPLIR_XFRM_INPUT
          /*in other cases we use read_lock_bh(), but vmalloc() can't be used in_interrrupt() context*/
		if( ( obj >= VMALLOC_START) && ( obj < VMALLOC_END ) )
			vfree( obj );
		else
#endif
			kfree( obj );
	}
}

/*****************************************************************************/
/* We need to check for 0 because of empty spinlock struct on UP systems. */
void * GCC_NOREGPARM mykmalloc(unsigned int size)
{
#ifdef KERNEL_MEMORY_LEAK_TEST
  struct mem_list *temp;
#endif
  void * p;
  if (!size)
    return NULL;

//  if (size > 131072)
//  {
//    myprintk("try allocating %d bytes of kmemory with limit of 128 Kb!\n", size);
//    return NULL;
//  }
  p = kmalloc(size, GFP_ATOMIC);
  if (!p)
  {
    myprintk("can't get %d bytes of free kmemory!\n", size);
#ifdef KERNEL_MEMORY_LEAK_TEST
    myprintk("total allocated %u bytes.\n", mem_allocated);
#endif
    return NULL;
  }

#ifdef KERNEL_MEMORY_LEAK_TEST
  temp = create_mem_list( p, size );
  spin_lock_bh( &mem_head_lock );
  ++mem_alloc_count;
  if ( temp )
  {
       mem_allocated += size;
       temp->func = __builtin_return_address(0);
       list_add_tail( (struct list_head *) temp, &mem_head );
  }
  spin_unlock_bh( &mem_head_lock );
#endif

  return p;
}
/*****************************************************************************/
void GCC_NOREGPARM mykfree(void * obj)
{
  if (obj)
  {
#ifdef KERNEL_MEMORY_LEAK_TEST
    struct list_head *pos = NULL;
    struct list_head *pos_del = NULL;
    spin_lock_bh( &mem_head_lock );
    ++mem_free_count;
    list_for_each_prev( pos, &mem_head )
    {
      if( ((struct mem_list*) pos)->ptr == obj )
      {
        mem_allocated -= ( (struct mem_list *) pos )->size;
        list_del( pos );
        pos_del = pos;
        break;
      }
    }
    spin_unlock_bh( &mem_head_lock );
    if ( pos_del )
        kfree( pos_del );
#endif
    kfree(obj);
  }
}
/*****************************************************************************/
/*
** Note that we are allowed to use this only if we are 100% sure that we are
** in user context. Never ever use it for spinlocks and such. --vel
*/
void * GCC_NOREGPARM myvmalloc(unsigned int size)
{
  void * p;
  p = vmalloc(size);
  if (size && !p)
  {
    myprintk("can't get %d bytes of free vmemory!\n", size);
    return NULL;
  }
  return p;
}
/*****************************************************************************/
void GCC_NOREGPARM myvfree(void * obj)
{
  if (obj)
    vfree ( obj );
}
/*****************************************************************************/
#if defined(IFACE_DRVIPLIR)
#    define DRVSELFNAME "drviplir"
#elif defined(IFACE_ITCSWD)
#    define DRVSELFNAME "itcswd"
#elif defined(IFACE_ITCSCRPT)
#    define DRVSELFNAME "itcscrpt"
#elif defined(IFACE_ITCSIPCLS)
#    define DRVSELFNAME "itcsipcls"
#else
#    error "What is my name ?"
#endif

int myprintk(const char * fmt, ...)
{
     char buf[768];
     va_list alist;

     /* We format message now */
     va_start(alist, fmt);
     vsnprintf(buf, sizeof(buf), fmt, alist);
     va_end(alist);

     /* All printk messages can have prefix like <X>,
      * where X is a char 0..7, d, c.
      * The KERN_ALERT macro correspond to the `<1>' prefix.
      * Also, you can see for linux/printk.h header.
      */
     return printk(KERN_ALERT DRVSELFNAME ": %s", buf);
}

int myprintkn(const char * fmt, ...)
{
     char buf[768];
     va_list alist;

     /* We format message now */
     va_start(alist, fmt);
     vsnprintf(buf, sizeof(buf), fmt, alist);
     va_end(alist);

     /* All printk messages can have prefix like <X>,
      * where X is a char 0..7, d, c.
      * The KERN_ALERT macro correspond to the `<1>' prefix.
      * Also, you can see for linux/printk.h header.
      */
     return printk(KERN_ALERT DRVSELFNAME ": %s\n", buf);
}
/*****************************************************************************/
char* GCC_NOREGPARM drv_strcpy( char* dest, const char* src )
{
     char *tmp = dest;

     while ((*dest++ = *src++) != '\0')
          /* nothing */;
     return tmp;

}
/*****************************************************************************/
unsigned int GCC_NOREGPARM drv_strlen( const char* s )
{
     const char *sc;

     for (sc = s; *sc != '\0'; ++sc)
          /* nothing */;
     return sc - s;

}
/*****************************************************************************/
void GCC_NOREGPARM mydumphex(char * s, int size)
{
     int rc;
     char buf[100] = "";
     size_t len;
     size_t offs;

     /* output words */
     for (offs = 0, len = 0; offs < (size & (~1)); offs += 2)
     {
          rc = snprintf(buf + len, sizeof(buf) - len, "%02X%02X ", (unsigned char)s[offs], (unsigned char)s[offs+1]);
          if (rc > 0)
          {
               len += rc;
          }
          if (offs % 16 >= 14)
          {
               myprintk("%s\n", buf);
               len = 0;
          }
     }
     /* add trailing byte */
     for (; offs < size; ++offs)
     {
          rc = snprintf(buf + len, sizeof(buf) - len, "%02X ", (unsigned char)s[offs]);
          if (rc > 0)
          {
               len += rc;
          }
     }
     if (len)
     {
          myprintk("%s\n", buf);
          len = 0;
     }
}
/*****************************************************************************/
int mysprintf(char * buf, const char * fmt, ...)
{
     int rc;
     va_list args;

     va_start(args, fmt);
     rc = vsprintf(buf, fmt, args);
     va_end(args);

     return rc;
}

int mysnprintf(char * buf, unsigned int size, const char * fmt, ...)
{
     int rc;
     va_list args;

     va_start(args, fmt);
     rc = vsnprintf(buf, size, fmt, args);
     va_end(args);

     return rc;
}

void* mymemcpy ( void* dest, const void* src, size_t count )
{
     char *tmp = (char *) dest, *s = (char *) src;

     while (count--)
          *tmp++ = *s++;

     return dest;

}

/*
** For 2.6.x kernels series in various cases we can't lookup symbol by name.
** This insanity is handled by Makefile: it looks up 'set_packet_filter' and
** 'nf_reinject' symbols in /proc/kallsyms and then runs 'sed' to replace this
** pseudo-symbols with the actual return values before compiling.
** For older kernels series we ues mymodlookup() mechanism.
*/
#ifndef HAVE_MYMODLOOKUP
#define NF_METHOD_ENABLED #NF_METHOD_ENABLED#
#endif /* #ifndef HAVE_MYMODLOOKUP */
int myis_netfilter_method()
{
#ifdef HAVE_MYMODLOOKUP
     return mymodlookup(NULL, "nf_reinject") ? 1 : 0;
#else /* #ifdef HAVE_MYMODLOOKUP */
     return NF_METHOD_ENABLED;
#endif /* #ifndef HAVE_MYMODLOOKUP */
}

#ifndef HAVE_MYMODLOOKUP
#define PATCH_METHOD_ENABLED #PATCH_METHOD_ENABLED#
#endif /* #ifndef HAVE_MYMODLOOKUP */
int myis_patch_method()
{
#ifdef HAVE_MYMODLOOKUP
     return mymodlookup(NULL, "set_packet_filter") ? 1 : 0;
#else /* #ifdef HAVE_MYMODLOOKUP */
     return PATCH_METHOD_ENABLED;
#endif /* #ifndef HAVE_MYMODLOOKUP */
}

#if PATCH_METHOD_ENABLED == 1
static struct sk_buff* GCC_NOREGPARM default_in_pf(struct sk_buff *skb);
static struct sk_buff* GCC_NOREGPARM default_out_pf(struct sk_buff *skb);

static void* GCC_NOREGPARM (*fast_drv_in)(void *skb, int flags);
static void* GCC_NOREGPARM (*fast_drv_out)(void *skb, int flags);

static int IsLocked = 0;

int GCC_NOREGPARM myset_packet_filter(void * _in_pf, void * _out_pf)
{
#ifdef HAVE_MYMODLOOKUP
     FILTFUNC patch_filter;

     patch_filter = (FILTFUNC)mymodlookup(NULL, "set_packet_filter");
     if (!patch_filter)
          return -1;
#else /* #ifdef HAVE_MYMODLOOKUP */
     if( !myis_patch_method() )
          return -1;
#endif /* #ifndef HAVE_MYMODLOOKUP */

    	if (_in_pf && _out_pf)
     {
          if (!IsLocked)
          {
               MOD_INC_USE_COUNT_LALA;
               IsLocked = 1;
          }

		fast_drv_in = _in_pf;
		fast_drv_out = _out_pf;

#define	in_pf default_in_pf
#define	out_pf default_out_pf

#ifdef HAVE_MYMODLOOKUP
     patch_filter(in_pf, out_pf);
#else /* #ifdef HAVE_MYMODLOOKUP */
     #SET_PACKET_FILTER#
#endif /* #ifndef HAVE_MYMODLOOKUP */

#undef	in_pf
#undef	out_pf
     }

	else
	{
/* Disable packet filtering methods. It is responibility of *patch* to perfrom syncronization between set/clear filtering method and packets capturing machinery*/
#define	in_pf NULL
#define	out_pf NULL

#ifdef HAVE_MYMODLOOKUP
     patch_filter(in_pf, out_pf);
#else /* #ifdef HAVE_MYMODLOOKUP */
     #SET_PACKET_FILTER#
#endif /* #ifndef HAVE_MYMODLOOKUP */

#undef	in_pf
#undef	out_pf

          if (IsLocked)
          {
               MOD_DEC_USE_COUNT_LALA;
               IsLocked = 0;
          }
     }

     return 0;
}
#else
int GCC_NOREGPARM myset_packet_filter(void * _in_pf, void * _out_pf)
{
     myprintk("myset_packet_filter is not used for netfilter\n");
     return -1;
}
#endif //PATCH_METHOD_ENABLED

int GCC_NOREGPARM mycondfrag_filter(VN_SKB skb, VN_DEVICE dev, int is_recv)
{
#ifdef CONFIG_IPLIR
#  if (IPLIR_PATCH_VERSION >= 2)
  /* Check if filter was set by someone. */
  if (l2_condfrag_filter == NULL)
    return 0;

  return (*l2_condfrag_filter)((struct sk_buff *)skb, (LINUX_DEVICE *)dev, is_recv);
#  else
  /* Old patch - no condfrag. */
  return 0;
#  endif
#else
  /*
  ** No patch - no condfrag. This (and the same above) should avoid
  ** l2_condfrag_filter symbol dependency when kernel is not patched or
  ** old version of the patch is used.
  ** I wonder why don't we do the same with set_packet_filter symbol. --vel
  */
  return 0;
#endif
}

#if !defined(CONFIG_NETFILTER)
int GCC_NOREGPARM myset_net_filter(VN_NFQFUNC nfq_in, VN_NFQFUNC nfq_out)
{
  return -1;
}

#else

NF_DRV_FUNC iface_drv_nf_in = NULL;
NF_DRV_FUNC iface_drv_nf_out = NULL;
/*
** Tiny internal netfilter hooks which always queue packet. Actual
** processing is done by queue handler. --vel
*/
/* These functions are used to hook all protocols that we can hook. --vel */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
unsigned int my_netfilter_in(unsigned int hooknum,
                             struct sk_buff * skb,
                             const struct net_device * in,
                             const struct net_device * out,
                             int (*okfn)(struct sk_buff *))
#else
unsigned int my_netfilter_in(unsigned int hooknum, struct sk_buff ** skb,
                             const struct net_device * in,
                             const struct net_device * out,
                             int (*okfn)(struct sk_buff *))
#endif
{
  udebug2(( "drviplir (nf_in): dev '%s'", mydev_fetch_name(in) ));
  return NF_QUEUE;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
unsigned int my_netfilter_out(unsigned int hooknum,
                              struct sk_buff * skb,
                              const struct net_device * in,
                              const struct net_device * out,
                              int (*okfn)(struct sk_buff *))
#else
unsigned int my_netfilter_out(unsigned int hooknum, struct sk_buff ** skb,
                              const struct net_device * in,
                              const struct net_device * out,
                              int (*okfn)(struct sk_buff *))
#endif
{
  udebug2(( "drviplir (nf_out): dev '%s'", mydev_fetch_name(out) ));
  return NF_QUEUE;
}

/*
** Queue handler for all protocols we can handle. There is some stuff we
** better do here, before passing packet to the core ... --vel
*/
/*
** This is going to become quite insane at some moment, but it can't be done
** otherwise until we drop support for older kernels. --vel
*/
#ifdef NF_USES_QUEUE_ENTRY
int my_netfilter_queue(struct nf_queue_entry * nfq, unsigned int queuenum)
#else
int my_netfilter_queue(struct sk_buff * skb, struct nf_info * nfq,
                       unsigned int queuenum,
                       void * data)
#endif
{
#ifdef NF_USES_QUEUE_ENTRY
  struct sk_buff * skb = nfq->skb;
#endif

  struct iphdr *iph;
  struct udphdr *udph;
  struct dst_entry *dst;
  struct neighbour *nghb;
  NOREF(iph);
  NOREF(udph);
  NOREF(dst);
  NOREF(nghb);

  udebug2(( "drviplir(nf_queue): skb %x nfinfo/nfq %x", skb, nfq ));
  /*
  ** Now with non-IP protocols, we must look at the protocols also,
  ** not only the hook number. For each protocol we do almost the same
  ** thing: we see if we have function of the driver core that we can
  ** call to pass packet to the core, and if we have one, we pass packet
  ** to it and look at return value; if the return value is non-zero,
  ** it means that the core did not handle the packet, and we immediately
  ** reinject it back to the stack. If the core returns 0, it means that
  ** the core has stolen the packet and will reinject in later; we do
  ** nothing. Finally, if there is no function to call, we drop the packet
  ** (it if done with nf_reniject() with NF_DROP, because it is more clean
  ** way that allows us not to mess with locks and refcounting - kernel
  ** will unlock and free everything for us. --vel
  */
  if ((nfq->pf == PF_INET) && (nfq->hook == NF_INET_PRE_ROUTING))
  {
#if !(defined(CONFIG_KALLSYMS) && KALLSYMS_LOOKUP_NAME_EXPORT)
  if(myip_rcv_finish == 0)
  {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
       iph = skb->nh.iph;
       udph = skb->h.uh;
#else
       iph = ip_hdr(skb);
       udph = udp_hdr(skb);
#endif
       if( iph->protocol == IPPROTO_UDP &&
           udph->source == htons(IPLIRCFG_PORT) &&
           udph->dest == htons(IPLIRCFG_PORT) )
       {
            myip_rcv_finish = nfq->okfn;
       }
  }
#endif
    /* Don't remove these, I bet we'll need them many times ... --vel */
    udebug2(("input queue ip"));
    if (iface_drv_nf_in)
    {
#ifdef  DRV_TRACE_RUNTIME       
       if ( trace_mask & TRACE_BIT_INTO_DRV )
       {
            MYDUMP_SKB_IPHDR( "nf_in", skb);
       }
       if ( trace_mask & TRACE_BIT_IN_SKB )
       {
            mydump_skb(skb);
       } 
#endif
      if (iface_drv_nf_in(skb, nfq))
      {
        NF_REINJECT_COMPAT(skb, nfq, NF_ACCEPT);
      }
      return 0;
    }
    else
    {
      NF_REINJECT_COMPAT(skb, nfq, NF_DROP);
      return 0;
    }
  }

  if ((nfq->pf == PF_INET) && (nfq->hook == NF_INET_POST_ROUTING))
  {
#if !(defined(CONFIG_KALLSYMS) && KALLSYMS_LOOKUP_NAME_EXPORT)
    if(myip_finish_output == 0)
    {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
         iph = skb->nh.iph;
         udph = skb->h.uh;
#else
         iph = ip_hdr(skb);
         udph = udp_hdr(skb);
#endif
         dst = kernel_skb_dst(skb);
         if( iph->protocol == IPPROTO_UDP &&
             udph->source == htons(IPLIRCFG_PORT) &&
             udph->dest == htons(IPLIRCFG_PORT) &&
             dst &&
             (nghb = DST_NEIGHBOUR(dst)) &&
             (nghb->type != RTN_BROADCAST) )
         {
              myip_finish_output = nfq->okfn;
         }
         if ( nghb )
              myneighbour_release(nghb);
    }
#endif
    /* Don't remove these, I bet we'll need them many times ... --vel */
    udebug2(("output queue ip"));
    if (iface_drv_nf_out)
    {
#ifdef DRV_TRACE_RUNTIME
      if ( trace_mask & TRACE_BIT_INTO_DRV )
      {
           MYDUMP_SKB_IPHDR( "nf_out", skb);
      }
      if ( trace_mask & TRACE_BIT_IN_SKB )
      {
           mydump_skb(skb);
      }      
#endif
      if (iface_drv_nf_out(skb, nfq))
      {
        NF_REINJECT_COMPAT(skb, nfq, NF_ACCEPT);
      }
      return 0;
    }
    else
    {
      NF_REINJECT_COMPAT(skb, nfq, NF_DROP);
      return 0;
    }
  }
#ifdef NF_FILTER_NONIP
#ifdef NF_FILTER_ARP
  if ((nfq->pf == NFPROTO_ARP) && (nfq->hook == NF_ARP_IN))
  {
//    printk("caught incoming ARP\n");
    if (iface_drv_nf_in)
    {
      if (iface_drv_nf_in(skb, nfq))
      {
        NF_REINJECT_COMPAT(skb, nfq, NF_ACCEPT);
      }
      return 0;
    }
    else
    {
      NF_REINJECT_COMPAT(skb, nfq, NF_DROP);
      return 0;
    }
  }

  if ((nfq->pf == NFPROTO_ARP) && (nfq->hook == NF_ARP_OUT))
  {
//    printk("caught outgoing ARP\n");
    if (iface_drv_nf_out)
    {
      if (iface_drv_nf_out(skb, nfq))
      {
        NF_REINJECT_COMPAT(skb, nfq, NF_ACCEPT);
      }
      return 0;
    }
    else
    {
      NF_REINJECT_COMPAT(skb, nfq, NF_DROP);
      return 0;
    }
  }
#endif // NF_FILTER_ARP

#ifdef CONFIG_IPV6
  if ((nfq->pf == PF_INET6) && (nfq->hook == NF_INET_PRE_ROUTING))
  {
    /* Don't remove these, I bet we'll need them many times ... --vel */
    udebug2(("input queue ip6"));
    if (iface_drv_nf_in)
    {
      if (iface_drv_nf_in(skb, nfq))
      {
        NF_REINJECT_COMPAT(skb, nfq, NF_ACCEPT);
      }
      return 0;
    }
    else
    {
      NF_REINJECT_COMPAT(skb, nfq, NF_DROP);
      return 0;
    }
  }

  if ((nfq->pf == PF_INET6) && (nfq->hook == NF_INET_POST_ROUTING))
  {
    /* Don't remove these, I bet we'll need them many times ... --vel */
    udebug2(("output queue ip6"));
    if (iface_drv_nf_out)
    {
      if (iface_drv_nf_out(skb, nfq))
      {
        NF_REINJECT_COMPAT(skb, nfq, NF_ACCEPT);
      }
      return 0;
    }
    else
    {
      NF_REINJECT_COMPAT(skb, nfq, NF_DROP);
      return 0;
    }
  }
#endif // CONFIG_IPV6
#endif // NF_FILTER_NONIP
  /*
  ** This should never happen, because in the queue handler we see only
  ** protocols that we hooked, and we handle them all above. But... --vel
  */
  myprintk("unknown hook, drop (please report to Infotecs)\n");
  NF_REINJECT_COMPAT(skb, nfq, NF_DROP);
  return 0;
}

/*
** Those dummy functions are needed to get ARP packet queued in the
** netfilter. Since 2.6.25, kernel has so called "address family info"
** (afinfo) structure for each protocols, which holds data and methods which
** are used to route packets of that protocol and calculate checksums
** for those packets. For some strange reason, netfilter does not allow
** queueing of packets belonging to protocols which do not have address family
** info entry. ARP protocol doesn't have it (actually, of all protocols, only
** ipv4 and ipv6 have it), so ARP packets cannot be queued by default.
** They can be hooked (using NFPROTO_ARP and NF_ARP_IN/NF_ARP_OUT), but
** when the hook handler returns NF_QUEUE, the packet simply gets dropped.
** To cirvumvent this, we register fake afinfo structure for the ARP protocols,
** where all data is zero and all functions do nothing. This should not
** break anything because ARP packets are not checksummed or routed anyway,
** but after this registration we can queue ARP packets and they arrive
** successfully to our queue handler, so we can handle them in usual way
** and use nf_reinject() to return them to the stack. --vel
*/
#ifdef NF_FILTER_NONIP
#ifdef NF_FILTER_ARP
__sum16 nf_afinfo_dummy_checksum(struct sk_buff * skb, unsigned int hook,
                                 unsigned int dataoff, u_int8_t protocol)
{
  return 0;
}

__sum16 nf_afinfo_dummy_checksum_partial(struct sk_buff * skb,
                                         unsigned int hook,
                                         unsigned int dataoff,
                                         unsigned int len,
                                         u_int8_t protocol)
{
  return 0;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,39)
int nf_afinfo_dummy_route(struct net *net, struct dst_entry **dst, struct flowi * fl, bool strict)
#else
int nf_afinfo_dummy_route(struct dst_entry **dst, struct flowi * fl)
#endif
{
  return 0;
}

void nf_afinfo_dummy_saveroute(const struct sk_buff * skb,
                               struct nf_queue_entry * entry)
{
}

int nf_afinfo_dummy_reroute(struct sk_buff * skb,
                            const struct nf_queue_entry * entry)
{
  return 0;
}
#endif // NF_FILTER_NONIP
#endif // NF_FILTER_ARP
/*
** These structures describe hooks. Now we fill them here, rather
** than assigning values in myset_net_filter() - it is faster and
** cleaner. --vel
*/
struct nf_hook_ops nf_ops_inet_in =
{
  .hook = my_netfilter_in,
  .pf = PF_INET,
  .hooknum = NF_INET_PRE_ROUTING,
  .priority = NF_IP_PRI_FIRST,
  .owner = THIS_MODULE
};

struct nf_hook_ops nf_ops_inet_out =
{
  .hook = my_netfilter_out,
  .pf = PF_INET,
  .hooknum = NF_INET_POST_ROUTING,
  .priority = NF_IP_PRI_FIRST,
  .owner = THIS_MODULE,
};

#ifdef NF_FILTER_NONIP
#ifdef NF_FILTER_ARP
/* Non-IP hooks are activated only on modern kernels. --vel */
struct nf_hook_ops nf_ops_arp_in =
{
  .hook = my_netfilter_in,
  .pf = NFPROTO_ARP,
  .hooknum = NF_ARP_IN,
  .priority = INT_MIN,
  .owner = THIS_MODULE
};

struct nf_hook_ops nf_ops_arp_out =
{
  .hook = my_netfilter_out,
  .pf = NFPROTO_ARP,
  .hooknum = NF_ARP_OUT,
  .priority = INT_MIN,
  .owner = THIS_MODULE
};
#endif //NF_FILTER_ARP

#ifdef CONFIG_IPV6
struct nf_hook_ops nf_ops_inet6_in =
{
  .hook = my_netfilter_in,
  .pf = PF_INET6,
  .hooknum = NF_INET_PRE_ROUTING,
  .priority = NF_IP6_PRI_FIRST,
  .owner = THIS_MODULE
};

struct nf_hook_ops nf_ops_inet6_out =
{
  .hook = my_netfilter_out,
  .pf = PF_INET6,
  .hooknum = NF_INET_POST_ROUTING,
  .priority = NF_IP6_PRI_FIRST,
  .owner = THIS_MODULE
};
#endif // CONFIG_IPV6
#endif // NF_FILTER_NONIP

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,14)
static struct nf_queue_handler nfq_ops =
{
     .outfn = my_netfilter_queue,
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,8,0)
     .name = "IPLIR_QUEUE",
#endif
};
#endif

#ifdef NF_FILTER_NONIP
#ifdef NF_FILTER_ARP
/* Fake address family info structure for ARP protocol, see above. --vel */
static struct nf_afinfo nf_arp_afinfo =
{
     .family = NFPROTO_ARP,
     .checksum = nf_afinfo_dummy_checksum,
     .checksum_partial = nf_afinfo_dummy_checksum_partial,
     .route = nf_afinfo_dummy_route,
     .saveroute = nf_afinfo_dummy_saveroute,
     .reroute = nf_afinfo_dummy_reroute,
     .route_key_size = 0
};
#endif // NF_FILTER_NONIP
#endif // NF_FILTER_ARP

/*
** This structure and array below are used to hold information about
** protocols for which netfilter hooks are already registered by us.
** It is possible that we registered half of the hooks, but then some hook
** or queue handler refuses to register. In this case we must unregister
** everything that we registered before, and the driver must refuse to load,
** because it can't work correctly with only half of the hooks.
** For this, we have reg_item for each protocol, which stores information
** about whether we registered all hooks for that protocol, and which
** function to call to unregister all hooks for that protocol. It is filled
** after successful registration of all hooks for that protocol.
** So, if some protocol fails, we can easily see which protocols were already
** registered, and call appropriate function to unregister them. --vel
*/
struct reg_item
{
  char reg;
  void (*unreg_func)(void);
};

/* PF_MAX is the highest possible protocol number, starting from 0. */
static struct reg_item protocol_reg_items[PF_MAX + 1];

/* This function unregisters all hooks and queue handler for IP protocol. */
void mynet_filter_unregister_ip(void)
{
  if (protocol_reg_items[PF_INET].reg == 0)
    return;
#if  LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23)
  nf_unregister_queue_handler(PF_INET);
#elif LINUX_VERSION_CODE < KERNEL_VERSION(3,8,0)
  nf_unregister_queue_handler(PF_INET, &nfq_ops);
#else
  nf_unregister_queue_handler();
#endif
  nf_unregister_hook(&nf_ops_inet_out);
  nf_unregister_hook(&nf_ops_inet_in);
  myprintk("Netfilter hooks for IP protocol unregistered.\n");
}

/* This function register hooks and queue handler for IP protocol. */
int mynet_filter_register_ip(void)
{
	int ret=0;

  /* Hook for incoming packets. */
  if (nf_register_hook(&nf_ops_inet_in) != 0)
  {
    myprintk("Failed to register IN hook for IP protocol.\n");
    return -1;
  }
  udebug(( "drviplir: register NF INET hook for incoming packets" ));

  /* Hook for ourgoing packets. */
  if (nf_register_hook(&nf_ops_inet_out) != 0)
  {
    myprintk("Failed to register OUT hook for IP protocol.\n");
    nf_unregister_hook(&nf_ops_inet_in);
    return -1;
  }
  udebug(( "drviplir: register NF INET hook for outcoming packets" ));

  /* Queue handler - different format for different kernels. */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,14) && LINUX_VERSION_CODE < KERNEL_VERSION(3,8,0)
  ret = nf_register_queue_handler(PF_INET, &nfq_ops);
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(3,8,0)
  nf_register_queue_handler(&nfq_ops);
#else
  ret = nf_register_queue_handler(PF_INET, my_netfilter_queue, NULL);
#endif
  if (ret != 0)
  {
    myprintk("Failed to register queue handler for IP protocol.\n");
    /*
    ** We don't try to deregister alien queue handler anymore, as it does not
    ** work with modern kernels, and it is stupid idea anyway. The Right Thing
    ** is to tell user to disable all other packet filters. --vel
    */
    if (ret == -EBUSY)
    {
      myprintk("Queue entry is occupied by somebody else. Probably you have\n");
      myprintk("another packet filter, please disable it before using ViPNet.\n");
    }
    nf_unregister_hook(&nf_ops_inet_in);
    nf_unregister_hook(&nf_ops_inet_out);
    return -1;
  }

  /* If we reached this, then everything is registered properly for IP. */
  protocol_reg_items[PF_INET].reg = 1;
  protocol_reg_items[PF_INET].unreg_func = mynet_filter_unregister_ip;
  myprintk("Netfilter hooks for IP protocol registered.\n");
  return 0;
}

#ifdef NF_FILTER_NONIP
#ifdef NF_FILTER_ARP
/* This function unregisters all hooks and queue handler for ARP protocol. */
void mynet_filter_unregister_arp(void)
{
  if (protocol_reg_items[NFPROTO_ARP].reg == 0)
  {
    return;
  }
#if  LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23)
  nf_unregister_queue_handler(NFPROTO_ARP);
#elif LINUX_VERSION_CODE < KERNEL_VERSION(3,8,0)
  nf_unregister_queue_handler(NFPROTO_ARP, &nfq_ops);
#else
  nf_unregister_queue_handler();
#endif
  
  /* ARP is special - don't forget to unregister afinfo. --vel */
  mynf_unregister_afinfo(&nf_arp_afinfo);
  nf_unregister_hook(&nf_ops_arp_out);
  nf_unregister_hook(&nf_ops_arp_in);
  myprintk("Netfilter hooks for ARP protocol unregistered.\n");
}

/* This function registers hooks and queue handler for ARP protocol. */
int mynet_filter_register_arp(void)
{
  int ret;

  if (nf_register_hook(&nf_ops_arp_in) != 0)
  {
    return -1;
  }

  if (nf_register_hook(&nf_ops_arp_out) != 0)
  {
    nf_unregister_hook(&nf_ops_arp_in);
    return -1;
  }

  if (mynf_register_afinfo(&nf_arp_afinfo) != 0)
  {
    nf_unregister_hook(&nf_ops_arp_out);
    nf_unregister_hook(&nf_ops_arp_in);
    return -1;
  }
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,8,0)
  ret = nf_register_queue_handler(NFPROTO_ARP, &nfq_ops);
#else
  nf_register_queue_handler( &nfq_ops);
  ret=0;
#endif
  if (ret != 0)
  {
    myprintk("Failed to register queue handler for ARP protocol.\n");
    if (ret == -EBUSY)
    {
       myprintk("Queue entry is occupied by somebody else. Probably you have\n");
       myprintk("another packet filter, please disable it before using ViPNet.\n");
    }
    mynf_unregister_afinfo(&nf_arp_afinfo);
    nf_unregister_hook(&nf_ops_arp_out);
    nf_unregister_hook(&nf_ops_arp_in);
    return -1;
  }

  /* Everything is registered for ARP. */
  protocol_reg_items[NFPROTO_ARP].reg = 1;
  protocol_reg_items[NFPROTO_ARP].unreg_func = mynet_filter_unregister_arp;
  myprintk("Netfilter hooks for ARP protocol registered.\n");
  return 0;
}
#endif // NF_FILTER_ARP

#ifdef CONFIG_IPV6
/* This function unregisters all hooks and queue handler for IPv6 protocol. */
void mynet_filter_unregister_ipv6(void)
{
  if (protocol_reg_items[PF_INET6].reg == 0)
    return;

#if  LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23)
  nf_unregister_queue_handler(PF_INET6);
#elif LINUX_VERSION_CODE < KERNEL_VERSION(3,8,0)
  nf_unregister_queue_handler(PF_INET6, &nfq_ops);
#else
  nf_unregister_queue_handler();
#endif

  nf_unregister_hook(&nf_ops_inet6_out);
  nf_unregister_hook(&nf_ops_inet6_in);
  myprintk("Netfilter hooks for IPv6 protocol unregistered.\n");
}

/* This function registers hooks and queue handler for IPv6 protocol. */
int mynet_filter_register_ipv6(void)
{
  int ret;

  if (nf_register_hook(&nf_ops_inet6_in) != 0)
  {
    return -1;
  }
  udebug(( "drviplir: register NF INET6 hook for incoming packets" ));

  if (nf_register_hook(&nf_ops_inet6_out) != 0)
  {
    nf_unregister_hook(&nf_ops_inet6_in);
    return -1;
  }
  udebug(( "drviplir: register NF INET6 hook for outcoming packets" ));

#if LINUX_VERSION_CODE < KERNEL_VERSION(3,8,0)
  ret = nf_register_queue_handler(PF_INET6, &nfq_ops);
#else
  nf_register_queue_handler( &nfq_ops);
  ret=0;
#endif

  if (ret != 0)
  {
    myprintk("Failed to register queue handler for IPv6 protocol.\n");
    /*
    ** We don't try to deregister alien queue handler anymore, as it does not
    ** work with modern kernels, and it is stupid idea anyway. The Right Thing
    ** is to tell user to disable all other packet filters. --vel
    */
    if (ret == -EBUSY)
    {
      myprintk("Queue entry is occupied by somebody else. Probably you have\n");
      myprintk("another packet filter, please disable it before using ViPNet.\n");
    }
    nf_unregister_hook(&nf_ops_inet6_in);
    nf_unregister_hook(&nf_ops_inet6_out);
    return -1;
  }

  /* If we reached this, then everything is registered properly for IP. */
  protocol_reg_items[PF_INET6].reg = 1;
  protocol_reg_items[PF_INET6].unreg_func = mynet_filter_unregister_ipv6;
  myprintk("Netfilter hooks for IPv6 protocol registered.\n");
  return 0;
}
#endif // CONFIG_IPV6
#endif //  NF_FILTER_NONIP

/*
** This function unregisters hooks and everything for all protocols that
** have been registered before. It goes though reg_item array and calls unregister
** function for every protocol that has the flag that it has been registered. --vel
*/
void mynet_filter_unregister_registered(void)
{
  int i;

  for (i = 0; i <= PF_MAX; i++)
    if (protocol_reg_items[i].reg)
    {
      protocol_reg_items[i].unreg_func();
      protocol_reg_items[i].reg = 0;
    }
}

// drviplir build phase:
// install-time checks indicated kernel is xfrm6-patched & ip6 sources are in distrib --
// 		going to build itcsipcls.ko
#if defined IPLIR_DISABLE_IPv6
	#ifndef KERN_IPCLS // ensure again (compile-time) it's supported kernel version
		#undef IPLIR_DISABLE_IPv6
	#endif
#endif

/* This is list of functions that register hooks and everything for each protocol. */
typedef int (*REGFUNC)(void);
REGFUNC protocol_autoreg[] =
{
  mynet_filter_register_ip,
#ifdef NF_FILTER_NONIP
#ifdef NF_FILTER_ARP
  mynet_filter_register_arp,
#endif
#ifdef CONFIG_IPV6
#ifndef IPLIR_DISABLE_IPv6
  mynet_filter_register_ipv6,
#endif
#endif
#endif
  NULL
};

/*
** This function actually sets or unsets packet filter for netfilter engine.
** Now it is quite simple. --vel
*/
int GCC_NOREGPARM myset_net_filter(VN_NFQFUNC nfq_in, VN_NFQFUNC nfq_out)
{
  int i, ret = 0;

  /*
  ** We register hooks only if both function for incoming and outgoing
  ** packets are provided by the driver core. It will not work otherwise
  ** anyway. --vel
  */
  if (nfq_in && nfq_out)
  {
    /* Save function given by the driver core. */
    iface_drv_nf_in = nfq_in;
    iface_drv_nf_out = nfq_out;

    /*
    ** Fill reg_item array with zeroes, to indicate that all protocols are
    ** unregistered. --vel
    */
    memset(protocol_reg_items, 0, sizeof(protocol_reg_items));

    /*
    ** Walk through protocol_autoreg array, and call each function that is there.
    ** each of those functions registers hooks and everything for one protocol
    ** that we support. --vel
    */
    for (i = 0; protocol_autoreg[i]; i++)
    {
      ret = protocol_autoreg[i]();
      /* If some protocol fails, stop. */
      if (ret < 0)
        break;
    }
    /*
    ** When some protocol fails, the register function clean up all hooks
    ** for that protocol (see above), but doesn't touch other protocols. So,
    ** we can be sure that there are no remaining hooks for protocol that
    ** failed, and that all other protocols are still registered. So we
    ** unregister all other protocols here. --vel
    */
    if (ret < 0)
    {
      mynet_filter_unregister_registered();
      iface_drv_nf_in = NULL;
      iface_drv_nf_out = NULL;
    }

    return ret;
  }
  else
  {
    /* If driver core provided null functions, we just unregister everything. */
    mynet_filter_unregister_registered();
    iface_drv_nf_in = NULL;
    iface_drv_nf_out = NULL;
    return 0;
  }
}
#endif

/*****************************************************************************/
int GCC_NOREGPARM mynetif_rx(VN_SKB skb)
{
  /* Never use netif_rx() here! It
  result in reduction of speed in packet processing.
  If we call netif_rx() instead of netif_receive_skb()
  packet is put in incoming kernel queue and when we use own
  packet queue in kernel thread context there is
  a large delay between this moment and the moment
  when packet come back to us with RECV_FLAG.
  This problem is only for patch packet hooking
  with kernel thread queue support. */

  /*
   *   bottom halves (softirq), ..
   *  netif_receive_skb()    softirq.
   *     ,     
   *     bh_lock_sock(sk), 
   *  struct sock *sk,   
   *   softirq    bh_lock_sock() 
   *     sk.
   */
  
#ifdef  DRV_TRACE_RUNTIME
  if ( trace_mask & TRACE_BIT_OUTOF_DRV )
  {
       MYDUMP_SKB_IPHDR( "netif_rx", skb);
  }
  if ( trace_mask & TRACE_BIT_OUT_SKB )
  {
       mydump_skb(skb);
  }  
#endif

  int res;
  local_bh_disable();
  res = netif_receive_skb((struct sk_buff *)skb);
  local_bh_enable();
  return res;
}

int GCC_NOREGPARM mynetif_rx_slow(VN_SKB skb)
{
     int res;
     res = netif_rx((struct sk_buff *)skb);
     return res;
}

int GCC_NOREGPARM mynetdev_refcnt( VN_DEVICE dev )
{
     LINUX_DEVICE * netdev = (LINUX_DEVICE *)dev;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,37)
     return netdev_refcnt_read( netdev );
#else
     return atomic_read(&netdev->refcnt);
#endif
}

void GCC_NOREGPARM mydev_drop_allrefs( VN_DEVICE dev )
{
     LINUX_DEVICE * netdev = (LINUX_DEVICE *)dev;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,37)
     /// we need to set reference counter for each cpu to zero
     int i; /// temporary cpu counter
     for_each_possible_cpu(i)
     {
          *per_cpu_ptr( netdev->pcpu_refcnt, i ) = 0;
     }
#else
     netdev->refcnt.counter = 0;
#endif
}

VN_INIFADDR GCC_NOREGPARM mydev_get_ifaddr( VN_DEVICE dev )
{
     struct in_ifaddr* addr = NULL;
     struct in_device* inetDev = myin_dev_get( dev );
     if( inetDev )
     {
          addr = inetDev->ifa_list;
          in_dev_put( inetDev );
     }

     return addr;
}

VN_INIFADDR GCC_NOREGPARM mydev_get_next_ifaddr( VN_INIFADDR prev )
{
     struct in_ifaddr* addr = (struct in_ifaddr*) prev;
     return ( addr ? addr->ifa_next : NULL );
}

void mymark_flow_cache_outdated(void)
{
#ifdef CONFIG_XFRM
// @todo Research this problem: flow_cache_genid isn't declared in 3.15.x and later versions.
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,15,0)
	myatomic_inc(&flow_cache_genid);
#endif
#endif
}

void GCC_NOREGPARM myinet_putpeer(VN_INET_PEER peer)
{
#if defined (KERN_IPCLS)
	struct inet_peer *p = (struct inet_peer *)peer;

	p->dtime = (__u32)jiffies;
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,18,0)
	smp_mb__before_atomic_dec();
#else
	smp_mb__before_atomic();
#endif // LINUX_VERSION_CODE < KERNEL_VERSION(3,18,0)
	atomic_dec(&p->refcnt);
#endif
}

int GCC_NOREGPARM my__ip6_local_out(VN_SKB vnskb)
{
	return my_ip6_local_out_krnface((struct sk_buff *)vnskb);
}

/*****************************************************************************/

static int tcp_local_port_engaged( unsigned short snum  )
{
     struct inet_hashinfo *hashinfo = &tcp_hashinfo;
     struct net *net = &init_net;
     struct inet_bind_hashbucket *head;
     struct inet_bind_bucket *tb;
     
     head = &hashinfo->bhash[ inet_bhashfn( net, snum, hashinfo->bhash_size ) ];

     spin_lock(&head->lock);

#if LINUX_VERSION_CODE < KERNEL_VERSION(3,9,11)
     struct hlist_node *node;
     inet_bind_bucket_for_each( tb, node, &head->chain )
#else     
     inet_bind_bucket_for_each( tb, &head->chain )
#endif
     {
          if ( net_eq( ib_net(tb), net) && tb->port == snum )
          {
               spin_unlock(&head->lock);
               return 1;
          }
     }
     spin_unlock(&head->lock);
     return 0;
}

static int udp_local_port_inuse( __u16 num, __be32 rcv_saddr, 
                                 const struct udp_hslot *hslot )
{
     struct sock *sk2;
     struct hlist_nulls_node *node;
     
     sk_nulls_for_each(sk2, node, &hslot->head)
     {
          // this will work only with init network namespace
          // TODO do for all namespaces if required
          //
          if ( net_eq( sock_net(sk2), &init_net )
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,33)            
              && ( udp_sk(sk2)->udp_port_hash == num ) 
              && ( inet_sk(sk2)->inet_rcv_saddr == rcv_saddr )
#else              
              && ( sk2->sk_hash == num )
#endif              
             )
          {
                return 1;
          }
     }
     return 0;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,33)            
static unsigned int udp4_portaddr_hash(struct net *net, __be32 saddr,
                       unsigned int port)
{
    return jhash_1word((__force u32)saddr, net_hash_mix(net)) ^ port;
}

static int udp_local_port_engaged( __u16 num, __be32 rcv_saddr )
{
     int res = 0;
     struct udp_hslot *hslot, *hslot2;
     struct udp_table *udptable = &udp_table;
     struct net *net = &init_net;
     
     hslot = udp_hashslot( udptable, net, num );

     unsigned int hash2_nulladdr = udp4_portaddr_hash( net, htonl(INADDR_ANY), num );
     unsigned int hash2_partial  = udp4_portaddr_hash( net, rcv_saddr, 0);
     
     spin_lock_bh(&hslot->lock);
     
     if ( hslot->count <= 10 ) 
     {
          res = udp_local_port_inuse ( num, rcv_saddr, hslot );
          spin_unlock_bh(&hslot->lock);
          return res;
     }
     
     unsigned int slot2 = hash2_partial ^ num;

     slot2          &= udptable->mask;
     hash2_nulladdr &= udptable->mask;

     hslot2 = udp_hashslot2( udptable, slot2 );
     
     if ( hslot->count < hslot2->count )
     {
          res = udp_local_port_inuse ( num, rcv_saddr, hslot );
          spin_unlock_bh(&hslot->lock);
          return res;
     }
     
     res = udp_local_port_inuse ( num, rcv_saddr, hslot2 );
     if ( res )
     {
          spin_unlock_bh(&hslot->lock);
          return res;
     }
     
     if ( hash2_nulladdr != slot2 )
     {
           hslot2 = udp_hashslot2( udptable, hash2_nulladdr );
           res = udp_local_port_inuse( num, rcv_saddr, hslot2 );
     }
     spin_unlock_bh(&hslot->lock);
     
     return res;
}
#else
int udp_local_port_engaged( __u16 num, __be32 rcv_saddr )
{
     int res = 0;
     struct udp_hslot *hslot;
     struct udp_table *udptable = &udp_table;
     struct net *net = &init_net;

     hslot = &udptable->hash[udp_hashfn(net, num)];
     spin_lock_bh(&hslot->lock);
     res = udp_local_port_inuse( num, rcv_saddr, hslot );
     spin_unlock_bh(&hslot->lock);
     
     return res;
}
#endif

int GCC_NOREGPARM my_is_ip4_local_port_inuse( unsigned char proto, unsigned short localPort, unsigned int localIp )
{
     if ( proto == IPPROTO_UDP )
     {
          return udp_local_port_engaged( (unsigned short)localPort, htonl(localIp) );
     }
     if ( proto == IPPROTO_TCP )
     {
          return tcp_local_port_engaged( (unsigned short)localPort );
     }     
     printk( KERN_ERR "Wrong proto = %d in my_is_ip4_local_port_inuse.\n", proto );
     return 0;
}

/*****************************************************************************/
int GCC_NOREGPARM mydev_queue_xmit(VN_SKB skbuf)
{
  struct sk_buff *skb = (struct sk_buff *)skbuf;

  struct ethhdr *ethh;
  struct dst_entry *dst;
  struct net_device *dev;
  struct rtable *rt;
  struct neighbour *nghb;
  u32 nexthop;
  int res;

  NOREF(dev);
  NOREF(rt);
  NOREF(nghb);
  NOREF(nexthop);
  NOREF(res);

#ifdef  DRV_TRACE_RUNTIME  
  if ( trace_mask & TRACE_BIT_OUTOF_DRV )
  {
       MYDUMP_SKB_IPHDR( "dev_queue_xmit", skb);
  }
  if ( trace_mask & TRACE_BIT_OUT_SKB )
  {
       mydump_skb(skb);
  }  
#endif

  if( ( mydev_fetch_header_len( skb->dev ) != 0 ) && skb->protocol == htons(ETH_P_IP))
  {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
    skb->mac.raw = skb->data;
#else
    skb_reset_mac_header(skb);
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
    skb_shinfo(skb)->gso_size = 0;
#endif
    ethh = eth_hdr(skb);
    // MAC ,      arp ,     MAC.
    if( ethh->h_dest[0] == 0 &&
        ethh->h_dest[1] == 0 &&
        ethh->h_dest[2] == 0 &&
        ethh->h_dest[3] == 0 &&
        ethh->h_dest[4] == 0 &&
        ethh->h_dest[5] == 0 )
    {
      myfill_skb(skb);
      dst = kernel_skb_dst(skb);
      if (dst)
      {
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,1,0)
        if ( dst->hh)
        {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
          int hh_alen;
          read_lock_bh(&dst->hh->hh_lock);
          hh_alen = HH_DATA_ALIGN(dst->hh->hh_len);
          mymemcpy(skb->data, dst->hh->hh_data, hh_alen);
          read_unlock_bh(&dst->hh->hh_lock);
          return dst->hh->hh_output(skb);
#else
          unsigned seq;
          do {
            int hh_alen;

            seq = read_seqbegin(&dst->hh->hh_lock);
            hh_alen = HH_DATA_ALIGN(dst->hh->hh_len);
            mymemcpy(skb->data, dst->hh->hh_data, hh_alen);
          } while (read_seqretry(&dst->hh->hh_lock, seq));
#endif // LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
        }
        else
#endif // LINUX_VERSION_CODE < KERNEL_VERSION(3,1,0)
        {

          myrcu_read_lock();

          nghb = DST_NEIGHBOUR(dst);
          if (nghb)
          {
            res = DST_NEIGHBOUR_OUTPUT(nghb);
            myrcu_read_unlock();
            myneighbour_release(nghb);
            return res;
          }
          myrcu_read_unlock();
        }
      }
    }
  }
  return dev_queue_xmit(skb);
}
/*****************************************************************************/
unsigned int GCC_NOREGPARM myeth_type_trans(VN_SKB skb, VN_DEVICE dev)
{
  return eth_type_trans((struct sk_buff *)skb, (LINUX_DEVICE *)dev);
}
/*****************************************************************************/
VN_DEVICE GCC_NOREGPARM mydev_get(const char * devname)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
  VN_DEVICE dev = dev_get_by_name(&init_net, devname);
#else
  VN_DEVICE dev = dev_get_by_name(devname);
#endif
  if( dev )
    udebug3(( "dev_get_by_name, name '%s',cnt = %d\n", devname,
              mynetdev_refcnt( dev ) ));
  else
    udebug3(( "dev_get_by_name, name '%s', NULL device\n", devname ));

  return dev;
}
/*****************************************************************************/
void GCC_NOREGPARM mydev_put(VN_DEVICE dev)
{
  dev_put(( LINUX_DEVICE * )dev);
  udebug3(( "dev_put, cnt = %d\n", mynetdev_refcnt( dev ) ));
}
/*****************************************************************************/
void GCC_NOREGPARM mydev_hold(VN_DEVICE dev)
{
  dev_hold(( LINUX_DEVICE * )dev);
  udebug3(( "dev_hold, cnt = %d\n", mynetdev_refcnt( dev ) ));
}
/*****************************************************************************/
int GCC_NOREGPARM myregister_chrdev(unsigned int major, const char * name, void * fops)
{
  udebug(( "itcs_drv: register chrdev major %d, name '%s', fops %x", major, name, fops ));
  if( name == NULL || fops == NULL )
    return -ENODEV;

  return register_chrdev(major, name, fops);
}
/*****************************************************************************/
int GCC_NOREGPARM myunregister_chrdev(unsigned int major, const char * name)
{
  udebug(( "itcs_drv: unregister chrdev major %d, name '%s'", major, name ));
  unregister_chrdev(major, name);
  return 0;
}
/*****************************************************************************/
VN_PROC_DIR_ENTRY GCC_NOREGPARM myproc_mkdir(const char *name)
{
  struct proc_dir_entry *p;
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
  p = create_proc_entry(name, S_IFDIR | S_IRUSR | S_IXUSR, NULL);
#else
  p = proc_mkdir(name, NULL);
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,30)
  if (p)
    p->owner = THIS_MODULE;
#endif
  return (VN_PROC_DIR_ENTRY)p;
}
/*****************************************************************************/
VN_PROC_DIR_ENTRY GCC_NOREGPARM mycreate_proc_entry(const char *name, VN_PROC_DIR_ENTRY parent,
                                                    void * read_proc, void * write_proc, void * data)
{
  struct proc_dir_entry *p;

  mode_t mode = ((read_proc)?(S_IRUSR):(0)) | ((write_proc)?(S_IWUSR):(0));
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0)
  p = create_proc_entry(name, mode, (struct proc_dir_entry *)parent);
  if (p)
  {
    p->data = data;
    p->read_proc = read_proc;
    p->write_proc = write_proc;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,30)
    p->owner = THIS_MODULE;
#endif
  }
#else
  struct file_operations my_ops = {
     .read = read_proc,
     .write = write_proc
  };
  p = proc_create(name, mode, (struct proc_dir_entry *)parent, &my_ops);
#endif
  return (VN_PROC_DIR_ENTRY)p;
}
/*****************************************************************************/
void GCC_NOREGPARM myremove_proc_entry(const char *name, VN_PROC_DIR_ENTRY parent)
{
  remove_proc_entry(name, parent);
}

/*****************************************************************************/
void GCC_NOREGPARM myip_send_check(void * ip)
{
  ip_send_check(ip);
}
/*****************************************************************************/
void mypanic(const char * fmt, ...)
{
  char b[1000];
  va_list args;

  va_start(args , fmt);
  vsprintf(b, fmt, args);
  va_end(args);
  panic(b);
}
/*****************************************************************************/
int GCC_NOREGPARM mydel_timer(VN_TIMER timer)
{
  if (timer)
    return del_timer(timer);
  return 0;
}

int GCC_NOREGPARM mydel_timer_sync(VN_TIMER timer)
{
  if (timer)
    return del_timer_sync(timer);
  return 0;
}
/*****************************************************************************/
void GCC_NOREGPARM myadd_timer(VN_TIMER timer)
{
  if (timer)
    add_timer(timer);
}
/*****************************************************************************/
void myschedule(void)
{
  schedule();
}

/*****************************************************************************/

/* FIXME: Uncomment next function when itcskrniface is available
 * myset_affinity sets the thread affinity.
 * Current thread will be bounded to "cpu_id" processor.
 *
 * int GCC_NOREGPARM myset_affinity(unsigned int cpu_id)
 * {
 *     return set_cpus_allowed_ptr(current, &cpumask_of_cpu(cpu_id));
 * }
 */
/*****************************************************************************/
unsigned long GCC_NOREGPARM mycopy_from_user(void * to, const void * from, unsigned long n)
{
  unsigned long r = 0;
  r =
    copy_from_user(to, from, n);
  return r;
}
/*****************************************************************************/
unsigned long GCC_NOREGPARM mycopy_to_user(void * to, const void * from, unsigned long n)
{
  unsigned long r = 0;
  r =
    copy_to_user(to, from, n);
  return r;
}
/*****************************************************************************/

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
typedef irqreturn_t (*HANDLER)(int, void *);
#else
typedef irqreturn_t (*HANDLER)(int, void *, struct pt_regs *);
#endif

int GCC_NOREGPARM myrequest_irq(unsigned int irq, void *handler,
                 const char * device, void * dev_id)
{
  return request_irq(irq, (HANDLER)handler,
#if  LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
                     IRQF_SHARED,
#else
                     SA_SHIRQ,
#endif
                     device, dev_id);
}
/*****************************************************************************/
void GCC_NOREGPARM myfree_irq(unsigned int irq, void * dev_id)
{
  free_irq(irq, dev_id);
}
/*****************************************************************************/
int mypcibios_present(void)
{
//  return pcibios_present();
  return 0;
}
/*****************************************************************************/
void * GCC_NOREGPARM myioremap(unsigned long offset, unsigned long size)
{
  return ioremap(offset, size);
}
/*****************************************************************************/
void GCC_NOREGPARM myiounmap(void * addr)
{
  iounmap(addr);
}
/*****************************************************************************/
int GCC_NOREGPARM myschedule_task(VN_TQSTRUCT task)
{
  struct mytimer_info * info = (struct mytimer_info *)task;
  return schedule_work(&info->ws);
}
/*****************************************************************************/
void GCC_NOREGPARM myflush_scheduled_tasks(void)
{
  flush_scheduled_work();
}
/*****************************************************************************/
int GCC_NOREGPARM mycancel_work_sync(VN_TQSTRUCT tq)
{
  if( NULL == ((struct mytimer_info *)tq)->fn )
  {
    /* task was not initialized, nothing to cancel */
    return 0;
  }
  return mycancel_work_sync_krnface(&((struct mytimer_info *)tq)->ws);
}
/*****************************************************************************/
unsigned long GCC_NOREGPARM myvirt_to_phys(volatile void * address)
{
  return virt_to_phys(address);
}
/*****************************************************************************/
unsigned long GCC_NOREGPARM my__get_free_pages(unsigned long gfp_order)
{
  return __get_free_pages(GFP_ATOMIC, gfp_order);
}
/*****************************************************************************/
void GCC_NOREGPARM myfree_pages(unsigned long addr, unsigned long order)
{
  free_pages(addr, order);
}
/*****************************************************************************/
#if !defined(__arm__)
int GCC_NOREGPARM mypci_read_config_base_addr0(VN_PCIDEV dev, unsigned int *val)
{
  return pci_read_config_dword((struct pci_dev *)dev, PCI_BASE_ADDRESS_0, val);
}
#endif
/*****************************************************************************/
unsigned int GCC_NOREGPARM myreadl(void * addr)
{
  return readl(addr);
}
/*****************************************************************************/
void GCC_NOREGPARM mywritel(unsigned int b, void * addr)
{
  writel(b, addr);
}
/*****************************************************************************/
void myreboot(void)
{
  myprintk( "Initiating system reboot.\n");
  myemergency_restart();
  myprintk( "Reboot didn't ?????\n");
}
/*****************************************************************************/

#if 0
/*****************************************************************************/
void GCC_NOREGPARM myarp_send(int type, int ptype, unsigned long dest_ip,
                VN_DEVICE dev, unsigned long src_ip,
                unsigned char *dest_hw, unsigned char *src_hw,
                unsigned char *target_hw)
{
  arp_send(type, ptype, dest_ip, (LINUX_DEVICE *)dev, src_ip,
           dest_hw, src_hw, target_hw);
}
/*****************************************************************************/
VN_NEIGHBOUR GCC_NOREGPARM my__neigh_lookup(VN_NEIGHTABLE tbl, const void *pkey,
                              VN_DEVICE dev, int create)
{
  return __neigh_lookup((struct neigh_table *)tbl, pkey,
                        (LINUX_DEVICE *)dev, create);
}
/*****************************************************************************/
int GCC_NOREGPARM myneigh_update(VN_NEIGHBOUR neigh, unsigned char *lladdr,
                   unsigned char new, int override, int arp)
{
  return neigh_update((struct neighbour *)neigh, lladdr, new, override, arp);
}
/*****************************************************************************/
void GCC_NOREGPARM myneigh_release(VN_NEIGHBOUR neigh)
{
  neigh_release((struct neighbour *)neigh);
}
#endif //0
/*****************************************************************************/
VN_TASK_STRUCT GCC_NOREGPARM mykthread_run( int GCC_NOREGPARM (*fn)(void *), const char *name )
{
     pid_t pid = -1;
     struct task_struct *tsk = NULL;
     /* Avoid C99 6.7.8, rule 10 violation */
     struct sched_param param = { .sched_priority = MAX_RT_PRIO - 2 };

     if (name == 0 )
     {  
          name = "mykthread_run";
     }
  
#if defined(KERNEL_USE_REGPARM)

     struct _univ_handle * data;
     data = (struct _univ_handle *)mykmalloc(sizeof(struct _univ_handle));
     if (!data)
     {
          udebug(( "mykthread_run: failed to allocate memory for thread handle data!" ));
          return NULL;
     }
     data->fn = fn;
     data->arg = 0;
     data->flclean = 1;
     
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,8,0)
     tsk = kthread_run(univ_handle, data, name ); 
#else
     pid = kernel_thread(univ_handle, data, CLONE_FS | CLONE_FILES | CLONE_SIGHAND);
#endif // VERSION
     
#else //REGPARM
     
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,8,0)
     tsk = kthread_run(fn, 0, name);
#else
     pid = kernel_thread(fn, 0, CLONE_FS | CLONE_FILES | CLONE_SIGHAND);
#endif // VERSION
     
#endif //REGPARM
     
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,8,0)
     if(IS_ERR(tsk))
     {
          printk (KERN_WARNING "mykthread_run: failed to start thread %s \n", name );
          return NULL;
     }
     if( tsk == NULL )
     {
          printk (KERN_WARNING "mykthread_run: tsk == NULL, failed to start thread %s \n", name );
          return NULL;
     }
     pid = tsk->pid;
#else
     if(pid < 0)
     {
          printk (KERN_WARNING "mykthread_run: failed to start thread %s \n", name );
          return NULL;
     }
     tsk = myfind_task_by_pid( pid );
     if( tsk == NULL )
     {
          printk (KERN_WARNING "mykthread_run: tsk == NULL, failed to start thread %s \n", name );
          return NULL;
     }
     if ( tsk->pid != pid )
     {
          printk(KERN_WARNING "mykthread_run: Wrong pid & tsk, tsk->pid != pid. Name = %s \n", name );
          return NULL;
     }
     myput_task_struct_lite(tsk);
#endif //REGPARM
     udebug(("kthread_run executed"));

#ifndef MAX_RT_PRIO
#define MAX_RT_PRIO 100
#endif
     /* Our threads should have prio smaller than "watchdog".
     * It will save us from uneeded "soft lockup" messages
     */
     
     set_user_nice( tsk, -20 );
     
     mysched_setscheduler( tsk, SCHED_FIFO, &param );
     return (VN_TASK_STRUCT)tsk;
}

int GCC_NOREGPARM mykthread_stop( VN_TASK_STRUCT tsk )
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,8,0)
     return kthread_stop( (struct task_struct *) tsk );
#else
     return 0;
#endif
}

void myfinish_kernel_thread( void )
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,8,0)
     while (!kthread_should_stop())
     {
          set_current_state( TASK_INTERRUPTIBLE );
          schedule();
     }
#endif
}

/*****************************************************************************/


int GCC_NOREGPARM myinterruptible_sleep_on(VN_WAITQUEUE p)
{
  WQHANDLEPTR wqh = (WQHANDLEPTR)p;
  if (!p)
    return -ERESTARTSYS;
  wqh->condition = 0;
  return wait_event_interruptible(*((LINUX_WAIT_QUEUE *)wqh->wq),
                                  wqh->condition);
}
/*****************************************************************************/
/*
 * Put current thread to sleep with EXCLUSIVE flag set on.
 * This flag allows waking up only one thread from the queue during
 * mywake_up_interruptible() and all threads during mywake_up_interruptible_all()
 * This function is exactly the same as myinterruptible_sleep_on() for kernels 2.4.x
 */

int GCC_NOREGPARM myinterruptible_sleep_on_exclusive(VN_WAITQUEUE p)
{
  WQHANDLEPTR wqh = (WQHANDLEPTR)p;
  if (!p)
    return -ERESTARTSYS;

  wqh->condition = 0;
  return wait_event_interruptible_exclusive(*((LINUX_WAIT_QUEUE *)wqh->wq),
                                  wqh->condition);
}
/*****************************************************************************/
void GCC_NOREGPARM myset_waitqueue_state(VN_WAITQUEUE q, unsigned long state)
{
  struct list_head *tmp;

  list_for_each(tmp,&((LINUX_WAIT_QUEUE *)q)->task_list) {
    wait_queue_t *curr = list_entry(tmp, wait_queue_t, task_list);

    ((struct task_struct *)(curr->private))->state = state;
  }
}
/*****************************************************************************/
int mysignal_pending_current()
{
    return signal_pending(current);
}

#ifdef DEBUG_F1024
int GCC_NOREGPARM myallow_signal(int sig)
{
    return allow_signal( sig);
}

VN_SIGINFO_T mysiginfo_alloc(void)
{
    return mykmalloc(sizeof(siginfo_t));
}

void GCC_NOREGPARM mysiginfo_free(VN_SIGINFO_T sig_inf)
{
    if ( sig_inf )
        mykfree(sig_inf);
}

unsigned long  GCC_NOREGPARM mydequeue_signal_lock_current(VN_SIGINFO_T info)
{
    return dequeue_signal_lock(current, &current->blocked, (siginfo_t *)info);
}

void myBUG(void )
{
//    printk("BUG: failure at %s:%d/%s()!\n", __FILE__, __LINE__, __FUNCTION__);
//    panic("BUG!");

    BUG();
}
// -- for debug f1024
#endif //DEBUG_F1024

/*****************************************************************************/
void GCC_NOREGPARM mywake_up_interruptible(VN_WAITQUEUE q)
{
  WQHANDLEPTR wqh = (WQHANDLEPTR)q;
  if( !q )
    return;
  wqh->condition = 1;
  wake_up_interruptible((LINUX_WAIT_QUEUE *)wqh->wq);
}

/*****************************************************************************/
void GCC_NOREGPARM mywake_up_interruptible_all(VN_WAITQUEUE q)
{
  WQHANDLEPTR wqh = (WQHANDLEPTR)q;
  if( !q )
    return;
  wqh->condition = 1;
  wake_up_interruptible_all((LINUX_WAIT_QUEUE *)wqh->wq);
}

/*****************************************************************************/
void GCC_NOREGPARM mydaemonize_thread(char *name)
{
#if  LINUX_VERSION_CODE >= KERNEL_VERSION(3,8,0)
	return;
#else
  daemonize(name);
#endif
}
/*****************************************************************************/
void myMOD_INC_USE_COUNT(void)
{
  MOD_INC_USE_COUNT_LALA;
}
/*****************************************************************************/
void myMOD_DEC_USE_COUNT(void)
{
  MOD_DEC_USE_COUNT_LALA;
}
/*****************************************************************************/
void GCC_NOREGPARM myspinlock_lock(VN_SPINLOCK lock)
{
  spin_lock_bh((spinlock_t *)lock);
}
/*****************************************************************************/
void GCC_NOREGPARM myspinlock_unlock(VN_SPINLOCK lock)
{
  spin_unlock_bh((spinlock_t *)lock);
}
/*****************************************************************************/
void GCC_NOREGPARM myspinlock_lock_irqsave(VN_SPINLOCK lock, unsigned long * flags)
{
  spin_lock_irqsave((spinlock_t *)lock, *flags);
}
/*****************************************************************************/
void GCC_NOREGPARM myspinlock_unlock_irqrestore(VN_SPINLOCK lock, unsigned long * flags)
{
  spin_unlock_irqrestore((spinlock_t *)lock, *flags);
}
/*****************************************************************************/
int GCC_NOREGPARM myspinlock_trylock(VN_SPINLOCK lock)
{
  return spin_trylock((spinlock_t *)lock);
}
/*****************************************************************************/
int GCC_NOREGPARM mymutex_lock(VN_MUTEX mutex)
{
/*
 * Changed from down_interruptible()
 * Normally we want mymutex_lock() to return with acquired lock. If you
 * need to interrupt mymutex_lock() during the wait for lock acquiring
 * use mymutex_lock_interruptible()
 */
#ifndef DEBUG_LOCK
  down(mutex);
#else
  int r;
  unsigned long j;
  char sb[256];
  // Try to acquire the lock for 10 seconds. If failef then we are deadlocked!
  for( j = jiffies; jiffies < j + 10 * HZ; )
  {
     r = down_trylock(mutex);
     if( r == 0 ) break;
     schedule();
  }
  if( r != 0 )
  {
    sprint_symbol(sb, (unsigned long)__builtin_return_address(0));
    myprintk("%s: lock {0x%x} at [%x] %s\n",
      __FUNCTION__, (unsigned long)mutex,
      (unsigned long)__builtin_return_address(0), sb);
    dump_stack();
  }
#endif
  return 0;
}
/*****************************************************************************/
int GCC_NOREGPARM mymutex_lock_interruptible(VN_MUTEX mutex)
{
  return down_interruptible(mutex);
}
/*****************************************************************************/
void GCC_NOREGPARM mymutex_unlock(VN_MUTEX mutex)
{
  up(mutex);
}
/*****************************************************************************/
int GCC_NOREGPARM mymutex_trylock(VN_MUTEX mutex)
{
  return down_trylock(mutex);
}
/*****************************************************************************/
void GCC_NOREGPARM myread_lock(VN_RWLOCK rwlock)
{
  VN_MYRWLOCK rw = (VN_MYRWLOCK)rwlock;
#ifdef USE_RWLOCK
#ifdef CONFIG_IPLIR_XFRM_INPUT
  read_lock_bh(&rw->lock);
#else
  read_lock(&rw->lock);
#endif
#else
#ifdef DEBUG_LOCK
  int i,r;
  unsigned long j;
  char sb[256];
  spin_lock(&rw->struct_lock);

  if( rw->pid == current->pid )
  {
       spin_unlock(&rw->struct_lock);
       sprint_symbol(sb, (unsigned long)__builtin_return_address(0));
       myprintk("%s: PID: %d recursive lock {0x%x} at [%x] %s\n",
            __FUNCTION__, current->pid,
            (unsigned long)(&rw->lock),
            (unsigned long)__builtin_return_address(0), sb);
       return;
  }
  spin_unlock(&rw->struct_lock);

#else //DEBUG_LOCK
  down_read(&rw->lock);
#endif //DEBUG_LOCK
#ifdef DEBUG_LOCK
  // Try to acquire the lock for 10 seconds. If failed then we are deadlocked!
  for( j = jiffies; jiffies < j + 10 * HZ; )
  {
     r = down_read_trylock(&rw->lock);
     if( r != 0 ) break;
     schedule();
  }
  if( r == 0 )
  {
    sprint_symbol(sb, (unsigned long)__builtin_return_address(0));
    myprintk("%s: PID: %d deadlock {0x%x} at [%x] %s\n",
      __FUNCTION__, current->pid,
      (unsigned long)(&rw->lock),
      (unsigned long)__builtin_return_address(0), sb);
    sprint_symbol(sb, (unsigned long)(rw->owner_ip));
    myprintk("%s: PID: %d previous %s lock {0x%x} at [%x] %s\n",
      __FUNCTION__, rw->owner_pid,
      (rw->read==-1)?"none":((rw->read)?"read":"write"),
      (unsigned long)(&rw->lock),
      (unsigned long)(rw->owner_ip), sb);
    dump_stack();
  }
  rw->owner_pid = current->pid;
  rw->owner_ip = __builtin_return_address(0);
  rw->read = 1;
#endif /* #ifdef DEBUG_LOCK */
#endif
}
/*****************************************************************************/
void GCC_NOREGPARM myread_unlock(VN_RWLOCK rwlock)
{
  VN_MYRWLOCK rw = (VN_MYRWLOCK)rwlock;
#ifdef USE_RWLOCK
#ifdef CONFIG_IPLIR_XFRM_INPUT
  read_unlock_bh(&rw->lock);
 #else
  read_unlock(&rw->lock);
#endif
#else
#ifdef DEBUG_LOCK
  char sb[256];
  spin_lock(&rw->struct_lock);

  if( rw->pid == current->pid )
  {
       spin_unlock(&rw->struct_lock);
       sprint_symbol(sb, (unsigned long)__builtin_return_address(0));
       myprintk("%s: PID: %d recursive lock {0x%x} at [%x] %s\n",
            __FUNCTION__, current->pid,
            (unsigned long)(&rw->lock),
            (unsigned long)__builtin_return_address(0), sb);
       return;
  }
  spin_unlock(&rw->struct_lock);
  rw->owner_pid = -1;
#endif /* #ifdef DEBUG_LOCK */
  up_read(&rw->lock);
#endif
}
/*****************************************************************************/
void GCC_NOREGPARM mywrite_lock(VN_RWLOCK rwlock)
{
  VN_MYRWLOCK rw = (VN_MYRWLOCK)rwlock;
#ifdef USE_RWLOCK
#ifdef CONFIG_IPLIR_XFRM_INPUT
  write_lock_bh(&rw->lock);
#else
  write_lock(&rw->lock);
#endif
#else
#ifdef DEBUG_LOCK
  int r;
  unsigned long j;
  char sb[256];
  spin_lock(&rw->struct_lock);
  if( rw->pid == current->pid )
  {
       rw->count++;
       myprintk("Recursive write lock... Current pid is %d\n", current->pid);
       spin_unlock(&rw->struct_lock);
       return;
  }
  spin_unlock(&rw->struct_lock);

#else
  down_write(&rw->lock);
#endif
#ifdef DEBUG_LOCK
  // Try to acquire the lock for 10 seconds. If failed then we are deadlocked!
  for( j = jiffies; jiffies < j + 10 * HZ; )
  {
     r = down_write_trylock(&rw->lock);
     if( r != 0 ) break;
     schedule();
  }
  if( r == 0 )
  {
    sprint_symbol(sb, (unsigned long)__builtin_return_address(0));
    myprintk("%s: PID: %d deadlock {0x%x} at [%x] %s\n",
      __FUNCTION__, current->pid,
      (unsigned long)(&rw->lock),
      (unsigned long)__builtin_return_address(0), sb);
    sprint_symbol(sb, (unsigned long)(rw->owner_ip));
    myprintk("%s: PID: %d previous %s lock {0x%x} at [%x] %s\n",
      __FUNCTION__, rw->owner_pid,
      (rw->read==-1)?"none":((rw->read)?"read":"write"),
      (unsigned long)(&rw->lock),
      (unsigned long)(rw->owner_ip), sb);
    dump_stack();
  }
  if(rw->owner_pid != -1)
  {
    sprint_symbol(sb, (unsigned long)(rw->owner_ip));
    myprintk("%s: PID: %d previous %s lock {0x%x} at [%x] %s\n",
      __FUNCTION__, rw->owner_pid,
      (rw->read)?"read":"write",
      (unsigned long)(&rw->lock),
      (unsigned long)(rw->owner_ip), sb);
  }
  rw->owner_pid = current->pid;
  rw->owner_ip = __builtin_return_address(0);
  rw->read = 0;
  // lock
  spin_lock(&rw->struct_lock);
  if(rw->pid != -1 || rw->count != 0)
  {
       myprintk("Warning write lock is broken. Pid=%d, Locked %d times. Current pid is %d\n", rw->pid, rw->count, current->pid);
	   dump_stack();
  }
  rw->pid = current->pid;
  rw->count = 1;
  //unlock
  spin_unlock(&rw->struct_lock);
#endif  /* #ifdef DEBUG_LOCK */
#endif
}
/*****************************************************************************/
void GCC_NOREGPARM mywrite_unlock(VN_RWLOCK rwlock)
{
  VN_MYRWLOCK rw = (VN_MYRWLOCK)rwlock;
#ifdef USE_RWLOCK
#ifdef CONFIG_IPLIR_XFRM_INPUT
  write_unlock_bh(&rw->lock);
#else
  write_unlock(&rw->lock);
#endif
#else
#ifdef DEBUG_LOCK
  spin_lock(&rw->struct_lock);
  if(rw->pid != current->pid || rw->count <= 0)
  {
       myprintk("Warning trying to unlock unlocked lock. Pid=%d, Locked %d times. Current pid is %d\n", rw->pid, rw->count, current->pid);
  }

  if( rw->count > 1 )
  {
       rw->count--;
       spin_unlock(&rw->struct_lock);
       return;
  }

  rw->count = 0;
  rw->pid = -1;
  spin_unlock(&rw->struct_lock);
  rw->owner_pid = -1;
#endif /* #ifdef DEBUG_LOCK */
  up_write(&rw->lock);
#endif
}

/*****************************************************************************/
void GCC_NOREGPARM myread_lock_bh(VN_RWLOCK rwlock)
{
  VN_MYRWLOCK rw = (VN_MYRWLOCK)rwlock;
#ifdef USE_RWLOCK
  read_lock_bh(&rw->lock);
#else
  // @todo
#endif
}
/*****************************************************************************/
void GCC_NOREGPARM myread_unlock_bh(VN_RWLOCK rwlock)
{
  VN_MYRWLOCK rw = (VN_MYRWLOCK)rwlock;
#ifdef USE_RWLOCK
  read_unlock_bh(&rw->lock);
#else
  // @todo
#endif
}
/*****************************************************************************/
void GCC_NOREGPARM mywrite_lock_bh(VN_RWLOCK rwlock)
{
  VN_MYRWLOCK rw = (VN_MYRWLOCK)rwlock;
#ifdef USE_RWLOCK
  write_lock_bh(&rw->lock);
#else
  // @todo
#endif
}
/*****************************************************************************/
void GCC_NOREGPARM mywrite_unlock_bh(VN_RWLOCK rwlock)
{
  VN_MYRWLOCK rw = (VN_MYRWLOCK)rwlock;
#ifdef USE_RWLOCK
  write_unlock_bh(&rw->lock);
#else
  // @todo
#endif
}

void mylocal_bh_disable()
{
     local_bh_disable();
}

void mylocal_bh_enable()
{
     local_bh_enable();
}

void mypreempt_disable()
{
     preempt_disable();
}

void mypreempt_enable()
{
     preempt_enable();
}



/*****************************************************************************/
int myin_softirq(void)
{
  return in_softirq();
}
/*****************************************************************************/
void GCC_NOREGPARM myoutb(int x, int y)
{
  outb(x, y);
}
/*****************************************************************************/
void GCC_NOREGPARM myfill_nfinfo(VN_SKB skb, VN_NFINFO nfinfo, int IsIncoming)
{
  struct sk_buff * skbuf = (struct sk_buff *)skb;
#ifdef NF_USES_QUEUE_ENTRY
  struct nf_queue_entry * nfi = (struct nf_queue_entry *)nfinfo;
  const struct nf_afinfo * afinfo;
#else
  struct nf_info * nfi = (struct nf_info *)nfinfo;
#endif
  /* We don't know how to fill non IPv4 packets */
  if( skbuf->protocol != htons(ETH_P_IP) )
    return;

  /* nf_info is already filled */
  if(IsIncoming == 1 && nfi->indev == skbuf->dev)
  {
    return;
  }
  if(IsIncoming == 0 && nfi->outdev == skbuf->dev)
  {
    return;
  }

  /* Decrease refcouting. */
  if (nfi->indev)
  {
       mydev_put(nfi->indev);       
  }
  if (nfi->outdev)
  {
       mydev_put(nfi->outdev);       
  }

  /* put refcount in case it was acuired before */
  if(nfi->elem && nfi->elem->owner)
  {
       module_put(nfi->elem->owner);       
  }

  /* Fill nf_info for incoming packet */
  if(IsIncoming == 1)
  {
    nfi->hook = NF_INET_PRE_ROUTING;
    nfi->indev = skbuf->dev;
    nfi->outdev = NULL;
    nfi->elem = &nf_ops_inet_in;
    nfi->okfn = myip_rcv_finish;
  }
  /* Fill nf_info for outgoing packet */
  else
  {
    nfi->hook = NF_INET_POST_ROUTING;
    nfi->indev = NULL;
    nfi->outdev = skbuf->dev;
    nfi->elem = &nf_ops_inet_out;
    nfi->okfn = myip_finish_output;
  }

  /* Put refcouting in order. --vel */
  if (nfi->indev)
  {
       mydev_hold(nfi->indev);
  }
  if (nfi->outdev)
  {
       mydev_hold(nfi->outdev);
  }   

#ifdef NF_USES_QUEUE_ENTRY
  /*
  ** This is why nf_queue_entry is bigger than its sizeof - additional
  ** routing info is saved at its end. --vel
  */
  afinfo = nf_get_afinfo(PF_INET);
  afinfo->saveroute(skbuf, nfi);
#endif

  /* Refcounting again, this time for module. --vel */
  if(nfi->elem && nfi->elem->owner)
    try_module_get(nfi->elem->owner);
}

void GCC_NOREGPARM myskb_dst_drop(VN_SKB skb)
{
     if (!kernel_skb_dst_is_empty((struct sk_buff *)skb))
     {
          kernel_skb_dst_drop((struct sk_buff *)skb);
          kernel_skb_dst_set((struct sk_buff *)skb, 0);
     }
}

/*****************************************************************************/
void GCC_NOREGPARM myfill_skb(VN_SKB skb)
{
     struct rtable *rt;
     struct sk_buff * skbuf = (struct sk_buff *)skb;
     const struct iphdr *iph = kernel_skb_ip_hdr(skbuf);

     /* Do not change anything in skb for non IP */
     if (skbuf->protocol != htons(ETH_P_IP))
     {
          return;
     }

     if (kernel_skb_dst_is_empty(skbuf))
     {
          if (0 == myip_check_route_output_key(&rt, iph->daddr))
          {
               kernel_skb_dst_set(skbuf, kernel_rtable_dst(rt));
          }
          else if (iph->daddr != 0xFFFFFFFF)
          {
               myprintk("ip_route_output_key() to %d.%d.%d.%d failed!!!\n",
                    (iph->daddr & 0xFF000000) >> 24, (iph->daddr & 0x00FF0000) >> 16,
                    (iph->daddr & 0x0000FF00) >> 8, iph->daddr & 0x000000FF
               );
          }
     }
}

/*****************************************************************************/
int GCC_NOREGPARM myskb_make_solid(VN_SKB *skb, VN_NFINFO nfinfo)
{
     struct sk_buff* sb = (struct sk_buff *)(*skb);
     if( skb_cloned( sb ) )
     {
    	  return pskb_expand_head(sb, 0, 0, GFP_ATOMIC);
     }
     return 0;
}


/*****************************************************************************/
void GCC_NOREGPARM mynf_reinject(VN_SKB skb, VN_NFINFO nfinfo)
{
#ifdef CONFIG_NETFILTER
#ifdef NF_USES_QUEUE_ENTRY
     struct nf_queue_entry * nfi = (struct nf_queue_entry *)nfinfo;
#else
     struct nf_info * nfi = (struct nf_info *)nfinfo;
#endif /* #ifdef NF_USES_QUEUE_ENTRY */
     struct sk_buff * skbuf = (struct sk_buff *)skb;
     unsigned int daddr = kernel_skb_ip_hdr(skbuf)->daddr;
     local_bh_disable();
     /*
     ** This macro gives us portable way of reinjecting, verdict is always
     ** "accept" here. --vel
     ** Except for the cases when the route is not found.
     */
     if (skbuf->protocol == htons(ETH_P_IP) && nfi->okfn == 0)
     {
          myprintk("Warning! Cannot return new skb. Drop packet.\n" );
          NF_REINJECT_COMPAT(skb, nfinfo, NF_DROP);
          local_bh_enable();
          return;
     }
     
     if ((nfi->pf == PF_INET) && (nfi->hook == NF_INET_POST_ROUTING))
     {
          if ( kernel_skb_dst_is_empty(skbuf) || kernel_skb_dst_metrics(skbuf) == 0 )
          {
               if (daddr != 0xFFFFFFFF)
               {
                    // @todo check, why do we have this message, currently we're just hiding it from user( there was prink trace)
                    udebug(("Warning! Cannot find route to %d.%d.%d.%d. Metrics = 0x%x. Drop packet.\n",
                         daddr & 0x000000FF, (daddr & 0x0000FF00) >> 8,
                         (daddr & 0x00FF0000) >> 16, (daddr & 0xFF000000) >> 24,
                         (kernel_skb_dst_is_empty(skbuf))?(0xFFFFFFFF):(kernel_skb_dst_metrics(skbuf))
                    ));
               }
               NF_REINJECT_COMPAT(skb, nfinfo, NF_DROP);
               local_bh_enable();
               return;
          }
     }
     
#ifdef  DRV_TRACE_RUNTIME     
     if ( trace_mask & TRACE_BIT_OUTOF_DRV )
     {
          MYDUMP_SKB_IPHDR( "nf_reinject", skb);
     }
     if ( trace_mask & TRACE_BIT_OUT_SKB )
     {
          mydump_skb(skb);
     }
#endif

     NF_REINJECT_COMPAT(skb, nfinfo, NF_ACCEPT);
     local_bh_enable();
#endif
}
/*****************************************************************************/
/*
** Clone nf_info structure and appropriate locks. This is needed when we
** have one packet came in and more than one packet must go out (due to
** fragmentation). Since nf_reinject() call frees nf_info structure we pass
** to it, we need to clone it for each fragment. We also redo most of
** the work done by nf_queue(), so that nf_reinject() will not be confused
** when it releases some device locks that were put by nf_queue(). --vel
*/
/*
** Now this function not only clones structure contents, but also allocates
** the structure. This is necessary because actual size of nf_queue_entry
** is not determined by it's sizeof() - the actual size is bigger.
** Here skb is the already cloned skb (NOT the original one, the caller must
** clone it by himself), nfi_src is the original nfinfo/nf_queue_entry
** (whichever is used), and nfi_dst is the pointer to variable that will be
** filled with cloned nfinfo/nf_queue_entry. The caller must not allocate
** nfi_dst, it is allocated here. --vel
*/
void GCC_NOREGPARM myclone_nf_info(VN_NFINFO nfi_src, VN_NFINFO * nfi_dst, VN_SKB skb )
{
#ifdef CONFIG_NETFILTER
#ifdef NF_USES_QUEUE_ENTRY
  struct nf_queue_entry * src, * dst;
  const struct nf_afinfo * afinfo;
  struct sk_buff * askb = (struct sk_buff *)skb;
#else
  struct nf_info * src, * dst;
#endif
  int nfq_size;

/*
** New kernels. We calculate the correct size of nf_queue_entry for given protocol
** using address family info for that protocol. --vel
*/
#ifdef NF_USES_QUEUE_ENTRY
  src = (struct nf_queue_entry *)nfi_src;
  afinfo = nf_get_afinfo(PF_INET);
  if (!afinfo)
    myprintk("clone: NULL afinfo\n");
  nfq_size = sizeof(struct nf_queue_entry);
  if (afinfo)
    nfq_size += afinfo->route_key_size;
#else
  /* For old kernels we just take sizeof. */
  src = (struct nf_info *)nfi_src;
  nfq_size = sizeof(struct nf_info);
#endif
  /* Allocate the clone and copy the contents. */
  dst = mykmalloc(nfq_size);
  /* drvlib memcpy() using there waits arguments in stack. GCC passes them in registers because CONFIG_REGPARM param is set.
   * mymemcpy() is a local copy of memcpy(). It is introduced in order
   * to eliminate argument passing error.
   * Newer kernels declare memcpy as __always_inline and therefor drvlib memcpy() isn't used. */
  mymemcpy(dst, src, nfq_size);
  /* For new kernels, skb is inside the nf-queue_entry, so we must fill it. --vel */
#ifdef NF_USES_QUEUE_ENTRY
  dst->skb = skb;
#endif
  /* Put refcouting in order. --vel */
  if (dst->indev)
    mydev_hold(dst->indev);
  if (dst->outdev)
    mydev_hold(dst->outdev);

#ifdef NF_USES_QUEUE_ENTRY
/* Bridged packets require additional handling if bridging is enabled. --vel */
#ifdef CONFIG_BRIDGE_NETFILTER
  if (askb->nf_bridge)
  {
    if (askb->nf_bridge->physindev)
      mydev_hold(askb->nf_bridge->physindev);
    if (askb->nf_bridge->physoutdev)
      mydev_hold(askb->nf_bridge->physoutdev);
  }
#endif
  /*
  ** This is why nf_queue_entry is bigger than its sizeof - additional
  ** routing info is saved at its end. --vel
  */
  afinfo->saveroute(askb, dst);
#endif

  /* Refcounting again, this time for module. --vel */
  try_module_get(dst->elem->owner);
  *nfi_dst = dst;
#endif
}

/* mycreate_nf_info is similair to myclone_nf_info. It fills VN_NFINFO with default values */
void GCC_NOREGPARM mycreate_nf_info(VN_NFINFO * nfi_dst,
                        VN_SKB skb, VN_DEVICE physindev, VN_DEVICE physoutdev)
{
#ifdef CONFIG_NETFILTER
#ifdef NF_USES_QUEUE_ENTRY
  struct nf_queue_entry * dst;
  const struct nf_afinfo * afinfo;
#else
  struct nf_info * dst;
#endif
  int nfq_size;

/*
** New kernels. We calculate the correct size of nf_queue_entry for given protocol
** using address family info for that protocol. --vel
*/
#ifdef NF_USES_QUEUE_ENTRY
  afinfo = nf_get_afinfo(PF_INET);
  if (!afinfo)
    myprintk("clone: NULL afinfo\n");
  nfq_size = sizeof(struct nf_queue_entry);
  if (afinfo)
    nfq_size += afinfo->route_key_size;
#else
  /* For old kernels we just take sizeof. */
  nfq_size = sizeof(struct nf_info);
#endif
  /* Allocate the clone and copy the contents. */
  dst = mykmalloc(nfq_size);
  /* Set default values */
#ifdef NF_USES_QUEUE_ENTRY
   dst->list.next = 0;
   dst->list.prev = 0;
   dst->skb = (struct sk_buff *)skb;
   dst->id = 0;
#endif
   dst->elem = 0;
   dst->pf = PF_INET; /* We create nf_info only for IPv4 packets */
   dst->hook = 0;
   dst->indev = 0;
   dst->outdev = 0;
   dst->okfn = 0;
  *nfi_dst = dst;
#endif
}
/*****************************************************************************/
/*
** Release all locks held by this packet, like nf_reinject() does. This is
** needed because sometimes after fragmentation the original packet doesn't
** go through nf_reinject() and goes to the list of reserved packets (so it's
** almost freed at that point). In such cases we need to manually release all
** locks put by nf_queue(). --vel
*/
/*
** Now this function is very simple - it calls nf_reinject with NF_DROP to
** free nfinfo/nf_queue_entry. However nf_reinject also frees skb, so this
** function is now renamed accordingly. Caller must not forget that skb is
** also freed after call to this function. --vel
*/
void GCC_NOREGPARM myfree_nf_info_and_skb( VN_NFINFO nfq, VN_SKB skb )
{
#ifdef CONFIG_NETFILTER
#ifdef NF_USES_QUEUE_ENTRY
  struct nf_queue_entry * nfi = (struct nf_queue_entry *)nfq;
#else
  struct nf_info * nfi = (struct nf_info *)nfq;
#endif // NF_USES_QUEUE_ENTRY
  /* skb is empty then we have to free only nfinfo */
  if( skb == NULL )
  {
    myfree_nf_info_only(nfi);
  }
  /* We have created this packet. So we don't need to pass it into kernel */
  else if( nfi->elem == 0 && nfi->okfn == 0 )
  {
       /// Reaching this part of code only means this dummy packet was created by us
       /// and reference counters weren't increased
       mykfree(nfq);
       mykfree_skb(skb);
  }
  else
  {
    NF_REINJECT_COMPAT(skb, nfq, NF_DROP);
  }
#endif
}
/*****************************************************************************/
void GCC_NOREGPARM myfree_nf_info_only(VN_NFINFO nfq)
{
#ifdef CONFIG_NETFILTER
     myrelease_nf_refs( nfq );
     kfree( nfq );
#endif
}

void GCC_NOREGPARM myrelease_nf_refs(VN_NFINFO nfq)
{
#ifdef CONFIG_NETFILTER
#ifdef NF_USES_QUEUE_ENTRY
  struct nf_queue_entry * nfi = (struct nf_queue_entry *)nfq;
#else
  struct nf_info * nfi = (struct nf_info *)nfq;
#endif // NF_USES_QUEUE_ENTRY

  /* Decrease refcouting. */
  if (nfi->indev)
  {
       mydev_put(nfi->indev);
  }
  if (nfi->outdev)
  {
       mydev_put(nfi->outdev);
  }
  /* put refcount in case it was acuired before */
  if(nfi->elem && nfi->elem->owner)
  {
       module_put(nfi->elem->owner);
  }

#endif // CONFIG_NETFILTER
}

/*****************************************************************************/
int GCC_NOREGPARM myfind_best_route( unsigned int destinIp, unsigned int *sourceIp,
                  unsigned int *nextHost, VN_DEVICE *dev )
{
     return myfind_best_route_and_set( destinIp, sourceIp, nextHost, dev, 0);
}
/*****************************************************************************/
int GCC_NOREGPARM myfind_best_route_and_set( u32 destinIp, u32 *sourceIp,
                  u32 *nextHost, VN_DEVICE *dev, VN_SKB skb )
{
     int rc;
     struct rtable *rt;

     if ( ( rc = myip_route_output_key( &rt, destinIp, sourceIp ) ) != 0 )
     {
          return rc;
     }


#if LINUX_VERSION_CODE < KERNEL_VERSION(3,8,0)
     *nextHost = rt->rt_gateway;
#else
     *nextHost  = rt_nexthop(rt,destinIp);
#endif
     *dev = (VN_DEVICE) kernel_rtable_dev(rt);
     if (skb) {
          kernel_skb_dst_set( ( struct sk_buff* )skb, ( struct dst_entry* )rt );
     }
     else
     // No need to keep this route
     {
          ip_rt_put(rt);
     }

     return 0;
}
/*****************************************************************************/
int GCC_NOREGPARM myneigh_lookup( VN_DEVICE device, unsigned int* ipAddress,
                  unsigned char* macAddress, unsigned int* macAddressSize )
{
     struct net_device* dev = (struct net_device*)device;
     struct neighbour* n;

     if( *macAddressSize < dev->addr_len )
          return -1;

     n = neigh_lookup(&arp_tbl, ipAddress, dev);
     if ( !n )
          return -2;

     read_lock_bh(&n->lock);
     mymemcpy(macAddress, n->ha, dev->addr_len);
     *macAddressSize = dev->addr_len;
     read_unlock_bh(&n->lock);
     neigh_release(n);

     return 0;
}
/*****************************************************************************/
VN_SKB GCC_NOREGPARM myip_defrag(VN_SKB skb)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
     return (VN_SKB)ip_defrag( (struct sk_buff *)skb, IP_DEFRAG_LOCAL_DELIVER );
#else
     if( ip_defrag( (struct sk_buff *)skb, IP_DEFRAG_LOCAL_DELIVER ) == 0 )
     {
          return skb;
     }
     else
     {
          return NULL;
     }
#endif
}
/*****************************************************************************/
int mysmp_processor_id(void)
{
  return smp_processor_id();
}
/*****************************************************************************/
int myin_interrupt(void)
{
  return in_interrupt();
}
/*****************************************************************************/
void GCC_NOREGPARM mydst_hold(VN_DST dst)
{
  dst_hold((struct dst_entry *)dst);
}
/*****************************************************************************/
void GCC_NOREGPARM mydst_release(VN_DST dst)
{
  dst_release((struct dst_entry *)dst);
}
/*****************************************************************************/
VN_DEV GCC_NOREGPARM mydst_dev(VN_DST mydst)
{
     struct dst_entry *dst = mydst;
     struct net_device *dev = dst->dev;
     if(!dst || !dev || !dev->name)
          return NULL;
     /* Ignore loopback devices */
     if (mydev_fetch_flags(dev) & myIFF_LOOPBACK())
          return NULL;
     return dst->dev;
}
/*****************************************************************************/
void GCC_NOREGPARM mydump_skb(VN_SKB skb)
{
#ifdef DEBUG_SHOW_STACK
  struct sk_buff * s = (struct sk_buff *)skb;
  ptrdiff_t length;
  const unsigned char* skb_tail_ptr = NULL;
  const unsigned char* skb_end_ptr = NULL;

  if (!s)
  {
       return;
  }

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
  skb_tail_ptr = s->tail;
  skb_end_ptr = s->end;
#else
  skb_tail_ptr = skb_tail_pointer(s);
  skb_end_ptr = skb_end_pointer(s);
#endif


  myprintk("dump_skb: skb addr %p shared/cloned %d/%d\n"
           , s
           , skb_shared(s), skb_cloned(s)
           );
  myprintk("dump_skb: pkt_type=%d protocol=%x ip_summed=%d len=%d data_len=%d csum=%08x\n",
           s->pkt_type, s->protocol, s->ip_summed, s->len, s->data_len, s->csum);

  myprintk("dump_skb: head=%x data=%x tail=%x end=%x\n",
           (s->head) ? s->head : NULL,
           (s->data) ? s->data : NULL,
           (s->tail) ? skb_tail_ptr : NULL,
           (s->end) ? skb_end_ptr : NULL);

  myprintk("dump_skb: shared_info: nr_frags=%d frag_list=0x%x\n",
           skb_shinfo(s) ? skb_shinfo(s)->nr_frags : 0,
           ( skb_shinfo(s) && skb_shinfo(s)->frag_list ) ? skb_shinfo(s)->frag_list : 0);

  if (s->data && s->tail)
  {
       length = skb_tail_ptr - s->data;

       myprintk("dump_skb: size (tail-data)=%d\n", length);
       myprintk("dump_skb: headlen=%d headroom=%d tailroom=%d\n",
                skb_headlen(s), skb_headroom(s), skb_tailroom(s));
  }

  myprintk("dump_skb: mac.raw=%x nh.raw=%x h.raw=%x\n",
           myskb_get_mac_raw(s),
           myskb_get_nh_raw(s),
           myskb_get_h_raw(s));

  myprintk("dump_skb: packet dump:\n");
  if ( myskb_is_nonlinear(s) )
  {
      length = skb_tail_ptr - s->data;
      mydumphex(s->data, length);
  }
  else
  {
       mydumphex(s->data, s->len);
  }
#endif /*DEBUG_SHOW_STACK*/
}
void GCC_NOREGPARM mydump_skb_iphdr(const char *label, const char *func, const int line, VN_SKB skb)
{
     struct sk_buff *s;
     struct rtable *rt;
     struct net_device *dev;

     if (!skb)
     {
          return;
     }

     s = (struct sk_buff *)skb;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,31)
     rt = 0;
     dev = 0;
#else
     rt = skb_rtable(s);
     dev = rt ? kernel_rtable_dev(rt) : 0;
#endif

#define XIP(v) (v)&0xFF, ((v) >> 8)&0xFF, ((v) >> 16)&0xFF, ((v) >> 24)&0xFF

     if( s->protocol == htons(ETH_P_IP) )
     {

          const struct iphdr *iph = kernel_skb_ip_hdr(s);
          myprintk("%s:%d %s ptr=%p, data=%p, dev=%p, shared/cloned=%d/%d, len=%u, id=%u, frag_off=%u, ttl=%u, proto=%u, check=%u, src=%d.%d.%d.%d, dst=%d.%d.%d.%d, dev-name=%p/%s, rt-dev=%p/%s\n",
               func, line, label,
               s,
               s->data,
               s->dev,
               skb_shared(s),
               skb_cloned(s),
               (unsigned)ntohs(iph->tot_len),
               (unsigned)iph->id,
               (unsigned)iph->frag_off,
               (unsigned)iph->ttl,
               (unsigned)iph->protocol,
               (unsigned)iph->check,
               XIP(iph->saddr),
               XIP(iph->daddr),
               s->dev,
               s->dev ? s->dev->name : "NULL",
               dev,
               dev ? dev->name : "NULL"
               );
          
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,21)
               if ( iph->protocol == IPPROTO_UDP )
               {
                    struct udphdr *udp_header;
                    if ( skb_transport_header(s) == skb_network_header(s) )
                    {
                         udp_header = (struct udphdr *)skb_network_header(s) + iph->tot_len;
                    }
                    else
                    {
                         udp_header = (struct udphdr *)skb_transport_header(s); 
                    }
                    myprintk("%s:%d %s udp src=%d dst=%d \n", func, line, label, 
                               ntohs((unsigned short int) udp_header->source), 
                               ntohs((unsigned short int) udp_header->dest) 
                            );
               }
               
               if ( iph->protocol == IPPROTO_TCP )
               {
                    struct tcphdr *tcp_header = (struct tcphdr *)skb_transport_header(s); 
                    
                    myprintk("%s:%d %s tcp src=%d dst=%d seq=%d aseq=%d \n", func, line, label, 
                              ntohs((unsigned short int) tcp_header->source), 
                              ntohs((unsigned short int) tcp_header->dest),
                              ntohl((unsigned int) tcp_header->seq),
                              ntohl((unsigned int) tcp_header->ack_seq)
                            );
               }
#endif

     }
     else
     {
          myprintk("%s:%d %s ptr=%p, data=%p, dev=%p, shared/cloned=%d/%d, dev-name=%p/%s, rt-dev=%p/%s\n",
               func, line, label,
               s,
               s->data,
               s->dev,
               skb_shared(s),
               skb_cloned(s),
               s->dev,
               s->dev ? s->dev->name : "NULL",
               dev,
               dev ? dev->name : "NULL"
               );
     }
#undef XIP
}

void GCC_NOREGPARM mydump_nfinfo(const char *func, const int line, VN_NFINFO nfinfo)
{
#ifdef NF_USES_QUEUE_ENTRY
     struct nf_queue_entry * nfi = (struct nf_queue_entry *)nfinfo;
     const struct nf_afinfo *afinfo = nf_get_afinfo(nfi->pf);
     myprintk("%s:%d ptr=%p, skb=%p, id=%d, pf=%d, afinfo=%p, indev=%p/%s, outdev=%p/%s\n"
          , func, line
          , nfi, nfi->skb, nfi->id
          , nfi->pf, afinfo
          , nfi->indev, nfi->indev ? nfi->indev->name : "NULL"
          , nfi->outdev, nfi->outdev ? nfi->outdev->name : "NULL"
          );
#else
     struct nf_info * nfi = (struct nf_info *)nfinfo;
     myprintk("%s:%d ptr=%p, pf=%d, indev=%p/%s, outdev=%p/%s\n"
          , func, line
          , nfi, nfi->pf
          , nfi->indev, nfi->indev ? nfi->indev->name : "NULL"
          , nfi->outdev, nfi->outdev ? nfi->outdev->name : "NULL"
          );
#endif /* #ifdef NF_USES_QUEUE_ENTRY */
}

/*****************************************************************************/
VN_ATOMIC myatomic_alloc(void)
{
  return mykmalloc(sizeof(atomic_t));
}
/*****************************************************************************/
void GCC_NOREGPARM myatomic_free(VN_ATOMIC a)
{
  if (a)
    mykfree(a);
}
/*****************************************************************************/
int GCC_NOREGPARM myatomic_read(VN_ATOMIC a)
{
  return atomic_read((atomic_t *)a);
}
/*****************************************************************************/
void GCC_NOREGPARM myatomic_set(VN_ATOMIC a, int value)
{
  atomic_set((atomic_t *)a, value);
}
/*****************************************************************************/
void GCC_NOREGPARM myatomic_inc(VN_ATOMIC a)
{
  atomic_inc((atomic_t *)a);
}
/*****************************************************************************/
void GCC_NOREGPARM myatomic_dec(VN_ATOMIC a)
{
  atomic_dec((atomic_t *)a);
}
/*****************************************************************************/
int GCC_NOREGPARM myatomic_dec_and_test(VN_ATOMIC a)
{
  return atomic_dec_and_test((atomic_t *)a);
}
/*****************************************************************************/
/*****************************************************************************/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,15)
#    define atomic_cmpxchg(v, oldval, newval) \
          ( (int)cmpxchg(&((v)->counter), (oldval), (newval)) )
#endif /* #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,15) */

/*****************************************************************************/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
#    define atomic_xchg(v, newval) \
          ( xchg(&((v)->counter), (newval)) )
#endif /* #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16) */

/*****************************************************************************/
int GCC_NOREGPARM myatomic_inc_return(VN_ATOMIC a)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
  return atomic_add_return(1, (atomic_t *)a);
#else
  return atomic_inc_return((atomic_t *)a);
#endif
}
/*****************************************************************************/
int GCC_NOREGPARM myatomic_dec_return(VN_ATOMIC a)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
  return atomic_add_return(-1, (atomic_t *)a);
#else
  return atomic_dec_return((atomic_t *)a);
#endif
}

#ifndef LOCK_PREFIX
#define LOCK_PREFIX LOCK
#endif
///*****************************************************************************/
long GCC_NOREGPARM myatomic32_add_return(volatile long * dest, long val)
{
#if defined(HAVE_KERNEL_ATOMIC_LONG_RETURN) &&  BITS_PER_LONG == 32
     return atomic_long_add_return(val, (atomic_long_t *)dest);
#else
     /* fallback for int, we assume sizeof(int) = 32bits */
     return (long)atomic_add_return(val, (atomic_t *)dest);
#endif
}

void GCC_NOREGPARM myatomic32_add(volatile long * dest, long val)
{
     myatomic32_add_return(dest, val);
}

long GCC_NOREGPARM myatomic32_inc_return(volatile long * dest)
{
     return myatomic32_add_return(dest, 1);
}
long GCC_NOREGPARM myatomic32_dec_return(volatile long * dest)
{
     return myatomic32_add_return(dest, -1);
}
long GCC_NOREGPARM myatomic32_cmpxchg(volatile long * dest, long old, long new)
{
      return atomic_cmpxchg((atomic_t *)dest, old, new);
}

long GCC_NOREGPARM myatomic32_xchg(volatile long * dest, long value)
{
/* we define l for kernel headers BUG workaround */
#define l dest
      return atomic_xchg((atomic_t *)dest, value);
#undef l
}

///*****************************************************************************/

#if defined(CONFIG_PPC32) || defined(CONFIG_PPC64)
static __always_inline unsigned long
powerpc_cmpxchg_u64(volatile unsigned long *p, unsigned long old, unsigned long new)
{
     unsigned long prev;

     __asm__ __volatile__ (
     PPC_RELEASE_BARRIER
"1:  ldarx     %0,0,%2          # __cmpxchg_u64\n\
     cmpd     0,%0,%3\n\
     bne-     2f\n\
     stdcx.     %4,0,%2\n\
     bne-     1b"
     PPC_ACQUIRE_BARRIER
     "\n\
2:"
     : "=&r" (prev), "+m" (*p)
     : "r" (p), "r" (old), "r" (new)
     : "cc", "memory");

     return prev;
}
#endif /* defined(CONFIG_PPC32) || defined(CONFIG_PPC64) */

static inline unsigned long long mycmpxchg64(
                                             volatile unsigned long long *ptr,
                                             unsigned long long oldval,
                                             unsigned long long newval)
{
#if defined(CONFIG_PPC32) || defined(CONFIG_PPC64)
     return powerpc_cmpxchg_u64(ptr, oldval, newval);
#else
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,25)
#ifdef CONFIG_X86

     volatile unsigned long long prev;
#if defined (__x86_64__)
     asm volatile(LOCK_PREFIX "cmpxchgq %2,%1"
                  : "=a" (prev), 
                    "+m" (*ptr)
                  : "r" (newval), 
                    "0" (oldval)
                  : "memory");
#else
     asm volatile(LOCK_PREFIX "cmpxchg8b %1"
                  : "=A" (prev),
                    "+m" (*ptr)
                  : "b" ((unsigned int)newval),
                    "c" ((unsigned int)(newval >> 32)),
                    "0" (oldval)
                  : "memory");
#endif
     return prev;

#else /* #ifdef CONFIG_X86 */
     volatile unsigned long long prev;
     unsigned long flags;

     local_irq_save(flags);
     prev = *ptr;
     if (prev == oldval)
     {
          *ptr = newval;
     }
     local_irq_restore(flags);
     return prev;
#endif /* #ifndef CONFIG_X86 */

#else /* #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,25) */
     return cmpxchg64(ptr, oldval, newval);
#endif
#endif /* defined(CONFIG_PPC32) || defined(CONFIG_PPC64) */
}

long long GCC_NOREGPARM myatomic64_add_return(volatile long long * dest, long val)
{
     long long cmp, set;

     do
     {
          cmp = *( long long * )dest;
          set = cmp + val;
     } while( cmp != mycmpxchg64( (volatile unsigned long long *)dest, (unsigned long long)cmp, (unsigned long long)set ) );

     return set;
}

void GCC_NOREGPARM myatomic64_add(volatile long long * dest, long val)
{
     myatomic64_add_return(dest, val);
}

long long GCC_NOREGPARM myatomic64_inc_return(volatile long long * dest)
{
     return myatomic64_add_return(dest, 1);
}
long long GCC_NOREGPARM myatomic64_dec_return(volatile long long * dest)
{
     return myatomic64_add_return(dest, -1);
}

long long GCC_NOREGPARM myatomic64_cmpxchg(volatile long long * dest, long long oldval, long long newval)
{
     return mycmpxchg64(dest, oldval, newval);
}

long long GCC_NOREGPARM myatomic64_xchg(volatile long long * dest, long long val)
{
     unsigned long long cmp;

     do {
          cmp = (unsigned long long) *dest;
     } while (cmp != mycmpxchg64((volatile unsigned long long *)dest, cmp, (unsigned long long)val));

     return cmp;
}


/*****************************************************************************
 * atomic_long
 ****************************************************************************/

void GCC_NOREGPARM my_atomic_long_inc(volatile long * dest)
{
#ifdef HAVE_KERNEL_ATOMIC_LONG
     atomic_long_inc((atomic_long_t *)dest);
#elif BITS_PER_LONG == 32
     myatomic32_add(dest, 1);
#elif BITS_PER_LONG == 64
     myatomic64_add(dest, 1);
#else
#    error "Oops, we can't define my_atomic_long_inc."
#endif
}

long GCC_NOREGPARM my_atomic_long_inc_return(volatile long * dest)
{
#ifdef HAVE_KERNEL_ATOMIC_LONG_RETURN
     return atomic_long_inc_return((atomic_long_t *)dest);
#elif BITS_PER_LONG == 32
     return myatomic32_inc_return(dest);
#elif BITS_PER_LONG == 64
     return myatomic64_inc_return(dest);
#else
#    error "Oops, we can't define my_atomic_long_inc_return."
#endif
}

void GCC_NOREGPARM my_atomic_long_add(volatile long * dest, long value)
{
#ifdef HAVE_KERNEL_ATOMIC_LONG
     atomic_long_add(value, (atomic_long_t *)dest);
#elif BITS_PER_LONG == 32
     myatomic32_add(dest, value);
#elif BITS_PER_LONG == 64
     myatomic64_add(dest, value);
#else
#    error "Oops, we can't define my_atomic_long_add."
#endif
}

long GCC_NOREGPARM my_atomic_long_add_return(volatile long * dest, long value)
{
#ifdef HAVE_KERNEL_ATOMIC_LONG_RETURN
     return atomic_long_add_return(value, (atomic_long_t *)dest);
#elif BITS_PER_LONG == 32
     return myatomic32_add_return(dest, value);
#elif BITS_PER_LONG == 64
     return myatomic64_add_return(dest, value);
#else
#    error "Oops, we can't define my_atomic_long_add_return."
#endif
}

void GCC_NOREGPARM my_atomic_long_dec(volatile long * dest)
{
#ifdef HAVE_KERNEL_ATOMIC_LONG
     atomic_long_dec((atomic_long_t *)dest);
#elif BITS_PER_LONG == 32
     myatomic32_add(dest, -1);
#elif BITS_PER_LONG == 64
     myatomic64_add(dest, -1);
#else
#    error "Oops, we can't define my_atomic_long_dec."
#endif
}

long GCC_NOREGPARM my_atomic_long_dec_return(volatile long * dest)
{
#ifdef HAVE_KERNEL_ATOMIC_LONG_RETURN
     return atomic_long_dec_return((atomic_long_t *)dest);
#elif BITS_PER_LONG == 32
     return myatomic32_dec_return(dest);
#elif BITS_PER_LONG == 64
     return myatomic64_dec_return(dest);
#else
#    error "Oops, we can't define my_atomic_long_dec_return."
#endif
}

void GCC_NOREGPARM my_atomic_long_sub(volatile long * dest, long value)
{
#ifdef HAVE_KERNEL_ATOMIC_LONG
     atomic_long_sub(value, (atomic_long_t *)dest);
#elif BITS_PER_LONG == 32
     myatomic32_add(dest, -value);
#elif BITS_PER_LONG == 64
     myatomic64_add(dest, -value);
#else
#    error "Oops, we can't define my_atomic_long_sub."
#endif
}

long GCC_NOREGPARM my_atomic_long_sub_return(volatile long * dest, long value)
{
#ifdef HAVE_KERNEL_ATOMIC_LONG_RETURN
     return atomic_long_sub_return(value, (atomic_long_t *)dest);
#elif BITS_PER_LONG == 32
     return myatomic32_add_return(dest, -value);
#elif BITS_PER_LONG == 64
     return myatomic64_add_return(dest, -value);
#else
#    error "Oops, we can't define my_atomic_long_sub_return."
#endif
}

long GCC_NOREGPARM my_atomic_long_xchg(volatile long * dest, long value)
{
#ifdef HAVE_KERNEL_ATOMIC_LONG_RETURN
     /* we define l for kernel headers BUG workaround */
#    define l dest
     return atomic_long_xchg((atomic_long_t *)dest, value);
#    undef l
#elif BITS_PER_LONG == 32
     return myatomic32_xchg(dest, value);
#elif BITS_PER_LONG == 64
     return myatomic64_xchg(dest, value);
#else
#    error "Oops, we can't define my_atomic_long_xchg."
#endif
}

long GCC_NOREGPARM my_atomic_long_cmpxchg(volatile long * dest, long oldval, long newval)
{
#ifdef HAVE_KERNEL_ATOMIC_LONG_RETURN
     return atomic_long_cmpxchg((atomic_long_t *)dest, oldval, newval);
#elif BITS_PER_LONG == 32
     return myatomic32_cmpxchg(dest, oldval, newval);
#elif BITS_PER_LONG == 64
     return myatomic64_cmpxchg(dest, oldval, newval);
#else
#    error "Oops, we can't define my_atomic_long_cmpxchg."
#endif
}


///*****************************************************************************/
/*
 * Recursive spinlock
 *
 * The idea behind this lock is the following:
 * you have some object, data or whatevere it is you
 * need to share between bottom half (softirq in our case)
 * or bottom halfs if SMP situation takes place and some
 * other process. Meanwhile this lock must not 'hang' if it
 * is called within the same logical context on the same CPU
 * (e.g. function uses this lock calls itself i.e. recursion
 * takes place). In case we do not need the recursion unconditional
 * lock can be used. Actually it works as read/write lock on
 * single CPU machine.
 *
 */

/*****************************************************************************/
void GCC_NOREGPARM mydisable_irq(int num)
{
  disable_irq_nosync(num);
}

void GCC_NOREGPARM myenable_irq(int num)
{
  enable_irq(num);
}

long int mygetpid(void)
{
  return current->tgid;
}

int mynum_online_cpus(void)
{
     return num_online_cpus();
}

int GCC_NOREGPARM myset_affinity(int cpu)
{
     return my__set_affinity(cpu);
}

unsigned long long myrdtsc(void)
{
// FIXME: this code isn't compiled for some kernels like 2.6.8...
/*
     uint32_t lo, hi;
     __asm__ __volatile__ (      // serialize
     "xorl %%eax,%%eax \n        cpuid"
     ::: "%rax", "%rbx", "%rcx", "%rdx");
   // We cannot use "=A", since this would use %rax on x86_64
     __asm__ __volatile__ ("rdtsc" : "=a" (lo), "=d" (hi));
//     __asm__ __volatile__ ("rdtsc" : "=a" (q));
//     return q;
      return (uint64_t)hi << 32 | lo;
*/

// @todo: not all platforms have tsc
  return 0;
}

void mydump_stack(void)
{
#ifdef DEBUG_SHOW_STACK
     myprintk("Show stack:\n");
     dump_stack();
#endif
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 19)
#define myhandle_sysrq(A) handle_sysrq(A, NULL, NULL)
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 36)
#define myhandle_sysrq(A) handle_sysrq(A, NULL)
#else
#define myhandle_sysrq(A) handle_sysrq(A)
#endif

void mydump_stack_all( void )
{
#ifdef DEBUG_SHOW_STACK
     myprintk("Show stack:\n");
     dump_stack();
#ifdef CONFIG_MAGIC_SYSRQ
     myprintk("Show Regs:");
     myhandle_sysrq('p');
     myprintk("Show backtrace of all active CPUs:");
     myhandle_sysrq('l');
     myprintk("Show State:");
     myhandle_sysrq('t');
     myprintk("Show Blocked State:");
     myhandle_sysrq('w');
#endif  /* CONFIG_MAGIC_SYSRQ */
#endif
}


// ------------------------------------------------------------------------------
typedef struct
{
	wait_queue_head_t wq_;

	int cond_;

}VnEvent;

VN_EVENT myevent_alloc( void )
{
	VnEvent* temp = (VnEvent* )mykmalloc( sizeof( *temp ) );

	if( temp )
	{
		init_waitqueue_head( &temp->wq_ );
		temp->cond_ = 0;
	}

	return ( VN_EVENT ) temp;
}

void GCC_NOREGPARM myevent_free( VN_EVENT event )
{
	if( event )
	{
		mykfree( event );
	}
}

int GCC_NOREGPARM myevent_wait(  VN_EVENT event )
{
	VnEvent* temp = ( VnEvent* ) event;

	if( !temp )
	{
		return -1;
	}

	{
		const int res = wait_event_interruptible( temp->wq_, temp->cond_ );

		if( 0 != res )
		{
			myprintk("wait event retrun error status %d.\n", res );
			return -2;
		}
	}

	return 0;
}

void GCC_NOREGPARM myevent_set(  VN_EVENT event )
{
	VnEvent* temp = ( VnEvent* ) event;

	if( temp )
	{
		temp->cond_ = 1;
		wake_up( &temp->wq_ );
	}
}

void GCC_NOREGPARM myevent_reset(  VN_EVENT event )
{
	VnEvent* temp = ( VnEvent* ) event;

	if( temp )
	{
		temp->cond_ = 0;
	}
}

// ------------------------------------------------------------------------------
// semaphore

VN_SEMAPHORE mysemaphore_alloc( void )
{
	 struct semaphore* sem = mykmalloc( sizeof(*sem ) );

	if( sem )
	{
		sema_init( sem, 0 );
	}

	return (VN_SEMAPHORE) sem;

}

void GCC_NOREGPARM mysemaphore_free( VN_SEMAPHORE sem )
{
	if( sem )
	{
		mykfree( sem );
	}
}

// semaphore actions
void GCC_NOREGPARM myup( VN_SEMAPHORE sem )
{
	struct semaphore* temp =  ( struct semaphore*) sem;

	up( temp );
}

void GCC_NOREGPARM mydown( VN_SEMAPHORE sem )
{
	struct semaphore* temp =  ( struct semaphore*) sem;

	down( temp );
}

#ifdef	CONFIG_LOCKMETER
void GCC_NOREGPARM _mymetered_enter(mymetered_info_t *lock_ptr, void *caller_pc)
{
    _metered_enter((metered_info_t *)lock_ptr, (void *)caller_pc);
}

void GCC_NOREGPARM _mymetered_leave(mymetered_info_t *lock_ptr)
{
    _metered_leave((metered_info_t *)lock_ptr);
}

void * GCC_NOREGPARM _current_address()
{
    return __builtin_return_address(0);
}
#endif	//CONFIG_LOCKMETER


// return "0" on success
int	GCC_NOREGPARM myis_task_present( const int taskPid )
{
	struct task_struct* tsk = myfind_task_by_pid( taskPid );

	if( NULL == tsk )
	{
		return -1;
	}

	myput_task_struct_lite(tsk);

	return 0;
}


#include "struct_access.c"

static void GCC_NOREGPARM (*_netdevice_event_handler)(unsigned long, VN_DEVICE);
static void GCC_NOREGPARM (*_inetaddr_event_handler)(unsigned long, VN_INIFADDR);

static int iplir_netdevice_event_handler(struct notifier_block *nb,
          unsigned long event, void *ptr)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 11, 0)
     _netdevice_event_handler(event, netdev_notifier_info_to_dev(ptr));
#else //LINUX_VERSION_CODE >= KERNEL_VERSION(3, 11, 0)
     _netdevice_event_handler(event, ptr);
#endif //LINUX_VERSION_CODE >= KERNEL_VERSION(3, 11, 0)
     return NOTIFY_DONE;
}

static int iplir_inet_event_handler(struct notifier_block *nb,
          unsigned long event, void *ptr)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 11, 0)
     // Here ptr is 'struct in_ifaddr *'
     _inetaddr_event_handler(event, ptr);
#else //LINUX_VERSION_CODE >= KERNEL_VERSION(3, 11, 0)
     _inetaddr_event_handler(event, ptr);
#endif //LINUX_VERSION_CODE >= KERNEL_VERSION(3, 11, 0)
     return NOTIFY_OK;
}

static struct notifier_block iplir_netdev_nb = {
     .notifier_call = iplir_netdevice_event_handler,
};

static struct notifier_block iplir_inet_nb = {
     .notifier_call = iplir_inet_event_handler,
};

int GCC_NOREGPARM myregister_netdevice_notifier(void GCC_NOREGPARM (*handler)(unsigned long, VN_DEVICE))
{
     _netdevice_event_handler = handler;
     return register_netdevice_notifier(&iplir_netdev_nb);
}

int GCC_NOREGPARM myregister_inetaddr_notifier( void (*handler)( unsigned long, VN_INIFADDR ) )
{
     _inetaddr_event_handler = handler;
     return register_inetaddr_notifier (&iplir_inet_nb);
}

void myunregister_netdevice_notifier(void)
{
     unregister_netdevice_notifier(&iplir_netdev_nb);
}

void myunregister_inetaddr_notifier(void)
{
     unregister_inetaddr_notifier(&iplir_inet_nb);
}

unsigned long myNETDEV_UNREGISTER(void)
{
     return NETDEV_UNREGISTER;
}

unsigned long myNETDEV_REGISTER(void)
{
     return NETDEV_REGISTER;
}

unsigned long myNETDEV_CHANGENAME(void)
{
     return NETDEV_CHANGENAME;
}

unsigned long myNETDEV_UP(void)
{
     return NETDEV_UP;
}

unsigned long myNETDEV_DOWN(void)
{
     return NETDEV_DOWN;
}

void myset_normal_scheduler(void)
{
     struct sched_param param = { .sched_priority = 0 };
     long pid = mygetpid();
     struct task_struct * tsk = myfind_task_by_pid( pid );
     if(tsk) {
          set_user_nice( tsk, 0 );
          mysched_setscheduler( tsk, SCHED_NORMAL, &param );
          myput_task_struct_lite(tsk);
     }
}

void mythread_yield(void)
{
     long pid = mygetpid();
     struct task_struct * tsk = myfind_task_by_pid( pid );
     if(tsk) 
     {
          set_user_nice( tsk, 0 );
          {
               struct sched_param param = { .sched_priority = 0 };
               mysched_setscheduler( tsk, SCHED_NORMAL, &param );
          }

          /*release ownership*/
          myschedule(); /*TODO: Can we hold task and release ownership at same time? Is it safe from view of task's destroying process? */

          /*restore scheduler class*/
          set_user_nice( tsk, -20 );
          {
               struct sched_param param = { .sched_priority = MAX_RT_PRIO - 2 };
               mysched_setscheduler( tsk, SCHED_FIFO, &param );
          }

          myput_task_struct_lite(tsk);
     }
}

VN_ADDR6 GCC_NOREGPARM my_flow_saddr(const VN_FLOW fl, unsigned short family)
{
#if defined (KERN_IPCLS)
         switch (family){
         case AF_INET:
                 return (VN_ADDR6)(&((struct flowi *)fl)->u.ip4.saddr);
         case AF_INET6:
                 return (VN_ADDR6)(&((struct flowi *)fl)->u.ip6.saddr);
         }
#endif
         return NULL;
}

VN_ADDR6 GCC_NOREGPARM my_flow_daddr(const VN_FLOW fl, unsigned short family)
{
#if defined (KERN_IPCLS)
         switch (family){
         case AF_INET:
                 return (VN_ADDR6)&((struct flowi *)fl)->u.ip4.daddr;
         case AF_INET6:
                 return (VN_ADDR6)&((struct flowi *)fl)->u.ip6.daddr;
         }
#endif
         return NULL;
}

static inline int
myaddr_match(const void *token1, const void *token2, int prefixlen)
{
	const u32 *a1 = token1;
	const u32 *a2 = token2;
	int pdw;
	int pbi;

	pdw = prefixlen >> 5;     /* num of whole u32 in prefix */
	pbi = prefixlen &  0x1f;  /* num of bits in incomplete u32 in prefix */

	if (pdw)
		if (mymemcmp(a1, a2, pdw << 2))
			 return 0;

	if (pbi) {
		uint mask;

		mask = htonl((0xffffffff) << (32 - pbi));

		if ((a1[pdw] ^ a2[pdw]) & mask)
			return 0;
	}

	return 1;
}

#if defined (IFACE_ITCSIPCLS) // bad.. because of flow_selector definition scope

int GCC_NOREGPARM my_selector_match(const struct flow_selector *sel, VN_FLOW flow,
	unsigned short family)
{
#if defined (KERN_IPCLS)
	const struct flowi *fl = (struct flowi *)flow;
	const struct flowi4 *fl4;
	const struct flowi6 *fl6;

	switch (family) {
	case AF_INET:

		fl4 = &fl->u.ip4;

		return  addr4_match(fl4->daddr, sel->daddr.a4, sel->prefixlen_d) &&
			addr4_match(fl4->saddr, sel->saddr.a4, sel->prefixlen_s) &&
			!((xfrm_flowi_dport(fl, &fl4->uli) ^ sel->dport) & sel->dport_mask) &&
			!((xfrm_flowi_sport(fl, &fl4->uli) ^ sel->sport) & sel->sport_mask) &&
			(fl4->flowi4_proto == sel->proto || !sel->proto);

	case AF_INET6:

		fl6 = &fl->u.ip6;

		return myaddr_match(&fl6->daddr, &sel->daddr, sel->prefixlen_d) &&
			myaddr_match(&fl6->saddr, &sel->saddr, sel->prefixlen_s) &&
			!((xfrm_flowi_dport(fl, &fl6->uli) ^ sel->dport) & sel->dport_mask) &&
			!((xfrm_flowi_sport(fl, &fl6->uli) ^ sel->sport) & sel->sport_mask) &&
			(fl6->flowi6_proto == sel->proto || !sel->proto);
	}
#endif
	return 0;
}
#endif // #if defined (IFACE_ITCSIPCLS)

#define FMT32LEN	8

#define fmt32(buf, be32)\
	(mysprintf(buf, "%02x%02x%02x%02x",\
		 (be32) & 0xFF,\
		((be32) >> 8) & 0xFF,\
		((be32) >> 16) & 0xFF,\
		((be32) >> 24) & 0xFF))

void GCC_NOREGPARM my_print_ip4_addr(unsigned long be_addr)
{
	unsigned char buf4[FMT32LEN+1];
	fmt32(buf4, be_addr);
	buf4[FMT32LEN] = '\0';
	myprintk("0x%8s\n", buf4);
}

void GCC_NOREGPARM my_print_ip6_addr(VN_ADDR6 addr)
{
	unsigned char i, off;
	unsigned char buf6[4*FMT32LEN+1];
	for(i = 0, off = 0; i < 4; i++, off+=FMT32LEN)
		fmt32(buf6+off, ((xfrm_address_t*)addr)->a6[i]);
	buf6[4*FMT32LEN] = '\0';
	myprintk("0x%32s\n", buf6);
}


#ifdef USE_FRWL_STATS

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,33)
struct frwl_counters
{
	struct frwl_stat_count   __percpu *stats;
	struct xnat_error_count  __percpu *errors;
};
#else
struct frwl_counters
{
    struct frwl_stat_count   *stats;
    struct xnat_error_count  *errors;
};
#endif

static struct frwl_counters frwl_ctr = {
		.stats  = NULL,
		.errors = NULL
};

// '1' if ON, '0' if OFF
static volatile unsigned int frwl_ctr_on = 0;

// return '0' on success
int GCC_NOREGPARM myfrwl_init_counters(void)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,33)
	frwl_ctr.stats = myalloc_percpu( struct frwl_stat_count );
	frwl_ctr.errors = myalloc_percpu( struct xnat_error_count );
#endif
    
	return (
		frwl_ctr.stats && frwl_ctr.errors
	) ;
}

void GCC_NOREGPARM myfrwl_flush_counters(void)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,33)
  
	struct frwl_stat_count *stats;
	struct xnat_error_count *errors;
	unsigned int cpu, field;

	for_each_possible_cpu(cpu)
	{
		stats = (struct frwl_stat_count *)per_cpu_ptr( frwl_ctr.stats, cpu );

		// memset?
		for (field = 0; field < _STAT_MAX; field++)
			stats->counter[field] = 0;

		errors = (struct xnat_error_count *)per_cpu_ptr( frwl_ctr.errors, cpu );

		for (field = 0; field < _ERROR_MAX; field++)
			errors->counter[field] = 0;
	}

	myprintk( "myfrwl_flush_counters done \n" );
#endif
}

void GCC_NOREGPARM myfrwl_toggle_counters(int on)
{
	frwl_ctr_on = on;

	myprintk( "myfrwl_toggle_counters to: %u \n", on );
}

void GCC_NOREGPARM myfrwl_fini_counters(void)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,33)
	if (frwl_ctr.stats)
		myfree_percpu(frwl_ctr.stats);

	if (frwl_ctr.errors)
		myfree_percpu(frwl_ctr.errors);
#endif    
}

void GCC_NOREGPARM myfrwl_inc_stat_counter(unsigned int field)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,33)
	if (frwl_ctr_on)
		this_cpu_inc( *((unsigned int*)frwl_ctr.stats + field) );
#endif    
}

void GCC_NOREGPARM myfrwl_inc_error_counter(unsigned int field)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,33)
	if (frwl_ctr_on)
		this_cpu_inc( *((unsigned int*)frwl_ctr.errors + field) );
#endif
}

/* order is less important here */
static const char* frwl_stat_labels[_STAT_MAX] =
{
	[STAT_ALG_PUSHED] 			= "alg pushed",
	[STAT_ALG_PULLED] 			= "alg pulled",
	[STAT_WAIT_RULE_INS]		= "wait rule inserted",
	[STAT_WAIT_RULE_FOUND]		= "wait rule found",
	[STAT_WAIT_RULE_PROCESSED]	= "wait rule processed",
	[STAT_WAIT_RULE_DEL]		= "wait rule removed",
	[STAT_ROUTE_FOUND]			= "static route found and processed",
	[STAT_NAT_RULE_FOUND]		= "static nat rule found and processed",
	[STAT_OPEN_RULE_FOUND]		= "open stat rule found and processed",
	[STAT_PROT_RULE_FOUND]		= "protected stat rule found and processed",
	[STAT_PROT_ENTER]			= "protected packet enter frwl",
	[STAT_OPEN_ENTER]			= "open packet enter frwl",
	[STAT_PROT_OR_OPEN_LEAVE]	= "protected/open packet leave frwl",
};

void GCC_NOREGPARM myfrwl_stats_proc_show(VN_SEQ m)
{
	struct seq_file *s = (struct seq_file *)m;

	struct frwl_stat_count *src, dst;
	unsigned int i, field, cnt;
	unsigned int *ptr = &dst;

	memset(&dst, 0, sizeof(struct frwl_stat_count));
    
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,33)
	for_each_possible_cpu(i)
	{
		src = (struct frwl_stat_count *)per_cpu_ptr( frwl_ctr.stats, i );

		for (field = 0; field < _STAT_MAX; field++)
			dst.counter[field] += src->counter[field];
	}
#endif

	for (field = 0; field < _STAT_MAX; field++)
	{
		seq_printf( s, " %-45s %d \n", frwl_stat_labels[field], dst.counter[field] );
	}
	return 0;
}


/* order is less important here */
static const char* frwl_error_labels[_ERROR_MAX] =
{
	[ERROR_DYN_RULE_NOT_FOUND_DROP]	= "dyn.rule not found thus drop",
	[ERROR_DROP]					= "drop",
	[ERROR_UNKNOWN_DROP]			= "drop due to unknw. reason",
	[ERROR_PROCESS_WAIT_RULE_DROP]	= "failed while processing waitrule",
	[ERROR_ROUTE_FAILED]			= "no route or failed to process route",
	[ERROR_NAT_FAILED]				= "failed to process stat. nat rule",
	[ERROR_OPEN_FILTER_FAILED]		= "open: no static rule or failed to process it",
	[ERROR_PROT_FILTER_FAILED]		= "prot: no static rule or failed to process it",
	[ERROR_CONS_LIMIT_REACHED]		= "cons limit reached",
	[ERROR_ALG_ROUTE_FAILED]		= "failed to pre-calc route for alg-ed packet",
};

void GCC_NOREGPARM myfrwl_errors_proc_show(VN_SEQ m)
{
	struct seq_file *s = (struct seq_file *)m;

	struct xnat_error_count *src, dst;
	unsigned int i, field, cnt;
	unsigned int *ptr = &dst;

	memset(&dst, 0, sizeof(struct xnat_error_count));
    
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,33)
	for_each_possible_cpu(i)
	{
		src = (struct xnat_error_count *)per_cpu_ptr( frwl_ctr.errors, i );

		for (field = 0; field < _ERROR_MAX; field++)
			dst.counter[field] += src->counter[field];
	}
#endif

	for (field = 0; field < _ERROR_MAX; field++)
	{
		seq_printf( s, " %-45s %d \n", frwl_error_labels[field], dst.counter[field] );
	}
	return 0;
}

#endif // #ifdef USE_FRWL_STATS


#if defined(CONFIG_PPC32) || defined(CONFIG_PPC64)
void * __memcpy_chk(void *dest, const void *src, size_t n)
{
	return memcpy(dest, src, n);
}

void * __memmove_chk(void *dest, const void *src, size_t n)
{
	return memmove(dest, src, n);
}
/*
long int strtol(const char *nptr, char **endptr, int base)
{
	return simple_strtol(nptr, endptr, base);
}
*/
#endif /* defined(CONFIG_PPC32) || defined(CONFIG_PPC64) */

#if defined ( CONFIG_IPLIR_XFRM_INPUT ) || defined ( IPLIR_XFRM_INPUT )
void GCC_NOREGPARM mynf_reset(VN_SKB skb)
{
          nf_reset ((struct sk_buff *)skb);
}

void GCC_NOREGPARM myskb_set_mac_header(VN_SKB skb, const int offset)
{
         skb_set_mac_header((struct sk_buff*) skb, offset);
}

void GCC_NOREGPARM myskb_reset_network_header(VN_SKB skb)
{
         skb_reset_network_header((struct sk_buff*) skb); 
}

VN_DST GCC_NOREGPARM myskb_dst_pop(VN_SKB skb)
{
          return skb_dst_pop((struct sk_buff *)skb);
}

void GCC_NOREGPARM myskb_dst_set_noref(VN_SKB skb, VN_DST dst)
{
          return skb_dst_set_noref((struct sk_buff *)skb, (struct dst_entry *)dst);
}

// Perform policy lookup for recv packet at channel layer
// return '1' if there ipsec policy, '0' otherwise
int GCC_NOREGPARM my_xfrm_lookup_fwd(VN_SKB skb)
{
	/* reworked version of nf_xfrm_me_harder */

	int err, ret;
	struct flowi fl;
	struct dst_entry *dst, *dst_new;
	struct sk_buff *skbuff = (struct sk_buff*)skb;
     struct net *net = dev_net(skbuff->dev);
	const struct iphdr *iph;

     if (!net->xfrm.policy_count[XFRM_POLICY_OUT])
          return 0;

	err = xfrm_decode_session( skbuff, &fl, AF_INET );
	if (err < 0)
		return 0;

	dst = skb_dst( skbuff );

	/* assuming no dst at rcv_in */
	if (likely(dst == NULL))
	{
		iph = ip_hdr(skb);

		ret = ip_route_input_noref(skb, iph->daddr, iph->saddr, iph->tos, skbuff->dev);

		if (unlikely(ret))
			return 0;

		dst = skb_dst( skbuff );
	}

     if (dst->flags & DST_NOXFRM)
     {
          ret = 0;
          goto dropdst;
     }

     skb_dst_force(skb);

#ifdef CONFIG_IPLIR_XFRM_OUT
	dst_new = xfrm_lookup_real( dev_net(dst->dev), dst, &fl, skbuff->sk, 0 );
#else
	dst_new = xfrm_lookup( dev_net(dst->dev), dst, &fl, skbuff->sk, 0 );
#endif

    if (IS_ERR(dst_new))
    {
    	ret = 1;
    	goto dropdst;
    }
    	
	ret = (dst_new &&
	     dst_new->xfrm && /* some transformation... */
	     dst_new->xfrm != IPLIR_NF_XFRM_MARK /* .. and it's not ours */);

     skb_dst_set(skb, dst_new);

dropdst: /* drop old dst */
	skb_dst_drop(skb);

	return ret;
}
#endif

#if PATCH_METHOD_ENABLED == 1

#define skb_dev(x)		(*(myskb_fetch_dev(x)))
#define skb_cb(x)		(*(myskb_fetch_cb(x)))
#define skb_protocol(x)  (myskb_fetch_protocol(x))
#define skb_pkt_type(x)         (skbuff_get(x, pkt_type))
#define neighbour_type(x)	(myneighbour_fetch_type(x))
#define dst_neighbour(x)        (mydst_fetch_neighbour(x))
#define dev_type(x)		(*(mydev_fetch_type(x)))
#define dev_flags(x)		(mydev_fetch_flags(x))

#define PASS_SKB( skb ) return skb
#define DROP_SKB( skb ) 	{ 	mydev_kfree_skb(skb); \
					  	return NULL; \
              			}
static int is_broadcast( struct sk_buff* skb )
{
	VN_DEVICE const dev = skb_dev( skb ) ;

	if ((skb_protocol(skb) == htons(myETH_P_IP())) || (skb_protocol(skb) == htons(myETH_P_IPV6())))
	{
		if (dev_type(dev) != myARPHRD_ETHER())
			return 0;

		if (skb_pkt_type(skb) == myPACKET_BROADCAST())
		{
			udebug3(("Is_Broadcast: pkt_type = BROADCAST"));
			return 1;
		}

		{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,31)
               struct rtable *rt = skb_rtable( skb );
#else
               void *rt = kernel_skb_dst( skb );
#endif
               if ( !rt ) 
               {
                    return 0;
               }
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,6,0)
               else
               {
                    return rt->rt_type == RTN_BROADCAST;
               }
#endif
               if ( neighbour_type( skb ) == myRTN_BROADCAST() )
                    return 1;
		}
		/* Packet has dst and neighbour, but the neighbour is not broadcast. */
		return 0;
	} /* IP handling */

	if (skb_protocol(skb) == htons(myETH_P_ARP()))
		return 1;

	return 0;
}

static struct sk_buff* GCC_NOREGPARM default_in_pf(struct sk_buff *skb)
{
	VN_DEVICE const dev = skb_dev(skb);

	/* Check for loopback interface - we don't interfere with packets
	** going through loopback interface.
	*/
	if (dev_flags(dev) & myIFF_LOOPBACK())
	{
		PASS_SKB(skb);
	}

	if (skb_cb(skb) == RECV_FLAG)
	{
		skb_cb(skb) = 0;
		PASS_SKB(skb);
	}

	{
		const int devtype  = dev_type(dev);
		if ( devtype != myARPHRD_ETHER() && !IsSupportedOtherMediumType( devtype ) )
		{
			PASS_SKB(skb);
		}
	}

	{
		int32_t const header_offset = mydev_fetch_header_len( dev );
		if( header_offset != 0 )
		{
			myskb_push( skb, header_offset );
		}
		myskb_set_network_header( skb, header_offset );
	}

	{
		const int flags = ( is_broadcast( skb )? FM_BROADCAST: 0 ) + ( myis_arp_protocol( skb )? FM_URGENT: 0 );
#ifdef DRV_TRACE_RUNTIME
        if ( trace_mask & TRACE_BIT_INTO_DRV )
        {
             MYDUMP_SKB_IPHDR( "default_in_pf", skb);
        }
        if ( trace_mask & TRACE_BIT_IN_SKB )
        {
             mydump_skb(skb);
        }        
#endif
		return fast_drv_in( skb, flags );
	}
}

static struct sk_buff* GCC_NOREGPARM default_out_pf(struct sk_buff *skb)
{
	VN_DEVICE const dev = skb_dev(skb);
	/* Check for loopback interface - we don't interfere with packets
	** going through loopback interface. Now we do it in much better manner.
	*/

	const int devtype  = dev_type(dev);
	if ( devtype != myARPHRD_ETHER() && !IsSupportedOtherMediumType( devtype ) )
	{
		PASS_SKB(skb);
	}

	if (dev_flags(dev) & myIFF_LOOPBACK())
	{
		PASS_SKB(skb);
	}

	if (skb_cb(skb) == SEND_FLAG)
	{
		PASS_SKB(skb);
	}

	{
		const int flags = ( is_broadcast( skb )? FM_BROADCAST: 0 + FM_URGENT );
#ifdef DRV_TRACE_RUNTIME
     if ( trace_mask & TRACE_BIT_INTO_DRV )
     {
          MYDUMP_SKB_IPHDR( "default_out_pf", skb);
     }
     if ( trace_mask & TRACE_BIT_IN_SKB )
     {
          mydump_skb(skb);
     }     
#endif
		return fast_drv_out( skb, flags );
	}
}
#endif //PATCH_METHOD_ENABLED
