#ifndef __LINUX_KRNIFACE_H__
#define __LINUX_KRNIFACE_H__

/* Same trick as the one in userland ... --vel */
#ifdef UDEBUG
#  define udebug(x) myprintkn x
#  define udebug_dump(x) mydumphex x
#  if UDEBUG > 0
#    define udebug1(x) myprintkn x
#    define udebug_dump1(x) mydumphex x
#  else
#    define udebug1(x)
#    define udebug1_dump1(x)
#  endif
#  if UDEBUG > 1
#    define udebug2(x) myprintkn x
#    define udebug_dump2(x) mydumphex x
#  else
#    define udebug2(x)
#    define udebug_dump2(x)
#  endif
#  if UDEBUG > 2
#    define udebug3(x) myprintkn x
#    define udebug_dump3(x) mydumphex x
#  else
#    define udebug3(x)
#    define udebug_dump3(x)
#  endif
#  if UDEBUG > 3
#    define udebug4(x) myprintkn x
#    define udebug_dump4(x) mydumphex x
#  else
#    define udebug4(x)
#    define udebug_dump4(x)
#  endif
#  define udebug_dump_skb(x) mydump_skb x
#  undef udebug_dump
#  undef udebug_dump1
#  undef udebug_dump2
#  undef udebug_dump3
#  undef udebug_dump4
#  define udebug_dump(x)
#  define udebug_dump1(x)
#  define udebug_dump2(x)
#  define udebug_dump3(x)
#  define udebug_dump4(x)
#else
#  define udebug(x)
#  define udebug_dump(x)
#  define udebug1(x)
#  define udebug_dump1(x)
#  define udebug2(x)
#  define udebug_dump2(x)
#  define udebug3(x)
#  define udebug_dump3(x)
#  define udebug4(x)
#  define udebug_dump4(x)
#  define udebug_dump_skb(x)
#endif

// trace_mask is a number of bits. 
// Each of them means some type(point) of tarce.
#ifdef  DRV_TRACE_RUNTIME
// trace headers packets are going in to driver
#define TRACE_BIT_INTO_DRV 1
// trace headers packets are going out of driver
#define TRACE_BIT_OUTOF_DRV 2
// dump skb packets are going in to driver
#define TRACE_BIT_IN_SKB 4
// dump skb packets are going out of driver
#define TRACE_BIT_OUT_SKB 8
#endif

#ifndef USE_FRWL_STATS
#define USE_FRWL_STATS
#endif

#define MYDUMP_SKB_IPHDR(label,skb) mydump_skb_iphdr(label, __func__, __LINE__, skb)
#define MYDUMP_NFINFO(n) mydump_nfinfo(__func__, __LINE__, n)

/*
** NOTE about use register arguments in function calls.
** Some linux 2.6 kernels has CONFIG_REGPARM configuration parameter.
** When this parameter set kernel is compliled with -mregparm=3 gcc option.
** This uses an different "application binary interface" (ABI) and passes the first
** three arguments of a function call in registers. This feature is only enabled for
** gcc-3.0 and later - earlier compilers generate incorrect output with certain
** kernel constructs when -mregparm=3 is used. Most linux distributions based on 2.6
** kernels (like SuSE, Fedora Core etc.) use this parameter by default. This will break
** binary modules comilped without this option. To solve this problem we must use special
** gcc function call attribute in kernel interface functions, because these routines
** is called from kernel-independent precompiled part of driver.
** In case when our precompiled functions registers and calls from kernel structures
** (such as kernel timers, threads, interrupt handlers etc.) we use functions stubs or
** universal handler mechanism, which allow to register abstract handler instead of
** real handler. When abstract handler is called it call real handler with proper ABI.
*/
#ifndef GCC_NOREGPARM
#if defined(LINUX_VERSION_CODE)
#if defined(__arm__)
      // modern/risc arch. always use registers in function calls
#     define GCC_NOREGPARM
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)) || \
     ((LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,19)) && (defined(CONFIG_REGPARM) && ( __GNUC__ >= 3 )))
#     define GCC_NOREGPARM __attribute__((regparm(0)))
#     define KERNEL_USE_REGPARM
#else
#     define GCC_NOREGPARM
#endif
#else  // !defined LINUX_VERSION_CODE
#     define GCC_NOREGPARM
#endif
#endif // define GCC_NOREGPARM

#ifdef KERNEL_USE_REGPARM
struct _univ_handle
{
  int GCC_NOREGPARM (* fn)(void *);
  void * arg;
  unsigned char flclean;
};
#endif




/* milestone points in theirs traversal order */
enum
{
	/* connection redirected to alg */
	STAT_ALG_PUSHED,

	/* alg accepted newly redirected connection */
	STAT_ALG_PULLED,

	/* created wait rule */
	STAT_WAIT_RULE_INS,

	/* wait rule found */
	STAT_WAIT_RULE_FOUND,

	/* found wait rule successfully processed */
	STAT_WAIT_RULE_PROCESSED,

	/* deleted either when found --or-- by timeout */
	STAT_WAIT_RULE_DEL,

	/* static route found and successfuly processed */
	STAT_ROUTE_FOUND,

	/* static nat rule found and successfuly processed */
	STAT_NAT_RULE_FOUND,

	/* found --and-- processed static rule */
	STAT_OPEN_RULE_FOUND,
	STAT_PROT_RULE_FOUND,

	/* open packet entered frwl */
	STAT_PROT_ENTER,

	/* protected packet entered firewall */
	STAT_OPEN_ENTER,

	/* protected --or-- open packet successfuly passed frwl */
	STAT_PROT_OR_OPEN_LEAVE,

	/* unused */
	_STAT_MAX
};

enum
{
	ERROR_DYN_RULE_NOT_FOUND_DROP,
	ERROR_DROP,
	ERROR_UNKNOWN_DROP,

	/* error while processing wait rule */
	ERROR_PROCESS_WAIT_RULE_DROP,

	/* we failed while processing static route --or-- no route found */
	ERROR_ROUTE_FAILED,

	/* we failed while processing _found_ static nat rule */
	ERROR_NAT_FAILED,

	/* no static rule --or-- we failed to process it */
	ERROR_OPEN_FILTER_FAILED,
	ERROR_PROT_FILTER_FAILED,

	ERROR_CONS_LIMIT_REACHED,

	ERROR_ALG_ROUTE_FAILED,

	/* unused */
	_ERROR_MAX
};


/** @def NOREF
 * Keeps the compiler from bitching about an unused parameter.
 */
#define NOREF(var)               (void)(var)
#define CONFIG_ITCSDRV_AS_MISC

struct myethhdr
{
	unsigned char	h_dest[6];
	unsigned char	h_source[6];
	unsigned short	h_proto;
} __attribute__((packed));

struct myarphdr
{
	unsigned short 	ar_hrd;
	unsigned short 	ar_pro;
	unsigned char 	ar_hln;
	unsigned char 	ar_pln;
	unsigned short 	ar_op;
} __attribute__((packed));

/* We have some typedefs here just for evidence. */
typedef void * VN_TIMER;
typedef void * VN_DEVICE;
typedef void * VN_SKB;
typedef void * VN_SPINLOCK;
typedef void * VN_PCIDEV;
typedef void * VN_NEIGHBOUR;
typedef void * VN_NEIGHTABLE;
typedef void * VN_MUTEX;
typedef void * VN_WAITQUEUE;
typedef void * VN_TQSTRUCT;
typedef void * VN_TASKQUEUE;
typedef void * VN_PTREGS;
typedef void * VN_INIFADDR;
typedef void * VN_INDEVICE;
typedef void * VN_STINODE;
typedef void * VN_STFILE;
typedef void * VN_FILEOPERATIONS;
typedef void * VN_NFQFUNC;
typedef void * VN_NFINFO;
typedef void * VN_DST;
typedef void * VN_ATOMIC;
typedef void * VN_RWLOCK;
typedef void * VN_KMEM_CACHE;
typedef void * VN_TASK_STRUCT;
typedef void * VN_PROC_DIR_ENTRY;
typedef void * VN_NET;
typedef void * VN_FLOWI;
typedef void * VN_FL4;
typedef void * VN_DEV;
typedef void * VN_SK;
typedef void * VN_WORKSTRUCT;
typedef void * VN_FLOW;
typedef void * VN_FLOW_CACHE_OBJECT;
typedef void * VN_INET_PEER;
typedef void * VN_ADDR6;
typedef void * VN_SEQ;

/* Dummy even address, for nf passing */
#define IPLIR_NF_XFRM_MARK         0x111111

//for debug
#ifdef DEBUG_F1024
typedef void * VN_SIGINFO_T;
#endif


#ifdef USE_FRWL_STATS

/* __percpu */
struct frwl_stat_count
{
	unsigned int counter[_STAT_MAX];
};

/* __percpu */
struct xnat_error_count
{
	unsigned int counter[_ERROR_MAX];
};

int GCC_NOREGPARM myfrwl_init_counters(void);
void GCC_NOREGPARM myfrwl_flush_counters(void);
void GCC_NOREGPARM myfrwl_toggle_counters(int on);
void GCC_NOREGPARM myfrwl_fini_counters(void);

void GCC_NOREGPARM myfrwl_inc_stat_counter(unsigned int stat_field);
void GCC_NOREGPARM myfrwl_inc_error_counter(unsigned int err_field);

void GCC_NOREGPARM myfrwl_stats_proc_show(VN_SEQ m);
void GCC_NOREGPARM myfrwl_errors_proc_show(VN_SEQ m);

#define FRWL_INC_STAT(field)		myfrwl_inc_stat_counter( field )
#define FRWL_INC_ERROR(field)		myfrwl_inc_error_counter( field )

#else // #ifdef USE_FRWL_STATS

#define FRWL_INC_STAT(field)
#define FRWL_INC_ERROR(field)

#endif // #ifndef USE_FRWL_STATS


typedef struct rec_spinlock {
	char rec_locks[16];
	VN_SPINLOCK lock;
} rec_spinlock_t;

void myunconditional_rec_spinlock_lock(rec_spinlock_t *);
void myunconditional_rec_spinlock_unlock(rec_spinlock_t *);
rec_spinlock_t * myrec_spinlock_init(void);
void myrec_spinlock_free(rec_spinlock_t *);

/// @brief Obtain a pointer to the method or variable exported by kernel or another module
/// @param[in] method name
void* GCC_NOREGPARM mysymbol_get( const char *symbol );
/// @brief drop the reference to the kernel symbol (and a possessing module)
/// @param[in] symbol name
void GCC_NOREGPARM mysymbol_put( const char *symbol );
void mycli(void);
void mysti(void);
void mymemleaktest(void);
void mymemleaktest_init(void);
void mymemleaktest_clean(void);
void * GCC_NOREGPARM mymalloc(unsigned int size);
void GCC_NOREGPARM myfree(void * obj);
void * GCC_NOREGPARM mykmalloc(unsigned int size);
void GCC_NOREGPARM mykfree(void * obj);
void * GCC_NOREGPARM myvmalloc(unsigned int size);
void GCC_NOREGPARM myvfree(void * obj);
int myprintk(const char * fmt, ...);
int myprintkn(const char * fmt, ...);
char* GCC_NOREGPARM drv_strcpy( char* dest, const char* src );
unsigned int GCC_NOREGPARM drv_strlen( const char* s );
void GCC_NOREGPARM mydumphex(char * s, int len);
int mysprintf(char * buf, const char * fmt, ...);
int mysnprintf(char * buf, unsigned int size, const char * fmt, ...);
int GCC_NOREGPARM myset_packet_filter(void *, void *);

/// \brief Calls l2_condfrag_filter (from the kernel driver) if it is initialized,
///        when we need to fragmentate packet before injecting it into the stack
int GCC_NOREGPARM mycondfrag_filter(VN_SKB, VN_DEVICE, int);
int GCC_NOREGPARM mynetif_rx(VN_SKB);
int GCC_NOREGPARM mynetif_rx_slow(VN_SKB);
int GCC_NOREGPARM mydev_queue_xmit(VN_SKB);
VN_DEVICE GCC_NOREGPARM mydev_get(const char *);
void GCC_NOREGPARM mydev_put(VN_DEVICE dev);
void GCC_NOREGPARM mydev_hold(VN_DEVICE dev);
/// @brief Obtain a reference amount, associated with device dev
int GCC_NOREGPARM mynetdev_refcnt( VN_DEVICE dev );
/// @brief Force set device reference counter to null
void GCC_NOREGPARM mydev_drop_allrefs( VN_DEVICE dev );
/// @brief Get IPv4 information, associated with net device
/// @return Pointer to in_ifaddr structure if succeeds, null pointer otherwise
VN_INIFADDR GCC_NOREGPARM mydev_get_ifaddr( VN_DEVICE dev );
VN_INIFADDR GCC_NOREGPARM mydev_get_next_ifaddr( VN_INIFADDR prev );

/// @brief Mark entirely ipsec flow cache outdated, forcing to resolve per-flow transformations
void mymark_flow_cache_outdated(void);

void GCC_NOREGPARM myinet_putpeer(VN_INET_PEER peer);

int GCC_NOREGPARM my__ip6_local_out(VN_SKB vnskb);

#if defined ( CONFIG_IPLIR_XFRM_INPUT ) || defined ( IPLIR_XFRM_INPUT )
/// \brief Resets Netfilter states for skb
void GCC_NOREGPARM mynf_reset(VN_SKB skb);
/// \brief Set mac header relative to 'data' buffer
void GCC_NOREGPARM myskb_set_mac_header(VN_SKB skb, const int offset);
/// \brief Sets network header field of skb to start of 'data' buffer
void GCC_NOREGPARM myskb_reset_network_header(VN_SKB skb);
/// \brief Drops current dst and returns its child
VN_DST GCC_NOREGPARM myskb_dst_pop(VN_SKB skb);
/// \brief Associate skb with dst, but do not update reference counter
void GCC_NOREGPARM myskb_dst_set_noref(VN_SKB skb, VN_DST dst);

/// \brief   IPsec     forward .
///
///  '1'   ,  '0'
///
int GCC_NOREGPARM my_xfrm_lookup_fwd(VN_SKB skb);
#endif
/*
** stubs to fill file_operations struct while charaster device register
** (see comments about CONFIG_REGPARM).
** DON'T use 'GCC_NOREGPARM' in these functions.
*/
struct inode;
struct file;

/// \brief    procfs .
///
///      Linux  
///     procfs .
///          procfs.
///
/// \param[out] page  page    ,     ,
///               .  page,       
///             . ,         copy_to_user.
/// \param[in] count     ,    
///               .
/// \param[in] start        (, 4KB), 
///               start  off.
/// \param[in] off        (, 4KB), 
///               start  off.
/// \param[out] eof  ,    ,   eof ( ).
/// \param[in] data       .
/// \return     
int mydrviplir_proc_read(char *page, char **start, long off, int count, int *eof, void *data);

/// \brief    procfs .
///
///      Linux  
///     procfs .
///          procfs.
///
/// \param[in] file    ,    .
/// \param[in] buffer   ,   .  ,  
///                 ,     
///               . ..    copy_from_user
/// \param[in] count      ,   buffer.
/// \param[in] data       .
/// \return     
int mydrviplir_proc_write(VN_STFILE file, const char *buffer, unsigned long count, void *data);


int GCC_NOREGPARM myregister_chrdev(unsigned int, const char *, void *);
int GCC_NOREGPARM myunregister_chrdev(unsigned int, const char *);

/// \brief      procfs.
///
///      procfs.
///
/// \param[in] name   
///
/// \return     
VN_PROC_DIR_ENTRY GCC_NOREGPARM myproc_mkdir(const char *name);
/// \brief      procfs.
///
///        procfs  
/// callback      .     .
///
/// \param[in] name   
/// \param[in] parent   ,     .
/// \param[in] read_proc   callback   .
/// \param[in] write_proc   callback   .
/// \param[in] data       .
/// \return     
VN_PROC_DIR_ENTRY GCC_NOREGPARM mycreate_proc_entry(const char *name, VN_PROC_DIR_ENTRY parent,
                                                    void * read_proc, void * write_proc, void * data);
/// \brief  /  procfs
///
///          procfs.
///
/// \param[in] name   /
/// \param[in] parent   ,     .
void GCC_NOREGPARM myremove_proc_entry(const char *name, VN_PROC_DIR_ENTRY parent);

void GCC_NOREGPARM myip_send_check(void * ip);
void mypanic(const char * fmt, ...);
int GCC_NOREGPARM mydel_timer(VN_TIMER timer);
int GCC_NOREGPARM mydel_timer_sync(VN_TIMER timer);
void GCC_NOREGPARM myadd_timer(VN_TIMER timer);
unsigned int GCC_NOREGPARM myeth_type_trans(VN_SKB skb, VN_DEVICE dev);
void myschedule(void);
void mythread_yield(void);	/*release cpu's ownership*/
/* FIXME: Uncomment next string when itcskrniface is available
 * int GCC_NOREGPARM myset_affinity(unsigned int cpu_id);
 */
unsigned long GCC_NOREGPARM mycopy_from_user(void * to, const void * from, unsigned long n);
unsigned long GCC_NOREGPARM mycopy_to_user(void * to, const void * from, unsigned long n);
int GCC_NOREGPARM myrequest_irq(unsigned int irq,
		  void * handler,
		  const char * device,
		  void * dev_id);
void GCC_NOREGPARM myfree_irq(unsigned int irq, void * dev_id);
void mymark_bh(void);
#define mypci_present mypcibios_present
int mypcibios_present(void);
void * GCC_NOREGPARM myioremap(unsigned long offset, unsigned long size);
void GCC_NOREGPARM myiounmap(void * addr);
void GCC_NOREGPARM myqueue_task(VN_TQSTRUCT bh_pointer, VN_TASKQUEUE bh_list);
int GCC_NOREGPARM myschedule_task(VN_TQSTRUCT task);
int GCC_NOREGPARM mycancel_work_sync(VN_TQSTRUCT tq);
void GCC_NOREGPARM myflush_scheduled_tasks(void);

#define myvirt_to_bus myvirt_to_phys
unsigned long GCC_NOREGPARM myvirt_to_phys(volatile void * address);
unsigned long GCC_NOREGPARM my__get_free_pages(unsigned long gfp_order);
void GCC_NOREGPARM myfree_pages(unsigned long addr, unsigned long order);
unsigned int GCC_NOREGPARM myreadl(void * addr);
void GCC_NOREGPARM mywritel(unsigned int b, void * addr);
#if !defined(__arm__)
int GCC_NOREGPARM mypci_read_config_base_addr0(VN_PCIDEV dev, unsigned int * val);
#endif
void myreboot(void);
#if 0
void GCC_NOREGPARM myarp_send(int, int, unsigned long, VN_DEVICE, unsigned long,
                unsigned char *, unsigned char *, unsigned char *);
VN_NEIGHBOUR GCC_NOREGPARM my__neigh_lookup(VN_NEIGHTABLE, const void *, VN_DEVICE, int);
int GCC_NOREGPARM myneigh_update(VN_NEIGHBOUR, unsigned char *, unsigned char, int, int);
void GCC_NOREGPARM myneigh_release(VN_NEIGHBOUR);
#endif

VN_TASK_STRUCT GCC_NOREGPARM mykthread_run( int GCC_NOREGPARM (*fn)(void *), const char *name );
int GCC_NOREGPARM mykthread_stop( VN_TASK_STRUCT tsk );
void myfinish_kernel_thread(void);

void GCC_NOREGPARM myinit_waitqueue_head(VN_WAITQUEUE p);
int GCC_NOREGPARM myinterruptible_sleep_on(VN_WAITQUEUE p);
int GCC_NOREGPARM myinterruptible_sleep_on_exclusive(VN_WAITQUEUE p);
void GCC_NOREGPARM myset_waitqueue_state(VN_WAITQUEUE p, unsigned long state);

int mysignal_pending_current(void);

int GCC_NOREGPARM myis_task_present( const int taskPid );
#ifdef DEBUG_F1024
VN_SIGINFO_T mysiginfo_alloc(void);
int GCC_NOREGPARM myallow_signal(int sig);
unsigned long  GCC_NOREGPARM mydequeue_signal_lock_current(VN_SIGINFO_T info);
void myBUG(void);
#endif

void GCC_NOREGPARM mywake_up_interruptible(VN_WAITQUEUE q);
void GCC_NOREGPARM mywake_up_interruptible_all(VN_WAITQUEUE q);
void GCC_NOREGPARM mydaemonize_thread(char *name);
void myMOD_INC_USE_COUNT(void);
void myMOD_DEC_USE_COUNT(void);
void GCC_NOREGPARM myspinlock_lock(VN_SPINLOCK);
void GCC_NOREGPARM myspinlock_unlock(VN_SPINLOCK);
void GCC_NOREGPARM myspinlock_lock_irqsave(VN_SPINLOCK, unsigned long *);
void GCC_NOREGPARM myspinlock_unlock_irqrestore(VN_SPINLOCK, unsigned long *);
int GCC_NOREGPARM myspinlock_trylock(VN_SPINLOCK);
int GCC_NOREGPARM mymutex_lock(VN_MUTEX);
int GCC_NOREGPARM mymutex_lock_interruptible(VN_MUTEX);
void GCC_NOREGPARM mymutex_unlock(VN_MUTEX);
int GCC_NOREGPARM mymutex_trylock(VN_MUTEX);
void GCC_NOREGPARM myread_lock(VN_RWLOCK);
void GCC_NOREGPARM myread_unlock(VN_RWLOCK);
void GCC_NOREGPARM mywrite_lock(VN_RWLOCK);
void GCC_NOREGPARM mywrite_unlock(VN_RWLOCK);
void GCC_NOREGPARM myread_lock_bh(VN_RWLOCK rwlock);
void GCC_NOREGPARM myread_unlock_bh(VN_RWLOCK rwlock);
void GCC_NOREGPARM mywrite_lock_bh(VN_RWLOCK rwlock);
void GCC_NOREGPARM mywrite_unlock_bh(VN_RWLOCK rwlock);
void GCC_NOREGPARM mylocal_bh_disable(void);
void GCC_NOREGPARM mylocal_bh_enable(void);
void GCC_NOREGPARM mypreempt_disable(void);
void GCC_NOREGPARM mypreempt_enable(void);
int myin_softirq(void);
void GCC_NOREGPARM myoutb(int, int);
int myis_netfilter_method(void);
int myis_patch_method(void);
void GCC_NOREGPARM myfill_nfinfo(VN_SKB skb, VN_NFINFO nfinfo, int IsIncoming);
void GCC_NOREGPARM myfill_skb(VN_SKB skb);

// Change skb data to be "non-cloned" and update *skb in any cases.
// For non-cloned packets myskb_make_solid do nothing.
// Return negative value on error
int GCC_NOREGPARM myskb_make_solid(VN_SKB *skb, VN_NFINFO nfinfo);

/// @brief      sk_buff
///
/// @param skb  sk_buff
///
///   ,    .    skb  , 
///   .
void GCC_NOREGPARM myskb_dst_drop(VN_SKB skb);
void GCC_NOREGPARM mynf_reinject(VN_SKB skb, VN_NFINFO nfinfo);
void GCC_NOREGPARM myclone_nf_info(VN_NFINFO nfi_src, VN_NFINFO * nfi_dst, VN_SKB skb);
void GCC_NOREGPARM mycreate_nf_info(VN_NFINFO * nfi_dst,
                  VN_SKB skb, VN_DEVICE physindev, VN_DEVICE physoutdev);
void GCC_NOREGPARM myfree_nf_info_and_skb(VN_NFINFO nfinfo, VN_SKB skb);
void GCC_NOREGPARM myfree_nf_info_only(VN_NFINFO nfq);
void GCC_NOREGPARM myrelease_nf_refs(VN_NFINFO nfq);
int GCC_NOREGPARM myfind_best_route( unsigned int destinIp, unsigned int *sourceIp, 
                  unsigned int *nextHost, VN_DEVICE *dev );
/// \brief Find route for destinIp and assign dst, if skb is supplied
int GCC_NOREGPARM myfind_best_route_and_set( unsigned int destinIp, unsigned int *sourceIp, 
                  unsigned int *nextHost, VN_DEVICE *dev, VN_SKB skb );
int GCC_NOREGPARM myneigh_lookup( VN_DEVICE device, unsigned int* ipAddress, 
                  unsigned char* macAddress, unsigned int* macAddressSize );
VN_SKB GCC_NOREGPARM myip_defrag(VN_SKB skb);

int mysmp_processor_id(void);
int myin_interrupt(void);
void GCC_NOREGPARM mydst_hold(VN_DST dst);
void GCC_NOREGPARM mydst_release(VN_DST dst);
VN_DEV GCC_NOREGPARM mydst_dev(VN_DST dst);
void GCC_NOREGPARM mydump_skb(VN_SKB skb);
void GCC_NOREGPARM mydump_skb_iphdr(const char *label, const char *func, const int line, VN_SKB skb);
void GCC_NOREGPARM mydump_nfinfo(const char *func, const int line, VN_NFINFO nfinfo);

VN_ATOMIC myatomic_alloc(void);
void GCC_NOREGPARM myatomic_free(VN_ATOMIC a);
int GCC_NOREGPARM myatomic_read(VN_ATOMIC a);
void GCC_NOREGPARM myatomic_set(VN_ATOMIC a, int value);
void GCC_NOREGPARM myatomic_inc(VN_ATOMIC a);
void GCC_NOREGPARM myatomic_dec(VN_ATOMIC a);
void GCC_NOREGPARM mydisable_irq(int num);
void GCC_NOREGPARM myenable_irq(int num);
int GCC_NOREGPARM myatomic_inc_return(VN_ATOMIC a);
int GCC_NOREGPARM myatomic_dec_return(VN_ATOMIC a);
int GCC_NOREGPARM myatomic_dec_and_test(VN_ATOMIC a);

/* new atomic_long routines. it's have correct names, and powerfull */
void GCC_NOREGPARM my_atomic_long_inc(volatile long * dest);
long GCC_NOREGPARM my_atomic_long_inc_return(volatile long * dest);
void GCC_NOREGPARM my_atomic_long_add(volatile long * dest, long value);
long GCC_NOREGPARM my_atomic_long_add_return(volatile long * dest, long value);
void GCC_NOREGPARM my_atomic_long_dec(volatile long * dest);
long GCC_NOREGPARM my_atomic_long_dec_return(volatile long * dest);
void GCC_NOREGPARM my_atomic_long_sub(volatile long * dest, long value);
long GCC_NOREGPARM my_atomic_long_sub_return(volatile long * dest, long value);
long GCC_NOREGPARM my_atomic_long_xchg(volatile long * dest, long value);
long GCC_NOREGPARM my_atomic_long_cmpxchg(volatile long * dest, long oldval, long newval);


void GCC_NOREGPARM myatomic32_add(volatile long * dest, long val);
long GCC_NOREGPARM myatomic32_add_return(volatile long * dest, long val);
long GCC_NOREGPARM myatomic32_inc_return(volatile long * dest);
long GCC_NOREGPARM myatomic32_dec_return(volatile long * dest);
long GCC_NOREGPARM myatomic32_xchg(volatile long * dest, long value);

void GCC_NOREGPARM myatomic64_add(volatile long long * dest, long val);
long long GCC_NOREGPARM myatomic64_add_return(volatile long long * dest, long val);
long long GCC_NOREGPARM myatomic64_inc_return(volatile long long * dest);
long long GCC_NOREGPARM myatomic64_dec_return(volatile long long * dest);

int GCC_NOREGPARM myset_net_filter(VN_NFQFUNC nfq_in, VN_NFQFUNC nfq_out);
long int  mygetpid(void);
int mynum_online_cpus(void );
int GCC_NOREGPARM myset_affinity(int cpu);
unsigned long long myrdtsc(void );

/// \brief   syslog    
///
///     syslog    .
///
void mydump_stack( void );

/// \brief   syslog     
///
///     syslog    
///       .
///
void mydump_stack_all( void );

// -----------------------------------------------------------------------
// VN_EVENT - many wait, one control

typedef void* VN_EVENT;	// event abstarction type

// event create/destroy
VN_EVENT myevent_alloc( void );
void GCC_NOREGPARM myevent_free( VN_EVENT event );

// event actions
int GCC_NOREGPARM myevent_wait( VN_EVENT event );

void GCC_NOREGPARM myevent_set( VN_EVENT event );
void GCC_NOREGPARM myevent_reset( VN_EVENT event );

// -----------------------------------------------------------------------
typedef void * VN_SEMAPHORE;

VN_SEMAPHORE mysemaphore_alloc( void );
void GCC_NOREGPARM mysemaphore_free( VN_SEMAPHORE sem );

// semaphore actions
void GCC_NOREGPARM myup( VN_SEMAPHORE event );
void GCC_NOREGPARM mydown( VN_SEMAPHORE event );

#if defined ( DEBUG_USE_LOCKMETER ) || defined( CONFIG_LOCKMETER )
/*  when compile stubs                 when compile drvkrniface.c */

typedef struct {
	/* required for LOCKMETER since all bits in lock    */
	/* are used and we need this storage for lock INDEX */
	volatile unsigned int index;
} mymetered_info_t;

#endif

#ifdef	DEBUG_USE_LOCKMETER

// when use lockemter oin common part, drvkrniface.c must contain next functions( target kernel must be compiled with CONFIG_LOCKMETER )

void GCC_NOREGPARM _mymetered_enter(mymetered_info_t *lock_ptr, void *caller_pc);
void GCC_NOREGPARM _mymetered_leave(mymetered_info_t *lock_ptr);
void * GCC_NOREGPARM _current_address(void);

#define MYDEFINE_METERED_INFO(A) mymetered_info_t metered_##A = { .index = 0 }
#define MYEXTERN_METERED_INFO(A) extern mymetered_info_t metered_##A
#define MYDEFINE_METERED_INFO_CPP(A) mymetered_info_t metered_##A = { 0 }
#define mymetered_enter(A)	\
		_mymetered_enter((mymetered_info_t *)&metered_##A, __builtin_return_address(0))
#define mymetered_leave(A)	\
		_mymetered_leave((mymetered_info_t *)&metered_##A)
#define mymetered_start(A)	\
		_mymetered_enter((mymetered_info_t *)&metered_##A, _current_address())

#else

#define MYDEFINE_METERED_INFO(A)
#define MYEXTERN_METERED_INFO(A)
#define MYDEFINE_METERED_INFO_CPP(A)
#define mymetered_enter(A)
#define mymetered_leave(A)
#define mymetered_start(A)

#endif


#if defined( IPLIR_XFRM_INPUT ) || defined( CONFIG_IPLIR_XFRM_INPUT )
/* To fit in unsigned char skb->cb->flags */
#define SEND_FLAG 127
#define RECV_FLAG 255
#else
#define SEND_FLAG 16436
#define RECV_FLAG 17565
#endif

#define FM_NONE		0
#define FM_PATCH 		1
#define FM_NETFILTER	2

/*flags from drvkrniface.c*/
#define FM_BROADCAST 	(1<<0)
#define FM_URGENT 		(1<<1)



#include "struct_access.h"

/**
 *   ,    
 *
 * @param handler - 
 * @return 0   ,  -   errno
 */
int GCC_NOREGPARM myregister_netdevice_notifier(void GCC_NOREGPARM (*handler)(unsigned long, VN_DEVICE));

/// @brief   ,     
/// @param handler - 
/// @return 0   ,  -   errno
int GCC_NOREGPARM myregister_inetaddr_notifier( void (*handler)( unsigned long, VN_INIFADDR ) );

/**
 *   ,    
 */
void myunregister_netdevice_notifier(void);

/// @brief   ,     
void myunregister_inetaddr_notifier(void);

///    NETDEV_UNREGISTER
unsigned long myNETDEV_UNREGISTER(void);
///    NETDEV_REGISTER
unsigned long myNETDEV_REGISTER(void);
///    NETDEV_CHANGENAME
unsigned long myNETDEV_CHANGENAME(void);
///    NETDEV_UP
unsigned long myNETDEV_UP(void);
///    NETDEV_DOWN
unsigned long myNETDEV_DOWN(void);
// Set normal scheduler and zero priority for current thread
void myset_normal_scheduler(void);

/// @brief Extract source IP4/IP6 adress from flow descriptor
/// @param fl - flow handle
/// @param family - address family
VN_ADDR6 GCC_NOREGPARM my_flow_saddr(const VN_FLOW fl, unsigned short family);

/// @brief Extract destination IP4/IP6 adress from flow descriptor
/// @param fl - flow handle
/// @param family - address family
VN_ADDR6 GCC_NOREGPARM my_flow_daddr(const VN_FLOW fl, unsigned short family);

void GCC_NOREGPARM my_print_ip4_addr(unsigned long be_addr);
void GCC_NOREGPARM my_print_ip6_addr(VN_ADDR6 addr);

/// @brief      
/// @param proto - ip ,     IPPROTO_UDP  IPPROTO_TCP
/// @param localPort -   
/// @param localIp -  ip    ,     IPPROTO_UDP
/// @return 1     ,  - 0
int GCC_NOREGPARM my_is_ip4_local_port_inuse( unsigned char proto, unsigned short localPort, unsigned int localIp );

#endif
