#include "struct_access.h"
#include "linux/slab.h"
#include "linux/tcp.h"
#include "linux/udp.h"
#include <linux/sched.h>
#include <linux/netdevice.h>
#include <linux/if_ether.h>

/*
 * skbuff_get()  - Get option from sk_buff structure
 */
unsigned int GCC_NOREGPARM skbuff_get(void *p, skbuff_types_u t)
{
     struct sk_buff *skb = (struct sk_buff *)p;

     switch (t)
     {
          case pkt_type:
               return skb->pkt_type;
          case len:
               return skb->len;
          default:
               return 0;
     }
}

/*
 * skbuff_get()  - Set option to sk_buff structure
 */
void GCC_NOREGPARM skbuff_set(void *p, skbuff_types_u t, unsigned long opt)
{
     struct sk_buff *skb = (struct sk_buff *)p;

     switch (t)
     {
          case pkt_type:
               skb->pkt_type = opt;
               break;
          case len:
               skb->len = opt;
               break;
     }
     return;
}

/* System time related stuff. */
long mytime(void)
{
  return myrealtime() - 60 * 60 * 24 * (365 * 10 + 2);
}

long myrealtime(void)
{
  return get_seconds();
}

extern struct timezone sys_tz;
long mylocaltime(void)
{
  return get_seconds() - sys_tz.tz_minuteswest * 60;
}


void GCC_NOREGPARM mygetutime(int * tv)
{
  struct timespec tt = CURRENT_TIME;
  tv[0] = tt.tv_sec;
  tv[1] = tt.tv_nsec;
}

void GCC_NOREGPARM mygetlongtime(unsigned long long * ptv)
{
  /* Warning! Dividing 64 by 64/32 sometimes cause unresolved __aeabi__uldivmod */
  struct timespec tt = CURRENT_TIME;
  *ptv = 1000ULL * tt.tv_sec + __udivdi3(tt.tv_nsec, 1000000UL);
}

unsigned long GCC_NOREGPARM myjiffies(void)
{
  return jiffies;
}

unsigned long GCC_NOREGPARM myhz(void)
{
  return HZ;
}

unsigned int mytimeval_sizeof(void)
{
  return sizeof(struct timeval);
}

/* Task queues. */

VN_TQSTRUCT myalloc_tq_struct(void)
{
  struct mytimer_info *tmp;
  tmp = (struct mytimer_info *)mykmalloc(sizeof(struct mytimer_info));
  memset(tmp, 0, sizeof(struct mytimer_info));
//  myprintk("%s: tq_struct %08X\n", __FUNCTION__, (int)tmp);
  return tmp;
}

void GCC_NOREGPARM myfree_tq_struct(VN_TQSTRUCT tq)
{
//  myprintk("%s: tq_struct %08X\n", __FUNCTION__, (int)tq);

#if defined(KERNEL_USE_REGPARM)
  void * fn_arg;
  fn_arg = ((struct mytimer_info *)tq)->arg;
#if defined (IFACE_ITCSIPCLS)
  mykfree(fn_arg);
#endif
#endif
  mykfree(tq);
}

int GCC_NOREGPARM myfill_tq_struct(VN_TQSTRUCT tq, void GCC_NOREGPARM (*routine)(void *), void * arg)
{
  void (* fn)(void *);
  void * fn_arg;

#if defined(KERNEL_USE_REGPARM)
  struct _univ_handle * data;
  data = (struct _univ_handle *)mykmalloc(sizeof(struct _univ_handle));
  if (!data)
  {
    udebug(( "itcs_drv: non mem for task handle data!" ));
    myprintk("%s: itcs_drv: non mem for task handle data!\n", __FUNCTION__);
    return -1;
  }
  data->fn = (int GCC_NOREGPARM (*)(void *)) routine;
  data->arg = arg;
  data->flclean = 1;

  fn = (void (* )(void *)) univ_handle;
  fn_arg = data;
#else
  fn = routine;
  fn_arg = arg;
#endif

  ((struct mytimer_info *)tq)->fn = fn;
  ((struct mytimer_info *)tq)->arg = fn_arg;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
//  myprintk("%s: task queue %08X, routine %08X, arg %08X, data %08X\n", __FUNCTION__, (int)tq, (int)routine, (int)arg, (int)data);
  INIT_WORK( &((struct mytimer_info *)tq)->ws, (void (*)(void *)) mytimer_handler, tq );
#else
  INIT_WORK( &((struct mytimer_info *)tq)->ws, (void (*)(struct work_struct *))mytimer_handler);
#endif // LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
  return 0;
}

void mytimer_handler(void *arg)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
     struct work_struct *work = (struct work_struct *)arg;
     struct mytimer_info *info =
		container_of(work, struct mytimer_info, ws);
#else
     struct mytimer_info *info = (struct mytimer_info *)arg;
//     myprintk("%s: fn %08X, fn->arg %08X, arg %08X\n", __FUNCTION__, (int)info->fn, (int)info->arg, (int)arg);
#endif
     info->fn(info->arg);
     return;
}

/* Wait queue. */
VN_WAITQUEUE myalloc_waitqueue(void)
{
  WQHANDLEPTR wqh;
  LINUX_WAIT_QUEUE *tmp;

  tmp = (LINUX_WAIT_QUEUE *)mykmalloc(sizeof(LINUX_WAIT_QUEUE));
  if (!tmp)
    return NULL;

  init_waitqueue_head(tmp);
  wqh = (WQHANDLEPTR)mykmalloc(sizeof(WQHANDLE));
  if (!wqh)
  {
    mykfree(tmp);
    return NULL;
  }
  wqh->wq = tmp;
  return wqh;
}

void GCC_NOREGPARM myfree_waitqueue(VN_WAITQUEUE wq)
{
  WQHANDLEPTR wqh = (WQHANDLEPTR)wq;
  if (!wq)
    return;
  if(wqh->wq)
    mykfree(wqh->wq);
  mykfree(wq);
}

/* sk_buff manupulation. */
VN_SKB GCC_NOREGPARM mydev_alloc_skb(unsigned int size)
{
  return dev_alloc_skb(size);
}

VN_SKB GCC_NOREGPARM myalloc_skb(unsigned int size)
{
#ifdef USE_RWLOCK
   /*
    * This function can be called when rw_lock is held. Therefore it can not sleep
    */
#define MYGFP_FLAGS GFP_ATOMIC
#else
   /*
    * This function is called from thread context and therefore can sleep
    * In this case flag GFP_KERNEL can be used for memory allocation.
    */
#define MYGFP_FLAGS GFP_KERNEL
#endif //USE_RWLOCK
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,14)
  return (VN_SKB)alloc_skb(size, MYGFP_FLAGS);

#elif LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
  return (VN_SKB)__alloc_skb(size, MYGFP_FLAGS, 0);
#else
  return (VN_SKB)__alloc_skb(size, MYGFP_FLAGS, 0, -1);
#endif
#undef MYGFP_FLAGS
}

void GCC_NOREGPARM mykfree_skb(VN_SKB skb)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,14)
  kfree_skb((struct sk_buff *)skb);
#else
  __kfree_skb((struct sk_buff *)skb);
#endif
}

void GCC_NOREGPARM mydev_kfree_skb(VN_SKB skb)
{
  dev_kfree_skb((struct sk_buff *)skb);
}

unsigned char * GCC_NOREGPARM myskb_put(VN_SKB skb, int len)
{
  return skb_put((struct sk_buff *)skb, len);
}

unsigned char * GCC_NOREGPARM myskb_push(VN_SKB skb, int len)
{
  return skb_push((struct sk_buff *)skb, len);
}

VN_SKB GCC_NOREGPARM myskb_copy_expand(VN_SKB skb, unsigned int size, unsigned int frontOffset, unsigned int backOffset)
{
   struct sk_buff *n;
   struct sk_buff *o = (struct sk_buff *)skb;
   unsigned int len = o->len;
   /* Temporarily change len in order to allocate skb with required size */
   o->len = size;
#ifdef USE_RWLOCK
   /*
    * This function can be called when rw_lock is held. Therefore it can not sleep
    */
   n = skb_copy_expand(o, frontOffset, backOffset, GFP_ATOMIC);
#else
   /*
    * This function is called from thread context and therefore can sleep
    * In this case flag GFP_KERNEL can be used for memory allocation.
    */
   n = skb_copy_expand(o, frontOffset, backOffset, GFP_KERNEL);
#endif //USE_RWLOCK
   /* Bring back original value */
   o->len = len;
   return (VN_SKB)n;
}

VN_SKB GCC_NOREGPARM myskb_create( unsigned int size, unsigned int frontOffset, unsigned int backOffset )
{
     /*
      * This function is called from thread context and therefore can sleep
      * In this case flag GFP_KERNEL can be used for memory allocation.
      */

     /*
      *      Allocate the copy buffer
      */
     struct sk_buff *n;
     /* We have to preallocate a buffer at least for Ethernet header (worst case). It should be hardware aligned. */
     if( frontOffset < HH_DATA_ALIGN(ETH_HLEN) )
     {
          frontOffset = HH_DATA_ALIGN(ETH_HLEN);
     }
     n = myalloc_skb(size + frontOffset + backOffset );
     if (!n)
     {
          return NULL;
     }

     /* Set the data pointer */
     skb_reserve(n, frontOffset);
     /* Set the tail pointer and length */
     skb_put(n, size);

     /// Using of PATCH method doesn't automatically lead to Ethernet protocol usage.
     /// So we don't need to offset network and transport header pointers here (it must be moved later in case of 802.3)
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,14)
     n->stamp.tv_sec     = 0;
     n->stamp.tv_usec    = 0;
#else
     n->tstamp.off_sec     = 0;
     n->tstamp.off_usec    = 0;
#endif
     n->h.raw              = n->data;
     n->nh.raw             = n->data;
     n->mac.raw            = n->data;
#else
     n->tstamp             = ktime_set(0,0);
     skb_reset_network_header(n);
     skb_reset_transport_header(n);
     skb_reset_mac_header(n);
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,14)
     n->fclone             = SKB_FCLONE_UNAVAILABLE;
#endif
     /* Enable local fragmentation */
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,16,0)
     n->local_df           = 1;
#else
     n->ignore_df          = 1;
#endif
     n->protocol           = htons(ETH_P_IP);
     return n;
}

void GCC_NOREGPARM myskb_reserve(VN_SKB skb, unsigned int len)
{
  skb_reserve((struct sk_buff *)skb, len);
}

int GCC_NOREGPARM myskb_expand_head(VN_SKB skb, int nhead, int ntail)
{
     struct sk_buff* sb = (struct sk_buff *)skb;
     if( ( (sb->data - sb->head) >= nhead ) && ( (sb->end - sb->tail) >= ntail) )
     {
          if( skb_cloned( sb ) )
          {
#ifdef USE_RWLOCK
               return pskb_expand_head(sb, nhead, ntail, GFP_ATOMIC);
#else
               return pskb_expand_head(sb, nhead, ntail, GFP_KERNEL);
#endif
          }
          return 0;
     }
     
#ifdef USE_RWLOCK
  /*
   * This function can be called when rw_lock is held. Therefore it can not sleep
   */
  return pskb_expand_head(sb, nhead, ntail, GFP_ATOMIC);
#else
  return pskb_expand_head(sb, nhead, ntail, GFP_KERNEL);
#endif

}

unsigned char * GCC_NOREGPARM myskb_pull(VN_SKB skb, unsigned int len)
{
     return pskb_pull((struct sk_buff *)skb, len);
}

void GCC_NOREGPARM myskb_move_mac_header(VN_SKB skb, int offset)
{
     struct sk_buff *ps = (struct sk_buff *)skb;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
     if( ps->mac.raw != 0 )
     {
          ps->mac.raw += offset;
     }
#else
     if( skb_mac_header_was_set(ps) )
     {
          ps->mac_header += offset;
     }
#endif
}

void GCC_NOREGPARM myskb_move_headers(VN_SKB skb, int offset)
{
     struct sk_buff *ps = (struct sk_buff *)skb;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
     ps->h.raw += offset;
     ps->nh.raw += offset;
     if( ps->mac.raw != 0 )
     {
          ps->mac.raw += offset;
     }
#else
     ps->transport_header += offset;
     ps->network_header += offset;
     if( skb_mac_header_was_set(ps) )
     {
          ps->mac_header += offset;
     }
#endif
}

int GCC_NOREGPARM myskb_trim(VN_SKB skb, unsigned int len)
{
     return pskb_trim((struct sk_buff *)skb, len);
}

int GCC_NOREGPARM myskb_linearize(VN_SKB skb)
{

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
     return __skb_linearize((struct sk_buff *)skb);
#elif defined(SLE_VERSION_CODE) && defined(SLE_VERSION)
#if (SLE_VERSION_CODE >= SLE_VERSION(10,1,0))
     // SLES 10 SP1,2+ with patched kernel 2.6.16 has __skb_linearize with one argument like 2.6.18
     return __skb_linearize((struct sk_buff *)skb);
#else // !SLES SP1,2
     return __skb_linearize((struct sk_buff *)skb, GFP_ATOMIC);
#endif
#else //!defined(SLE_VERSION_CODE) && !defined(SLE_VERSION)
     return __skb_linearize((struct sk_buff *)skb, GFP_ATOMIC);
#endif
}

unsigned char GCC_NOREGPARM myskb_is_nonlinear(VN_SKB skb)
{
  return (skb_shinfo((struct sk_buff *)skb)->nr_frags ||
          skb_shinfo((struct sk_buff *)skb)->frag_list) ? 1 : 0;
}

int GCC_NOREGPARM myskb_check_and_linearize( VN_SKB skb )
{
	return myskb_is_nonlinear( skb )? myskb_linearize( skb ): 0;
}

unsigned char ** GCC_NOREGPARM myskb_fetch_dev(VN_SKB skb)
{
  return (unsigned char **)(&(((struct sk_buff *)skb)->dev));
}

void GCC_NOREGPARM myskb_set_dev(VN_SKB skb, VN_DEVICE dev)
{
  ((struct sk_buff *)skb)->dev = (struct net_device *)dev;
}

unsigned char * GCC_NOREGPARM myskb_fetch_dst(VN_SKB skb)
{
  struct dst_entry * dst;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,31)
  dst = skb_dst((struct sk_buff *)skb);
#else
  dst = ((struct sk_buff *)skb)->dst;
#endif
  return (unsigned char *)dst;
}

struct neighbour * GCC_NOREGPARM mydst_fetch_neighbour(VN_SKB skb)
{
  struct neighbour * ret;
  struct dst_entry * dst;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,31)
  dst = skb_dst((struct sk_buff *)skb);
#else
  dst = ((struct sk_buff *)skb)->dst;
#endif
  if (!dst)
  {
    return 0;
  }
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,0,21)
  myrcu_read_lock();
#if  LINUX_VERSION_CODE < KERNEL_VERSION(3,3,0)
  ret = dst_get_neighbour(dst);
#elif LINUX_VERSION_CODE < KERNEL_VERSION(3,6,0)
  ret = dst_get_neighbour_noref(dst);
#else

  ret = dst_neigh_lookup_skb(dst, skb);
#endif
  myrcu_read_unlock();
#else
  ret = (dst)->neighbour;
#endif
  return ret;
}

void GCC_NOREGPARM myneighbour_release(void *neigh) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,6,0)
          neigh_release((struct neighbour *)neigh);
#endif
}

int GCC_NOREGPARM myneighbour_fetch_type(VN_SKB skb)
{
     int type = 0;
     struct neighbour * n = mydst_fetch_neighbour(skb);
     if (n)
     {
         type = (n)->type;
         myneighbour_release((void *)n);
     }
     return type;
}
#ifdef CONFIG_IPLIR_XFRM_OUT
unsigned char * GCC_NOREGPARM myskb_fetch_cb(VN_SKB skb)
{
  return (char *)(&IPCB((struct sk_buff *)skb)->flags);
}
#else
unsigned int * GCC_NOREGPARM myskb_fetch_cb(VN_SKB skb)
{
  return (int *)(((struct sk_buff *)skb)->cb);
}
#endif
unsigned char ** GCC_NOREGPARM myskb_fetch_tail(VN_SKB skb)
{
  return (unsigned char **)(&(((struct sk_buff *)skb)->tail));
}

unsigned char ** GCC_NOREGPARM myskb_fetch_head(VN_SKB skb)
{
  return (unsigned char **)(&(((struct sk_buff *)skb)->head));
}

unsigned char ** GCC_NOREGPARM myskb_fetch_end(VN_SKB skb)
{
  return (unsigned char **)(&(((struct sk_buff *)skb)->end));
}

unsigned char ** GCC_NOREGPARM myskb_fetch_data(VN_SKB skb)
{
  return (unsigned char **)(&(((struct sk_buff *)skb)->data));
}

unsigned short GCC_NOREGPARM myskb_fetch_protocol(VN_SKB skb)
{
  return ((struct sk_buff *)skb)->protocol;
}

void GCC_NOREGPARM myskb_set_protocol(VN_SKB skb, unsigned short proto)
{
  ((struct sk_buff *)skb)->protocol = proto;
}

int GCC_NOREGPARM myis_ip_prtocol(VN_SKB skb)
{
  return ((struct sk_buff *)skb)->protocol == htons(ETH_P_IP);
}

unsigned char * GCC_NOREGPARM myskb_get_nh_raw(VN_SKB skb)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
  return (unsigned char *)( ((struct sk_buff *)skb)->nh.raw );
#else
  return skb_network_header((struct sk_buff *)skb);
#endif
}

unsigned char * GCC_NOREGPARM myskb_get_h_raw(VN_SKB skb)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
  return (unsigned char *)( ((struct sk_buff *)skb)->h.raw );
#else
  return skb_transport_header((struct sk_buff *)skb);
#endif
}

unsigned char * GCC_NOREGPARM myskb_get_mac_raw(VN_SKB skb)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
  return (unsigned char *)( ((struct sk_buff *)skb)->mac.raw );
#else
  return skb_mac_header((struct sk_buff *)skb);
#endif
}

unsigned char * GCC_NOREGPARM myskb_get_inner_mac_raw(VN_SKB skb)
{
#if defined(KERN_IPCLS)
  return skb_inner_mac_header((struct sk_buff *)skb);
#else
  return 0;
#endif
}

unsigned char ** GCC_NOREGPARM myskb_fetch_sk(VN_SKB skb)
{
  return (unsigned char **)(&(((struct sk_buff *)skb)->sk));
}

unsigned char GCC_NOREGPARM myskb_fetch_ip_summed(VN_SKB skb)
{
  return ((struct sk_buff *)skb)->ip_summed;
}

void GCC_NOREGPARM myskb_set_ip_summed(VN_SKB skb, unsigned char ip_summed)
{
  ((struct sk_buff *)skb)->ip_summed = ip_summed;
}

unsigned int * GCC_NOREGPARM myskb_fetch_len(VN_SKB skb)
{
  return &(((struct sk_buff *)skb)->len);
}

unsigned int * GCC_NOREGPARM myskb_fetch_data_len(VN_SKB skb)
{
  return &(((struct sk_buff *)skb)->data_len);
}

unsigned int * GCC_NOREGPARM myskb_fetch_csum(VN_SKB skb)
{
  return &(((struct sk_buff *)skb)->csum);
}

int GCC_NOREGPARM myskb_network_offset(VN_SKB skb)
{
	return myskb_get_nh_raw(skb) - *myskb_fetch_data(skb);
}

int GCC_NOREGPARM myskb_cow_data(VN_SKB skb_out, int tailbits, VN_SKB *trailer_out)
{
	return myskb_cow_data_krnface((struct sk_buff*)skb_out, tailbits, (struct sk_buff**)trailer_out);
}

unsigned char * GCC_NOREGPARM myskb_tail_pointer(VN_SKB skb)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,22)
	return skb_tail_pointer((struct sk_buff *)skb);
#else
	return ((struct sk_buff *)skb)->tail;
#endif
}

unsigned char * GCC_NOREGPARM mypskb_put(VN_SKB skb, VN_SKB tail, int len)
{
	if (tail != skb) {
		((struct sk_buff *)skb)->data_len += len;
		((struct sk_buff *)skb)->len += len;
	}
	return skb_put(tail, len);
}

int GCC_NOREGPARM myskb_copy_bits(VN_SKB skb, int offset, void *to, int len)
{
	return skb_copy_bits((struct sk_buff *)skb, offset, to, len);
}

int GCC_NOREGPARM my_skb_check_space(VN_SKB vnskb)
{
#if defined (KERN_IPCLS)
	struct sk_buff *skb = (struct sk_buff *)vnskb;

	struct dst_entry *dst = (struct dst_entry *)myskb_fetch_dst(skb);
	int nhead = dst->header_len + LL_RESERVED_SPACE(dst->dev)
		 - skb_headroom(skb);
	int ntail = dst->dev->needed_tailroom - skb_tailroom(skb);

	if (nhead <= 0) {
		 if (ntail <= 0)
				 return 0;
		 nhead = 0;
	} else if (ntail < 0)
		 ntail = 0;

	return myskb_expand_head(skb, nhead, ntail);
#else
	return 0;
#endif
}

void GCC_NOREGPARM my_skb_mac_header_rebuild(VN_SKB vnskb)
{
	struct sk_buff *skb = (struct sk_buff *)vnskb;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
	if (skb->mac.raw != 0) {
		const unsigned char *old_mac = skb->mac.raw;

		skb->mac.raw -= skb->mac_len;
		mymemmove(skb->mac.raw, old_mac, skb->mac_len);
	}
#else
	if (skb_mac_header_was_set(skb)) {
		const unsigned char *old_mac = skb_mac_header(skb);

		skb_set_mac_header(skb, -skb->mac_len);
		mymemmove(skb_mac_header(skb), old_mac, skb->mac_len);
	}
#endif
}

int GCC_NOREGPARM myskb_need_csum_receive(VN_SKB skbuff)
{
  struct sk_buff *skb = (struct sk_buff *)skbuff;
  return ( CHECKSUM_NONE == skb->ip_summed  ? 1 : 0 );
}


int GCC_NOREGPARM myskb_need_csum_transmit(VN_SKB skbuff)
{
  struct sk_buff *skb = (struct sk_buff *)skbuff;
  // Çíà÷åíèå NONE áóäåò ó ëîêàëüíîãî ïàêåòà, äëÿ êîòîðîãî ñòåê ïðîòîêîëîâ(L4,L3) ïîñ÷èòàë ñóììó (øòàòíàÿ ñèòóàöèÿ).
  // Çíà÷åíèå UNNECESSARY áóäåò ó îòêðûòîãî òðàíçèòíîãî ïàêåòà (âûñòàâëÿåòñÿ ïðè ïðèåìå íàìè æå).
  // Çíà÷åíèå UNNECESSARY áóäåò èçìåíåíî íàìè íà NONE èëè PARTIAL ïåðåä âîçâðàùåíèåì ïàêåòà ÿäðó.
  return ( CHECKSUM_NONE == skb->ip_summed || CHECKSUM_UNNECESSARY == skb->ip_summed ? 0 : 1 );
}

void GCC_NOREGPARM myskb_ok_csum(VN_SKB skbuff)
{
  struct sk_buff *skb = (struct sk_buff *)skbuff;
  skb->ip_summed = CHECKSUM_UNNECESSARY;
  skb->csum = 0;
}

void GCC_NOREGPARM myskb_calc_csum(VN_SKB skbuff)
{
     struct sk_buff *skb = (struct sk_buff *)skbuff;
     const struct iphdr *iph;

     if(myis_patch_method())
     {
          int32_t nh_offset = mydev_fetch_header_len( skb->dev );
          myskb_set_network_header( skb, nh_offset );
     }

     iph = kernel_skb_ip_hdr( skb );

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
     if(myis_patch_method())
     {
          uint16_t header_offset = mydev_fetch_header_len( skb->dev );
          skb->h.raw = skb->data + header_offset + sizeof(struct iphdr);
     }
     else
     {
          skb->h.raw = skb->data + sizeof(struct iphdr);
     }
#else
     if(myis_patch_method())
     {
          uint16_t header_offset = mydev_fetch_header_len( skb->dev );
          skb_set_transport_header(skb, header_offset + sizeof(struct iphdr));
     }
     else
     {
          skb_set_transport_header(skb, sizeof(struct iphdr));
     }
#endif

#ifdef CHECKSUM_HW
     skb->ip_summed = CHECKSUM_HW;
#else /* #ifdef CHECKSUM_HW */
     skb->ip_summed = CHECKSUM_PARTIAL;
#endif /* #ifndef CHECKSUM_HW */
     if( iph->protocol == IPPROTO_UDP )
     {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
          skb->csum = offsetof(struct udphdr, check);
#else /* LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22) */
          skb->csum_start = skb_transport_header(skb) - skb->head;
          skb->csum_offset = offsetof(struct udphdr, check);
#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22) */
     }
     else if( iph->protocol == IPPROTO_TCP )
      {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
          skb->csum = offsetof(struct tcphdr, check);
#else /* LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22) */
          skb->csum_start = skb_transport_header(skb) - skb->head;
          skb->csum_offset = offsetof(struct tcphdr, check);
#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22) */
      }
}

void GCC_NOREGPARM myskb_calc_udp_csum_v6(VN_SKB skbuff, int partial)
{
#if defined (KERN_IPCLS)
	struct sk_buff *skb = (struct sk_buff *)skbuff;
	struct ipv6hdr *iph = ipv6_hdr(skb);
	struct udphdr *uh;
	u32 csum;
	unsigned int offset;


	if (partial)
	{
		skb_set_transport_header(skb, sizeof(struct ipv6hdr));
		uh = udp_hdr(skb);

		uh->check = ~csum_ipv6_magic(&iph->saddr, &iph->daddr, ntohs(uh->len), IPPROTO_UDP, 0);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
		skb->csum = offsetof(struct udphdr, check);
#else /* LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22) */
		skb->csum_start = skb_transport_header(skb) - skb->head;
		skb->csum_offset = offsetof(struct udphdr, check);
#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22) */
#ifdef CHECKSUM_HW
		skb->ip_summed = CHECKSUM_HW;
#else /* #ifdef CHECKSUM_HW */
		skb->ip_summed = CHECKSUM_PARTIAL;
#endif /* #ifndef CHECKSUM_HW */
	}
	else
	{
		offset = sizeof(struct ipv6hdr);
		skb_set_transport_header(skb, offset);
		uh = udp_hdr(skb);

		// calc checksum of IP payload as per RFCxxx
		csum = skb_checksum(skb, offset, skb->len - offset, 0);

		// calc checksum of IPv6 header
		uh->check = csum_ipv6_magic(&iph->saddr, &iph->daddr, ntohs(uh->len), IPPROTO_UDP, csum);
	}
#endif
}

/*
 * Return pointer to IPLIR header
 */
unsigned char * GCC_NOREGPARM myskb_insert_udp_hdr_v6(VN_SKB skbuff, unsigned short sport_be, unsigned short dport_be, int touchh)
{
#if defined (KERN_IPCLS)
	struct sk_buff *skb = (struct sk_buff *)skbuff;
	struct udphdr *udph;
	int err;

	// realloc space, move layer-2 header back - do real insertion
	if (touchh)
	{
		int nhead = sizeof(struct udphdr);
		struct ipv6hdr *oldh;

		if (nhead < skb_headroom(skb)) {
			err = myskb_expand_head(skb, nhead, 0);
			if (err != 0)
				return err;
		}

		oldh = ipv6_hdr(skb);
		skb->network_header -= sizeof(struct udphdr);
		mymemmove(ipv6_hdr(skb), oldh, sizeof(struct ipv6hdr));
		skb->transport_header = skb->network_header + sizeof(struct ipv6hdr);

		ipv6_hdr(skb)->payload_len += htons(sizeof(struct udphdr));
		ipv6_hdr(skb)->nexthdr = IPPROTO_UDP;
	}

	// place UDP header at existing room, no touch to other headers

	udph = udp_hdr(skb);
	udph->source = sport_be;
	udph->dest 	 = dport_be;
	udph->len    = 0x0;
	udph->check  = 0x0;
	// keep pointer to UDP length field
	skb->inner_mac_header = (u8*)udph + offsetof(struct udphdr, len);
	// point it to skb->data (IPLIR header)
	skb->transport_header += sizeof(struct udphdr);

	return skb->transport_header;
#endif
	return NULL;
}

void GCC_NOREGPARM myskb_remove_udp_hdr_v6(VN_SKB skbuff)
{
#if defined (KERN_IPCLS)
	struct ipv6hdr *oldh;
	struct sk_buff *skb = (struct sk_buff *)skbuff;

	// remove encap. header
	myskb_pull(skb, sizeof(struct udphdr));
	skb_reset_transport_header(skb);
	// move back ip6 header
	oldh = ipv6_hdr(skb);
	skb_set_network_header(skb, -sizeof(struct ipv6hdr));
	mymemmove((u8*)ipv6_hdr(skb), (u8*)oldh, sizeof(struct ipv6hdr));
	ipv6_hdr(skb)->payload_len = htons(skb->len);
	// we assuming so (unsafely)
	ipv6_hdr(skb)->nexthdr = 0xf1;
#endif
}

void GCC_NOREGPARM myskb_clear_csum(VN_SKB skbuff)
{
  struct sk_buff *skb = (struct sk_buff *)skbuff;
  skb->ip_summed = CHECKSUM_NONE;
  skb->csum = 0;
}


int GCC_NOREGPARM myskb_size_overhead_mtu(VN_SKB skbuff)
{
  struct sk_buff *skb = (struct sk_buff *)skbuff;
  return ( *myskb_fetch_len(skb) > *mydev_fetch_mtu( *myskb_fetch_dev(skb) ) ? 1 : 0 );
}

void GCC_NOREGPARM myskb_set_network_header(VN_SKB skbuff, int offset)
{
  struct sk_buff *skb = (struct sk_buff *)skbuff;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
  skb->nh.raw = skb->data + offset;
#else
  skb_set_network_header(skb, offset);
#endif
}

int GCC_NOREGPARM myskb_create_secpath(VN_SKB skbuff)
{
#if defined( CONFIG_IPLIR_XFRM_INPUT )
    struct sk_buff *skb = (struct sk_buff *)skbuff;

    if (skb->sp)
    {
       if (likely(skb->sp->len > 0))
            // secpath_put(skb->sp);
            return 1; /* don't release previous (IPsec) secpath, return error instead */
       else
            return 0;
    }

    skb->sp = secpath_dup(NULL);
    if (unlikely(!skb->sp)) 
         return 1;

    /* iplir decap. marker (used in iplir-xfrm4 patch) */
    skb->sp->xvec[0] = IPLIR_NF_XFRM_MARK;

    return 0;
#else
    return 1;
#endif
}

void * GCC_NOREGPARM myskb_get_secpath_iplir_storage(VN_SKB skbuff)
{
    void *ret = NULL;
    struct sk_buff *skb = (struct sk_buff *)skbuff;

#if defined( CONFIG_IPLIR_XFRM_INPUT )
    /* ensure SP allocated and used by us (iplir) */
    if( !skb->sp || skb->sp->xvec[0] != IPLIR_NF_XFRM_MARK )
        return ret;

    /* pointer to storage */
    ret = (void*)&skb->sp->xvec[1];
#endif
    return ret;
}

int GCC_NOREGPARM myskb_is_orig_secpath_exist(VN_SKB skbuff)
{
    int ret = 0;

#if defined( CONFIG_IPLIR_XFRM_INPUT )
    struct sk_buff *skb = (struct sk_buff *)skbuff;

    /* ensure SP allocated and used not by us (iplir) */
    ret = ( skb->sp && (skb->sp->len > 0) && (skb->sp->xvec[0] != IPLIR_NF_XFRM_MARK) );
#endif
    return ret;
}

#if 0
unsigned char ** GCC_NOREGPARM myskb_fetch_physindev(VN_SKB skb)
{
  return (unsigned char **)(&(((struct sk_buff *)skb)->physindev));
}

unsigned char ** GCC_NOREGPARM myskb_fetch_physoutdev(VN_SKB skb)
{
  return (unsigned char **)(&(((struct sk_buff *)skb)->physoutdev));
}
#endif
void GCC_NOREGPARM myskb_clear_dst(VN_SKB skb)
{
     struct dst_entry * dst =(struct dst_entry *)myskb_fetch_dst(skb);

     if (dst
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,35)
          && !skb_dst_is_noref(skb)
#endif
          )
          dst_release(dst);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,31)
     skb_dst_set((struct sk_buff *)skb, NULL);
#else
     ((struct sk_buff *)skb)->dst = NULL;
#endif

     /* with NETFILTER and 2.6.32 on outgoing reversed local packets we have cached socket that
        prevents it to complete properly so that we have to clean it along with dst entry.
        it is unclear what to do with destructor field which is surely will be sock_wfree()
        in that case; need more research */
     if (((struct sk_buff *)skb)->sk)
     {
          /*sock_put(((struct sk_buff *)skb)->sk);*/
          sock_wfree((struct sk_buff *)skb);
          ((struct sk_buff *)skb)->sk = NULL;
          ((struct sk_buff *)skb)->destructor = NULL;
     }
}

/* net_device manupulation. */
unsigned short * GCC_NOREGPARM mydev_fetch_type(VN_DEVICE dev)
{
  return &(((LINUX_DEVICE *)dev)->type);
}

unsigned int GCC_NOREGPARM mydev_fetch_flags(VN_DEVICE dev)
{
  return ((LINUX_DEVICE *)dev)->flags;
}

unsigned char * GCC_NOREGPARM mydev_fetch_name(VN_DEVICE dev)
{
  return ((LINUX_DEVICE *)dev)->name;
}

unsigned short GCC_NOREGPARM mydev_fetch_header_len( VN_DEVICE dev )
{
     if( ((LINUX_DEVICE *)dev)->type == ARPHRD_ETHER )
     {
          return ETH_HLEN;
     }
     /// @todo add another L2 protocol header management

     return 0;
}

unsigned * GCC_NOREGPARM mydev_fetch_mtu(VN_DEVICE dev)
{
  return &(((LINUX_DEVICE *)dev)->mtu);
}

unsigned char ** GCC_NOREGPARM mydev_fetch_ip_ptr(VN_DEVICE dev)
{
  return (unsigned char **)(&(((LINUX_DEVICE *)dev)->ip_ptr));
}

unsigned char * GCC_NOREGPARM mydev_fetch_dev_addr(VN_DEVICE dev)
{
  return (((LINUX_DEVICE *)dev)->dev_addr);
}

unsigned char ** GCC_NOREGPARM mynf_info_fetch_outdev(VN_NFINFO nfinfo)
{
#ifdef NF_USES_QUEUE_ENTRY
  return (unsigned char **)(&(((struct nf_queue_entry *)nfinfo)->outdev));
#else
  return (unsigned char **)(&(((struct nf_info *)nfinfo)->outdev));
#endif
}

unsigned char ** GCC_NOREGPARM mynf_info_fetch_indev(VN_NFINFO nfinfo)
{
#ifdef NF_USES_QUEUE_ENTRY
  return (unsigned char **)(&(((struct nf_queue_entry *)nfinfo)->indev));
#else
  return (unsigned char **)(&(((struct nf_info *)nfinfo)->indev));
#endif
}

/* Interface structures. */
unsigned char ** GCC_NOREGPARM myin_dev_fetch_ifa_list(VN_INDEVICE indev)
{
  return (unsigned char **)(&(((struct in_device *)indev)->ifa_list));
}

unsigned char ** GCC_NOREGPARM myin_ifaddr_fetch_ifa_next(VN_INIFADDR ifa)
{
  return (unsigned char **)(&(((struct in_ifaddr *)ifa)->ifa_next));
}

unsigned char * GCC_NOREGPARM myin_ifaddr_fetch_ifa_label(VN_INIFADDR ifa)
{
  return (((struct in_ifaddr *)ifa)->ifa_label);
}

unsigned int * GCC_NOREGPARM myin_ifaddr_fetch_ifa_local(VN_INIFADDR ifa)
{
  return &(((struct in_ifaddr *)ifa)->ifa_local);
}

unsigned int GCC_NOREGPARM myin_ifaddr_fetch_ifa_broadcast( VN_INIFADDR ifa )
{
     return ((struct in_ifaddr *)ifa)->ifa_broadcast;
}

VN_DEVICE GCC_NOREGPARM myin_ifaddr_fetch_netdev(VN_INIFADDR ifa)
{
     return ((struct in_ifaddr*)ifa)->ifa_dev->dev;
}

/* Neighbour and it's contants. */
unsigned char * GCC_NOREGPARM myneighbour_fetch_nud_state(VN_NEIGHBOUR neigh)
{
  return &(((struct neighbour *)neigh)->nud_state);
}

int myNUD_INCOMPLETE(void)
{
  return NUD_INCOMPLETE;
}

int myNUD_NOARP(void)
{
  return NUD_NOARP;
}

int myNUD_FAILED(void)
{
  return NUD_FAILED;
}

int myNUD_REACHABLE(void)
{
  return NUD_REACHABLE;
}

VN_NEIGHTABLE myarp_tbl(void)
{
  return &arp_tbl;
}

/* PCI device. */
unsigned int * GCC_NOREGPARM mypci_dev_fetch_irq(VN_PCIDEV dev)
{
  return &(((struct pci_dev *)dev)->irq);
}

/* File operations. */
VN_FILEOPERATIONS myalloc_fops(void)
{
  struct file_operations *tmp;

  tmp = (struct file_operations *)mykmalloc(sizeof(struct file_operations));
  return tmp;
}

void GCC_NOREGPARM myfree_fops(VN_FILEOPERATIONS fop)
{
  mykfree(fop);
}

void GCC_NOREGPARM myfill_fops(VN_FILEOPERATIONS fop, void * open,
                               void * release, void * ioctl, void * write)
{
  memset(fop, 0, sizeof(struct file_operations));
  ((struct file_operations *)fop)->owner = THIS_MODULE;
  ((struct file_operations *)fop)->open = open;
  ((struct file_operations *)fop)->release = release;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36)
  ((struct file_operations *)fop)->unlocked_ioctl = ioctl;
#else
  ((struct file_operations *)fop)->ioctl = ioctl;
#endif
#ifdef HAVE_COMPAT_IOCTL
  ((struct file_operations *)fop)->compat_ioctl = ioctl;
#endif
#ifdef HAVE_COMPAT_IOCTL
  ((struct file_operations *)fop)->compat_ioctl = ioctl;
#endif
  ((struct file_operations *)fop)->write = write;
}

/* Timer structures. */
VN_TIMER myget_timer(void)
{
  struct timer_list * tm;

  tm = (struct timer_list *)mykmalloc(sizeof(struct timer_list));
  if (tm)
  {
#if defined(TIMER_INITIALIZER)
    INIT_LIST_HEAD(&tm->entry);
    init_timer(tm);
#else
    tm->prev = NULL;
    tm->next = NULL;
#endif
#if defined(KERNEL_USE_REGPARM)
//    myprintk("%s: got %08X timer_list\n", __FUNCTION__, (int)tm);
    tm->data = (unsigned long)mykmalloc(sizeof(struct _univ_handle));
    if (!tm->data)
    {
      udebug(( "itcs_drv: no memory for timer handle!" ));
      mykfree(tm);
      return NULL;
    }
//    myprintk("%s: got univ_handler %08X\n", __FUNCTION__, (int)tm->data);
    ((struct _univ_handle *)tm->data)->flclean = 0;
#endif
  }
  return tm;
}

void GCC_NOREGPARM myfree_timer(VN_TIMER timer)
{
  if (timer)
  {
    struct timer_list * tm = (struct timer_list *)timer;
#if defined(KERNEL_USE_REGPARM)
    if (((struct _univ_handle *)tm->data) != NULL)
      mykfree(((struct _univ_handle *)tm->data));
#endif
    mykfree(tm);
  }
}

void GCC_NOREGPARM myset_timer(VN_TIMER timer, int secondtenths, void* data,
                 void GCC_NOREGPARM (* function)(void*))
{
  if (timer)
  {
    struct timer_list * timer_entry = (struct timer_list *)timer;
#if defined(KERNEL_USE_REGPARM)
    if (!((struct _univ_handle *)timer_entry->data))
    {
      udebug(( "itcs_drv: timer handle data not initialized!" ));
      return;
    }
    ((struct _univ_handle *)timer_entry->data)->fn = (int GCC_NOREGPARM (*)(void *))function;
    ((struct _univ_handle *)timer_entry->data)->arg = data;
    timer_entry->function = (void (*)(unsigned long))univ_handle;
//    myprintk("%s: timer %08X, function %08X, data %08X\n", __FUNCTION__, (int)timer, (int)function, (int)data);
#else
    timer_entry->function = (void (*)(unsigned long)) function;
    timer_entry->data = (unsigned long) data;
#endif
    timer_entry->expires = secondtenths * HZ / 10 + jiffies;
  }
}

int GCC_NOREGPARM mymod_timer(VN_TIMER timer, int secondtenths )
{
  if (timer)
  {
    struct timer_list * timer_entry = (struct timer_list *)timer;
    return mod_timer(timer_entry, secondtenths * HZ / 10 + jiffies);
  }
  return 0;
}

/* Mutex. */
VN_MUTEX mymutex_alloc(void)
{
  struct semaphore *tmp;
  tmp = (struct semaphore *)mykmalloc(sizeof(struct semaphore));
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,37)
    init_MUTEX(tmp);
#else
    sema_init(tmp, 1);
#endif
  return tmp;
}

void GCC_NOREGPARM mymutex_free(VN_MUTEX mutex)
{
  mykfree(mutex);
}

VN_MUTEX mymutex_alloc_locked(void)
{
  struct semaphore *tmp;
  tmp = (struct semaphore *)mykmalloc(sizeof(struct semaphore));
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,37)
    init_MUTEX_LOCKED(tmp);
#else
    sema_init(tmp, 0);
#endif

  return tmp;
}

/* Spinlock. */
VN_SPINLOCK myspinlock_alloc(void)
{
  spinlock_t *tmp;

  tmp = (spinlock_t *)mykmalloc(sizeof(spinlock_t));
  if ( tmp )
    spin_lock_init(tmp);

  return tmp;
}

void GCC_NOREGPARM myspinlock_free(VN_SPINLOCK lock)
{
  mykfree(lock);
}

/* rwlock. */
VN_RWLOCK myrwlock_alloc(void)
{
  VN_MYRWLOCK tmp = (VN_MYRWLOCK )mykmalloc(sizeof(MYRWLOCK));
  if(tmp)
  {
#ifdef USE_RWLOCK
       rwlock_init(&tmp->lock);
#else /* #ifdef USE_RWLOCK */
       init_rwsem(&tmp->lock);
#ifdef DEBUG_LOCK
       tmp->count = 0;
       tmp->pid = -1;
       spin_lock_init(&tmp->struct_lock);
       tmp->owner_pid = -1;
       tmp->owner_ip = 0;
       tmp->read = -1;
#endif /* #idef DEBUG_LOCK */
#endif /* #ifndef USE_RWLOCK */
  }
  return tmp;
}

void GCC_NOREGPARM myrwlock_free(VN_RWLOCK rw_lock)
{
  mykfree(rw_lock);
}

#ifdef SLAB_LEAK_TEST
atomic_t kmem_alloc_num = ATOMIC_INIT(0);

void mykmem_print_stats() {
     myprintk("Currently %d elements is allocated\n", atomic_read(&kmem_alloc_num));
}

static void mykmem_leak_test(void *ptr) {
     return;
}
#endif

VN_KMEM_CACHE GCC_NOREGPARM mykmem_cache_create(const char *name, int size, int offset )
{
     return (VN_KMEM_CACHE) kmem_cache_create( name, size, offset, SLAB_HWCACHE_ALIGN
#ifdef SLAB_LEAK_TEST
          /* pass the ctor parameter to get SLUB to allocate a separate slab */
          , mykmem_leak_test
#else
          , NULL
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23)
          ,NULL
#endif
     );
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,23)
     #define kmem_cache_t struct kmem_cache
#endif
void GCC_NOREGPARM mykmem_cache_destroy( VN_KMEM_CACHE cache )
{
     kmem_cache_destroy( (kmem_cache_t *) cache );
}

int GCC_NOREGPARM mykmem_cache_shrink( VN_KMEM_CACHE cache )
{
     return kmem_cache_shrink( (kmem_cache_t *) cache );
}

void* GCC_NOREGPARM mykmem_cache_alloc_kernel( VN_KMEM_CACHE cache )
{
#ifdef SLAB_LEAK_TEST
     atomic_inc(&kmem_alloc_num);
#endif
#ifdef USE_RWLOCK
     /*
      * This function can be called when rw_lock is held. Therefore it can not sleep
      */
     return kmem_cache_alloc( (kmem_cache_t *) cache, GFP_ATOMIC );
#else /* #ifdef USE_RWLOCK */
     return kmem_cache_alloc( (kmem_cache_t *) cache, GFP_KERNEL );
#endif /* #ifndef USE_RWLOCK */
}

void* GCC_NOREGPARM mykmem_cache_alloc_atomic( VN_KMEM_CACHE cache )
{
#ifdef SLAB_LEAK_TEST
     atomic_inc(&kmem_alloc_num);
#endif
     return kmem_cache_alloc( (kmem_cache_t *) cache, GFP_ATOMIC );
}

void GCC_NOREGPARM mykmem_cache_free( VN_KMEM_CACHE cache, void* obj )
{
#ifdef SLAB_LEAK_TEST
     atomic_dec(&kmem_alloc_num);
#endif
     kmem_cache_free( (kmem_cache_t *) cache, obj );
}

int GCC_NOREGPARM mytask_get_pid( VN_TASK_STRUCT tsk )
{
     return ((struct task_struct *)tsk)->pid;
}

/* Various constants. */
int myPAGE_SIZE(void)
{
  return PAGE_SIZE;
}

int myENODEV(void)
{
  return ENODEV;
}

int myEINVAL(void)
{
  return EINVAL;
}

int myEFAULT(void)
{
  return EFAULT;
}

int myERESTARTSYS(void)
{
  return ERESTARTSYS;
}

int myARPOP_REQUEST(void)
{
  return ARPOP_REQUEST;
}

int myETH_P_IP(void)
{
  return ETH_P_IP;
}

int myETH_P_ARP(void)
{
  return ETH_P_ARP;
}

int myETH_P_IPV6(void)
{
  return ETH_P_IPV6;
}

int myARPHRD_ETHER(void)
{
  return ARPHRD_ETHER;
}

int myARPHRD_PPP(void)
{
  return ARPHRD_PPP;
}

int myARPHRD_SLIP(void)
{
  return ARPHRD_SLIP;
}

int myARPHRD_CSLIP(void)
{
  return ARPHRD_CSLIP;
}

int myARPHRD_SLIP6(void)
{
  return ARPHRD_SLIP6;
}

int myARPHRD_CSLIP6(void)
{
  return ARPHRD_CSLIP6;
}

// IPv6-in-IPv4 tunnels
int myARPHRD_SIT(void)
{
  return ARPHRD_SIT;
}


int myETH_HLEN(void)
{
  return ETH_HLEN;
}

int myETH_ALEN(void)
{
  return ETH_ALEN;
}

int myIFF_LOOPBACK(void)
{
  return IFF_LOOPBACK;
}

int myCHECKSUM_UNNECESSARY(void)
{
  return CHECKSUM_UNNECESSARY;
}

int myCHECKSUM_NONE(void)
{
  return CHECKSUM_NONE;
}

int myCHECKSUM_COMPLETE(void)
{
#if defined(CHECKSUM_COMPLETE)
  return CHECKSUM_COMPLETE;
#elif defined(CHECKSUM_HW)
  return CHECKSUM_HW;
#else
  return -1;
#endif
}

int myPACKET_HOST(void)
{
  return PACKET_HOST;
}

int myPACKET_BROADCAST(void)
{
  return PACKET_BROADCAST;
}

int myPACKET_LOOPBACK(void)
{
  return PACKET_LOOPBACK;
}

int myRTN_BROADCAST(void)
{
  return RTN_BROADCAST;
}

// all supported medium types but Ethernet
int GCC_NOREGPARM IsSupportedOtherMediumType( int medium )
{
	switch( medium )
	{
		case ARPHRD_PPP:
		case ARPHRD_SLIP:
		case ARPHRD_SLIP6:
		case ARPHRD_CSLIP6:
		case ARPHRD_SIT:
		{
			return 1;
		}
	}

	return 0;
}

unsigned long long GCC_NOREGPARM __udivdi3(unsigned long long a, unsigned long b)
{
  unsigned long long res=a;
  do_div(res,b);
  return res;
}

unsigned long GCC_NOREGPARM __umoddi3(unsigned long long a, unsigned long b)
{
  unsigned long long res=a;
  unsigned long mod;
  mod = do_div(res,b);
  return mod;
}

void GCC_NOREGPARM myskb_set_transport_header(VN_SKB skbuff, int offset)
{
     struct sk_buff *skb = (struct sk_buff *)skbuff;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
     skb->h.raw = skb->data + offset;
#else
     skb_set_transport_header( skb, offset );
#endif
}

int GCC_NOREGPARM myis_arp_protocol( VN_SKB skbuff )
{
     enum
     {
    	  ETH_P_ARP_NETTOHOST = 0x100 * ( ETH_P_ARP & 0xFF ) + ( ETH_P_ARP >> 8 )
     };
	 return ( ((struct sk_buff *)skbuff )->protocol == ETH_P_ARP_NETTOHOST ? 1 : 0 );
}
