//watchdog driver
/** The module name. */
#define WD_DEV_NAME         "itcswd"


extern int GCC_NOREGPARM wddev_open(VN_STINODE inode, VN_STFILE file);
extern int GCC_NOREGPARM wddev_release(VN_STINODE inode, VN_STFILE file);
extern int GCC_NOREGPARM wddev_ioctl( int cmd, unsigned long arg );

/*
** stubs to fill file_operations struct while charaster device register
** (see comments about CONFIG_REGPARM)
*/
int my_wddev_open( struct inode *inode, struct file *file)
{
    return wddev_open((VN_STINODE)inode, (VN_STFILE)file);
}

int my_wddev_release( struct inode *inode, struct file *file)
{
    return wddev_release((VN_STINODE)inode, (VN_STFILE)file);
}

#ifdef HAVE_UNLOCKED_IOCTL
long my_wddev_ioctl( struct file *file, unsigned int cmd, unsigned long arg)
{
    return wddev_ioctl(cmd, arg);
}
#else
int my_wddev_ioctl( struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
    return wddev_ioctl(cmd, arg);
}
#endif


#ifdef CONFIG_ITCSDRV_AS_MISC
/** The file_operations structure. */
static struct file_operations gFileOpsWD =
{
owner:      THIS_MODULE,
open:       my_wddev_open,
release:    my_wddev_release,
#ifdef HAVE_UNLOCKED_IOCTL
unlocked_ioctl: my_wddev_ioctl,
#else
ioctl:      my_wddev_ioctl,
#endif
#ifdef HAVE_COMPAT_IOCTL
compat_ioctl: my_wddev_ioctl,
#endif
};

/** The miscdevice structure. */
static struct miscdevice g_WDDev =
{
minor:      MISC_DYNAMIC_MINOR,
name:       WD_DEV_NAME,
fops:       &gFileOpsWD,
# if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 17)
devfs_name: WD_DEV_NAME
# endif
};

/* static struct notifier_block itcswd_notifier = { */
/* 	.notifier_call	= watchdog_notify_sys, */
/* }; */
#else
static int wddev_major_number = 0;
#endif

static int itcswd_major = -1;
module_param(itcswd_major, int, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP);
MODULE_PARM_DESC(itcswd_major, "Major number for watchdogdriver");


extern int itcswd_real_init_module(void);
extern void itcswd_real_cleanup_module(void);

int __init itcswd_init_module(void)
{
	int rc=0;

	/* rc = register_reboot_notifier(&softdog_notifier); */
	/* if (rc) { */
	/* 	printk(KERN_ERR PFX */
	/* 		   "cannot register reboot notifier (err=%d)\n", ret); */
	/* 	return rc; */
	/* } */


#ifdef CONFIG_ITCSDRV_AS_MISC
    rc = misc_register(&g_WDDev);
    if (rc)
    {
        printk(KERN_ERR " Can't register " WD_DEV_NAME " device! rc=%d\n", rc);
        return rc;
    }
#else  /* !CONFIG_ITCSDRV_AS_MISC */
    /*
     * Register character device.
     */

    if (itcswd_major != -1)
    {
        if ((itcswd_major <= 0) || (itcswd_major > 254))
        {
            myprintk("WatchDogDrv: device major number is incorrectly specified (%d)\n",
                     itcswd_major);
            return myENODEV();
        }
        wddev_major_number = itcswd_major;
    }

    rc = register_chrdev((dev_t)wddev_major_number, WD_DEV_NAME, &gFileOpsWD);
    if (rc < 0)
    {

        myprintk("WatchDogDrv: cannot register device (code %d)\n", rc);
        return myENODEV();
    }

    /*
     * Save returned module major number
     */
    if ( wddev_major_number == 0 )
        wddev_major_number = rc;
    rc = 0;

# ifdef CONFIG_DEVFS_FS
    /*
     * Register a device entry
     */
    if (devfs_mk_cdev(MKDEV(wddev_major_number, 0), S_IFCHR | ITCS_DEV_FMASK, WD_DEV_NAME) != 0)
    {
        myprintk("devfs_register failed!\n");
        return myENODEV();
    }
# endif
    myprintk("WatchDogDrv: device number = %d, device node = /dev/"
             WD_DEV_NAME "\n", wddev_major_number);

#endif /* !CONFIG_ITCSDRV_AS_MISC */


    rc = itcswd_real_init_module();
    if  ( rc)
    {
        /** Failed, cleanup and return the error code.
         */
        myprintk("itcswd_real_init_module returning %#x)\n", rc);
#if defined(CONFIG_DEVFS_FS) && !defined(CONFIG_ITCSDRV_AS_MISC)
        devfs_remove(WD_DEV_NAME);
#endif

#ifdef CONFIG_ITCSDRV_AS_MISC
        misc_deregister(&g_WDDev);
#else
        unregister_chrdev(wddev_major_number, WD_DEV_NAME);
#endif
        return rc;
    }
    myprintk("ViPNet WatchDog Driver successfully loaded\n");

    return rc;
}

static void __exit itcswd_cleanup_module(void)
{
	int rc=0;

    itcswd_real_cleanup_module();
    myprintk("Unload\n");
    NOREF(rc);

/** I Don't think it's possible to unload a driver which processes have
 * opened, at least we'll blindly assume that here.
 */
#ifdef CONFIG_ITCSDRV_AS_MISC
    rc = misc_deregister(&g_WDDev);
    if (rc < 0)
    {
        myprintk("misc_deregister failed with rc=%#x\n", rc);
    }
#else  /* !CONFIG_ITCSDRV_AS_MISC */
# ifdef CONFIG_DEVFS_FS
    /*
     * Unregister a device entry
     */
    devfs_remove(WD_DEV_NAME);
# endif /* devfs */
    unregister_chrdev(wddev_major_number, WD_DEV_NAME);
#endif /* !CONFIG_ITCSDRV_AS_MISC */
}

module_init(itcswd_init_module);
module_exit(itcswd_cleanup_module);


